                     /*-----------------------------------*/
                     /*             >>>Slip<<<            */
                     /*            Theo D'Hondt           */
                     /*     VUB Software Languages Lab    */
                     /*             (c) 2010              */
                     /*-----------------------------------*/
                     /*      version 14: multicores       */
                     /*-----------------------------------*/
                     /*              compiler             */
                     /*-----------------------------------*/

#include "SlipMain.h"

#include "SlipCompile.h"
#include "SlipDictionary.h"
#include "SlipStack.h"

/*------------------------------ private prototypes --------------------------*/

static NIL_type compile_expression_C(EXP_type);
static NIL_type     compile_inline_C(EXP_type);
static NIL_type   compile_operands_C(PAI_type,
                                     TXT_type);
static NIL_type compile_parameters_C(PAI_type,
                                     TXT_type);
static NIL_type   compile_sequence_C(PAI_type,
                                     TXT_type);

/*----------------------------------------------------------------------------*/
/*-------------------------------- application -------------------------------*/
/*----------------------------------------------------------------------------*/

static NIL_type compile_application_C(PAI_type Application)
  { APL_type compiled_application;
    EXP_type compiled_operator,
             operands,
             operator;
    PAI_type application;
    VEC_type compiled_operands;
    TAG_type tag;
    operands = Application->cdr;
    tag = Grammar_Tag(operands);
    switch (tag)
      { case NUL_tag:
          operator = Application->car;
          compile_expression_C(operator);
          Main_Claim_Default();
          compiled_operator = Stack_Peek();
          compiled_application = make_APL_M(compiled_operator,
                                            Grammar_Empty_Vector);
          Main_Release_Default();
          Stack_Poke(compiled_application);
          return;
        case PAI_tag:
          application = Stack_Push_C(Application);
          operator = application->car;
          compile_expression_C(operator);
          application = Stack_Reduce();
          operands = application->cdr;
          compile_operands_C(operands,
                             Main_Application_String);
          Main_Claim_Default();
          compiled_operands = Stack_Pop();
          compiled_operator = Stack_Peek();
          compiled_application = make_APL_M(compiled_operator,
                                            compiled_operands);
          Main_Release_Default();
          Stack_Poke(compiled_application);
          return;
        default:
          Main_Error_Text(ITF_error,
                          Main_Application_String); }}

/*----------------------------------------------------------------------------*/
/*----------------------------------- begin ----------------------------------*/
/*----------------------------------------------------------------------------*/

static NIL_type compile_begin_C(PAI_type Operands)
  { BEG_type compiled_begin;
    EXP_type compiled_sequence;
    TAG_type tag;
    tag = Grammar_Tag(Operands);
    switch (tag)
      { case NUL_tag:
          Main_Error_Text(AL1_error,
                          Main_Begin_String);
        case PAI_tag:
          compile_sequence_C(Operands,
                             Main_Begin_String);
          Main_Claim_Default();
          compiled_sequence = Stack_Peek();
          compiled_begin = make_BEG_M(compiled_sequence);
          Main_Release_Default();
          Stack_Poke(compiled_begin);
          return;
        default:
          Main_Error_Text(ITF_error,
                          Main_Begin_String); }}

/*----------------------------------------------------------------------------*/
/*---------------------------------- define ----------------------------------*/
/*----------------------------------------------------------------------------*/

static NIL_type compile_define_function_C(EXP_type Operands)
  { EXP_type compiled_body,
             compiled_definition,
             residue;
    NBR_type frame_size,
             offset,
             param_count;
    PAI_type body,
             operands,
             parameters,
             pattern;
    SYM_type symbol,
             vararg;
    TAG_type tag;
    UNS_type raw_frame_size,
             raw_offset,
             raw_param_count;
    tag = Grammar_Tag(Operands);
    switch (tag)
      { case NUL_tag:
          Main_Error_Text(MSP_error,
                          Main_Define_String);
        case PAI_tag:
          Stack_Push_C(Operands);
          Main_Claim_Default();
          operands = Stack_Peek();
          pattern = operands->car;
          symbol = pattern->car;
          if (!is_SYM(symbol))
            Main_Error_Text(IVV_error,
                            Main_Define_String);
          raw_offset = Dictionary_Define_M(symbol);
          Dictionary_Enter_Nested_Scope_M();
          Main_Release_Default();
          body = operands->cdr;
          if (!is_PAI(body))
            Main_Error_Text(AL1_error,
                            Main_Define_String);
          parameters = pattern->cdr;
          compile_parameters_C(parameters,
                               Main_Define_String);
          raw_param_count = Dictionary_Get_Frame_Size();
          residue = Stack_Peek();
          tag = Grammar_Tag(residue);
          switch (tag)
            { case NUL_tag:
                Stack_Zap();
                operands = Stack_Peek();
                body = operands->cdr;
                compile_sequence_C(body,
                                   Main_Define_String);
                raw_frame_size = Dictionary_Exit_Nested_Scope();
                offset = make_NBR(raw_offset);
                param_count = make_NBR(raw_param_count);
                frame_size = make_NBR(raw_frame_size);
                Main_Claim_Default();
                compiled_body = Stack_Pop();
                operands = Stack_Peek();
                pattern = operands->car;
                symbol = pattern->car;
                compiled_definition = make_DFF_M(symbol,
                                                 offset,
                                                 param_count,
                                                 frame_size,
                                                 compiled_body);
                Main_Release_Default();
                Stack_Poke(compiled_definition);
                return;
              case SYM_tag:
                Main_Claim_Default();
                vararg = Stack_Pop();
                Dictionary_Define_M(vararg);
                Main_Release_Default();
                operands = Stack_Peek();
                body = operands->cdr;
                compile_sequence_C(body,
                                   Main_Define_String);
                raw_frame_size = Dictionary_Exit_Nested_Scope();
                offset = make_NBR(raw_offset);
                param_count = make_NBR(raw_param_count);
                frame_size = make_NBR(raw_frame_size);
                Main_Claim_Default();
                compiled_body = Stack_Pop();
                operands = Stack_Peek();
                pattern = operands->car;
                symbol = pattern->car;
                compiled_definition = make_DFZ_M(symbol,
                                                 offset,
                                                 param_count,
                                                 frame_size,
                                                 compiled_body);
                Main_Release_Default();
                Stack_Poke(compiled_definition);
                return;
              default:
                Main_Error_Text(IPA_error,
                                Main_Define_String); }
        default:
          Main_Error_Text(ITF_error,
                          Main_Define_String); }}

static NIL_type compile_define_variable_C(PAI_type Operands)
  { DFV_type compiled_definition;
    EXP_type compiled_expression,
             expression,
             residue;
    NBR_type offset;
    PAI_type expressions,
             operands;
    SYM_type symbol;
    TAG_type tag;
    UNS_type raw_offset;
    expressions = Operands->cdr;
    tag = Grammar_Tag(expressions);
    switch (tag)
      { case NUL_tag:
          Main_Error_Text(TFX_error,
                          Main_Define_String);
        case PAI_tag:
          residue = expressions->cdr;
          tag = Grammar_Tag(residue);
          switch (tag)
            { case NUL_tag:
                operands = Stack_Push_C(Operands);
                expressions = operands->cdr;
                expression = expressions->car;
                compile_expression_C(expression);
                Main_Claim_Default();
                operands = Stack_Reduce();
                symbol = operands->car;
                raw_offset = Dictionary_Define_M(symbol);
                offset = make_NBR(raw_offset);
                compiled_expression = Stack_Peek();
                compiled_definition = make_DFV_M(offset,
                                                 compiled_expression);
                Main_Release_Default();
                Stack_Poke(compiled_definition);
                return;
              case PAI_tag:
                Main_Error_Text(TMX_error,
                                Main_Define_String);
              default:
                Main_Error_Text(ITF_error,
                                Main_Define_String); }
        default:
          Main_Error_Text(ITF_error,
                          Main_Define_String); }}

static NIL_type compile_define_C(PAI_type Operands)
  { EXP_type pattern;
    TAG_type tag;
    tag = Grammar_Tag(Operands);
    switch (tag)
      { case NUL_tag:
          Main_Error_Text(MSP_error,
                          Main_Define_String);
        case PAI_tag:
          pattern = Operands->car;
          tag = Grammar_Tag(pattern);
          switch (tag)
            { case PAI_tag:
                compile_define_function_C(Operands);
                return;
              case SYM_tag:
                compile_define_variable_C(Operands);
                return;
              default:
                Main_Error_Text(IVP_error,
                                Main_Define_String); }
        default:
          Main_Error_Text(ITF_error,
                          Main_Define_String); }}

/*----------------------------------------------------------------------------*/
/*------------------------------------ if ------------------------------------*/
/*----------------------------------------------------------------------------*/

static NIL_type compile_if_C(PAI_type Operands)
  { EXP_type alternative,
             compiled_alternative,
             compiled_consequent,
             compiled_if,
             compiled_predicate,
             consequent,
             predicate,
             residue;
    PAI_type alternatives,
             clauses,
             operands;
    TAG_type tag;
    tag = Grammar_Tag(Operands);
    switch (tag)
      { case NUL_tag:
          Main_Error_Text(TFX_error,
                          Main_If_String);
        case PAI_tag:
          clauses = Operands->cdr;
          tag = Grammar_Tag(clauses);
          switch (tag)
            { case NUL_tag:
                Main_Error_Text(TFX_error,
                                Main_If_String);
              case PAI_tag:
                operands = Stack_Push_C(Operands);
                predicate = operands->car;
                compile_expression_C(predicate);
                operands = Stack_Reduce();
                clauses = operands->cdr;
                alternatives = clauses->cdr;
                tag = Grammar_Tag(alternatives);
                switch (tag)
                  { case NUL_tag:
                      consequent = clauses->car;
                      compile_inline_C(consequent);
                      Main_Claim_Default();
                      compiled_consequent = Stack_Pop();
                      compiled_predicate = Stack_Peek();
                      compiled_if = make_IFZ_M(compiled_predicate,
                                               compiled_consequent);
                      Main_Release_Default();
                      Stack_Poke(compiled_if);
                      return;
                    case PAI_tag:
                      residue = alternatives->cdr;
                      if (!is_NUL(residue))
                        Main_Error_Text(TMX_error,
                                        Main_If_String);
                      clauses = Stack_Push_C(clauses);
                      consequent = clauses->car;
                      compile_inline_C(consequent);
                      clauses = Stack_Reduce();
                      alternatives = clauses->cdr;
                      alternative = alternatives->car;
                      compile_inline_C(alternative);
                      Main_Claim_Default();
                      compiled_alternative = Stack_Pop();
                      compiled_consequent = Stack_Pop();
                      compiled_predicate = Stack_Peek();
                      compiled_if = make_IFF_M(compiled_predicate,
                                               compiled_consequent,
                                               compiled_alternative);
                      Main_Release_Default();
                      Stack_Poke(compiled_if);
                      return;
                    default:
                      Main_Error_Text(ITF_error,
                                      Main_If_String); }
              default:
                Main_Error_Text(ITF_error,
                                Main_If_String); }
        default:
          Main_Error_Text(ITF_error,
                          Main_If_String); }}

/*----------------------------------------------------------------------------*/
/*----------------------------- inline expression ----------------------------*/
/*----------------------------------------------------------------------------*/

static NIL_type compile_inline_C(EXP_type Expression)
  { EXP_type compiled_expression,
             expression;
    NBR_type frame_size;
    THK_type thunk;
    UNS_type raw_frame_size;
    Stack_Push_C(Expression);
    Dictionary_Enter_Nested_Scope_C();
    expression = Stack_Peek();
    compile_expression_C(expression);
    raw_frame_size = Dictionary_Exit_Nested_Scope();
    if (raw_frame_size == 0)
      { Stack_Zap();
        expression = Stack_Pop();
        compile_expression_C(expression);
        return; }   
    frame_size = make_NBR(raw_frame_size);
    Main_Claim_Default();
    compiled_expression = Stack_Pop();
    thunk = make_THK_M(compiled_expression,
                       frame_size);
    Main_Release_Default();
    Stack_Poke(thunk); }
    
/*----------------------------------------------------------------------------*/
/*------------------------------ inline sequence -----------------------------*/
/*----------------------------------------------------------------------------*/

static NIL_type compile_inline_sequence_C(PAI_type Sequence,
                                          TXT_type Error_string)
  { EXP_type compiled_sequence;
    NBR_type frame_size;
    PAI_type sequence;
    THK_type thunk;
    UNS_type raw_frame_size;
    Stack_Push_C(Sequence);
    Dictionary_Enter_Nested_Scope_C();
    sequence = Stack_Peek();
    compile_sequence_C(sequence,
                       Error_string);
    raw_frame_size = Dictionary_Exit_Nested_Scope();
    if (raw_frame_size == 0)
      { Stack_Zap();
        sequence = Stack_Pop();
        compile_sequence_C(sequence,
                           Error_string);
        return; }   
    frame_size = make_NBR(raw_frame_size);
    Main_Claim_Default();
    compiled_sequence = Stack_Pop();
    thunk = make_THK_M(compiled_sequence,
                       frame_size);
    Main_Release_Default();
    Stack_Poke(thunk); }
                        
/*----------------------------------------------------------------------------*/
/*----------------------------------- lambda ---------------------------------*/
/*----------------------------------------------------------------------------*/

static NIL_type compile_lambda_C(PAI_type Operands)
  { EXP_type compiled_body,
             compiled_lambda,
             parameters,
             residue;
    NBR_type frame_size,
             param_count;
    PAI_type body,
             operands;
    SYM_type vararg;
    TAG_type tag;
    UNS_type raw_frame_size,
             raw_param_count;
    tag = Grammar_Tag(Operands);
    switch (tag)
      { case NUL_tag:
          Main_Error_Text(TFX_error,
                          Main_Lambda_String);
        case PAI_tag:
          Stack_Push_C(Operands);
          Dictionary_Enter_Nested_Scope_C();
          operands = Stack_Peek();
          body = operands->cdr;
          if (!is_PAI(body))
            Main_Error_Text(AL1_error,
                            Main_Lambda_String);
          parameters = operands->car;
          compile_parameters_C(parameters,
                               Main_Lambda_String);
          raw_param_count = Dictionary_Get_Frame_Size();
          residue = Stack_Peek();
          tag = Grammar_Tag(residue);
          switch (tag)
            { case NUL_tag:
                Stack_Zap();
                operands = Stack_Peek();
                body = operands->cdr;
                compile_sequence_C(body,
                                   Main_Lambda_String);
                raw_frame_size = Dictionary_Exit_Nested_Scope();
                param_count = make_NBR(raw_param_count);
                frame_size = make_NBR(raw_frame_size);
                Main_Claim_Default();
                compiled_body = Stack_Pop();
                compiled_lambda = make_LMB_M(param_count,
                                             frame_size,
                                             compiled_body);
                Main_Release_Default();
                Stack_Poke(compiled_lambda);
                return;
              case SYM_tag:
                Main_Claim_Default();
                vararg = Stack_Pop();
                Dictionary_Define_M(vararg);
                Main_Release_Default();
                operands = Stack_Peek();
                body = operands->cdr;
                compile_sequence_C(body,
                                   Main_Lambda_String);
                raw_frame_size = Dictionary_Exit_Nested_Scope();
                param_count = make_NBR(raw_param_count);
                frame_size = make_NBR(raw_frame_size);
                Main_Claim_Default();
                compiled_body = Stack_Pop();
                compiled_lambda = make_LMZ_M(param_count,
                                             frame_size,
                                             compiled_body);
                Main_Release_Default();
                Stack_Poke(compiled_lambda);
                return;
              default:
                Main_Error_Text(IPA_error,
                                Main_Lambda_String); }
        default:
          Main_Error_Text(ITF_error,
                          Main_Lambda_String); }}

/*----------------------------------------------------------------------------*/
/*--------------------------------- operands ---------------------------------*/
/*----------------------------------------------------------------------------*/

static NIL_type compile_operands_C(PAI_type Operands,
                                   TXT_type Error_string)
  { EXP_type compiled_expression,
             expression;
    PAI_type operands;
    UNS_type count;
    VEC_type compiled_operands;
    operands = Operands;
    count = 0;
    do
      { operands = Stack_Push_C(operands);
        expression = operands->car;
        compile_expression_C(expression);
        operands = Stack_Reduce();
        operands = operands->cdr;
        ++count; }
    while (is_PAI(operands));
    if (!is_NUL(operands))
      Main_Error_Text(ITF_error,
                      Error_string);
    Main_Claim_Dynamic(count);
    compiled_operands = make_VEC_M(count);
    Main_Release_Dynamic(count);
    while (count > 0)
      { compiled_expression = Stack_Pop();
        compiled_operands[count--] = compiled_expression; }
    Stack_Push_C(compiled_operands); }

/*----------------------------------------------------------------------------*/
/*-------------------------------- parameters --------------------------------*/
/*----------------------------------------------------------------------------*/

static NIL_type compile_parameters_C(PAI_type Parameters,
                                     TXT_type Error_string)
  { EXP_type parameter;
    PAI_type parameters;
    parameters = Stack_Push_C(Parameters);
    while (is_PAI(parameters))
      { Main_Claim_Default();
        parameters = Stack_Peek();
        parameter = parameters->car;
        if (!is_SYM(parameter))
          Main_Error_Text(IPA_error,
                          Error_string);
        parameters = parameters->cdr;
        Stack_Poke(parameters);
        Dictionary_Define_M(parameter);
        Main_Release_Default(); }}

/*----------------------------------------------------------------------------*/
/*---------------------------------- protect ---------------------------------*/
/*----------------------------------------------------------------------------*/

static NIL_type compile_protect_C(PAI_type Operands)
  { EXP_type compiled_sequence;
    LCK_type lock;
    PRT_type compiled_protect;
    TAG_type tag;
    tag = Grammar_Tag(Operands);
    switch (tag)
      { case NUL_tag:
          Main_Error_Text(AL1_error,
                          Main_Protect_String);
        case PAI_tag:
          compile_sequence_C(Operands,
                             Main_Protect_String);
          Main_Claim_Default();
          lock = Main_Make_Lock_M();
          compiled_sequence = Stack_Peek();
          compiled_protect = make_PRT_M(compiled_sequence,
                                        lock);
          Main_Release_Default();
          Stack_Poke(compiled_protect);
          return;
        default:
          Main_Error_Text(ITF_error,
                          Main_Protect_String); }}

/*----------------------------------------------------------------------------*/
/*----------------------------------- quote ----------------------------------*/
/*----------------------------------------------------------------------------*/

static NIL_type compile_quote_C(PAI_type Operands)
  { EXP_type expression,
             residue;
    QUO_type compiled_quotation;
    TAG_type tag;
    tag = Grammar_Tag(Operands);
    switch (tag)
      { case NUL_tag:
          Main_Error_Text(E1X_error,
                          Main_Quote_String);
        case PAI_tag:
          expression = Operands->car;
          residue    = Operands->cdr;
          if (!is_NUL(residue))
            Main_Error_Text(TMX_error,
                            Main_Quote_String);
          Stack_Push_C(expression);
          Main_Claim_Default();
          expression = Stack_Peek();
          compiled_quotation = make_QUO_M(expression);
          Main_Release_Default();
          Stack_Poke(compiled_quotation);   /* clone!!! */
          return;
        default:
          Main_Error_Text(ITF_error,
                          Main_Quote_String); }}

/*----------------------------------------------------------------------------*/
/*--------------------------------- sequence ---------------------------------*/
/*----------------------------------------------------------------------------*/

static NIL_type compile_sequence_C(PAI_type Sequence,
                                   TXT_type Error_string)
  { EXP_type compiled_expression,
             expression;
    PAI_type sequence;
    SEQ_type compiled_sequence;
    UNS_type count;
    sequence = Sequence->cdr;
    if (is_NUL(sequence))
      { expression = Sequence->car;
        compile_expression_C(expression);
        return; }
    sequence = Sequence;
    count = 0;
    do
      { sequence = Stack_Push_C(sequence);
        expression = sequence->car;
        compile_expression_C(expression);
        sequence = Stack_Reduce();
        sequence = sequence->cdr;
        ++count; }
    while (is_PAI(sequence));
    if (!is_NUL(sequence))
      Main_Error_Text(ITF_error,
                      Error_string);
    Main_Claim_Dynamic(count);
    compiled_sequence = make_SEQ_M(count);
    Main_Release_Dynamic(count);
    while (count > 0)
      { compiled_expression = Stack_Pop();
        compiled_sequence[count--] = compiled_expression; }
    Stack_Push_C(compiled_sequence); }

/*----------------------------------------------------------------------------*/
/*------------------------------------ set -----------------------------------*/
/*----------------------------------------------------------------------------*/

static NIL_type compile_set_C(PAI_type Operands)
  { EXP_type compiled_expression,
             compiled_set,
             expression;
    NBR_type offset,
             scope;
    PAI_type expressions,
             operands,
             residue;
    SYM_type variable;
    TAG_type tag;
    UNS_type raw_offset,
             raw_scope;
    tag = Grammar_Tag(Operands);
    switch (tag)
      { case NUL_tag:
          Main_Error_Text(MSV_error,
                                 Main_Set_String);
        case PAI_tag:
          variable = Operands->car;
          if (!is_SYM(variable))
            Main_Error_Text(IVV_error,
                            Main_Set_String);
          expressions = Operands->cdr;
          tag = Grammar_Tag(expressions);
          switch (tag)
            { case NUL_tag:
                Main_Error_Text(E1X_error,
                                Main_Set_String);
              case PAI_tag:
                residue = expressions->cdr;
                if (!is_NUL(residue))
                  Main_Error_Text(TMX_error,
                                  Main_Set_String);
                operands = Stack_Push_C(Operands);
                expressions = operands->cdr;
                expression = expressions->car;
                compile_expression_C(expression);
                operands = Stack_Reduce();
                variable = operands->car;
                raw_offset = Dictionary_Lexical_Address(variable,
                                                        &raw_scope);
                if (raw_offset == 0)
                  Main_Error_Symbol(VNF_error,
                                    variable);
                offset = make_NBR(raw_offset);
                Main_Claim_Default();
                compiled_expression = Stack_Peek();
                if (raw_scope == 0)
                  compiled_set = make_STL_M(offset,
                                            compiled_expression);
                else
                  { scope = make_NBR(raw_scope);
                    compiled_set = make_STG_M(scope,
                                              offset,
                                              compiled_expression); }
                Main_Release_Default();
                Stack_Poke(compiled_set);
                return;
              default:
                Main_Error_Text(ITF_error,
                                Main_Set_String); }
        default:                                
          Main_Error_Text(ITF_error,
                          Main_Set_String); }}

/*----------------------------------------------------------------------------*/
/*------------------------------------ spawn ---------------------------------*/
/*----------------------------------------------------------------------------*/

static NIL_type compile_spawn_C(PAI_type Operands)
  { EXP_type compiled_spawn,
             compiled_expression,
             expression,
             residue;
    TAG_type tag;
    tag = Grammar_Tag(Operands);
    switch (tag)
      { case NUL_tag:
          Main_Error_Text(E1X_error,
                          Main_Spawn_String);
        case PAI_tag:
          residue = Operands->cdr;
          if (!is_NUL(residue))
            Main_Error_Text(E1X_error,
                            Main_Spawn_String);
          expression = Operands->car;
          compile_expression_C(expression);
          Main_Claim_Default();
          compiled_expression = Stack_Peek();
          compiled_spawn = make_SPN_M(compiled_expression);
          Main_Release_Default();
          Stack_Poke(compiled_spawn);
          return;
        default:
          Main_Error_Text(ITF_error,
                          Main_Spawn_String); }}

/*----------------------------------------------------------------------------*/
/*---------------------------------- symbol ----------------------------------*/
/*----------------------------------------------------------------------------*/

static NIL_type compile_symbol_C(SYM_type Symbol)
  { EXP_type compiled_variable;
    NBR_type offset,
             scope;
    UNS_type raw_offset,
             raw_scope;
    raw_offset = Dictionary_Lexical_Address(Symbol,
                                            &raw_scope);
    if (raw_offset == 0)
      Main_Error_Symbol(VNF_error,
                        Symbol);
    offset = make_NBR(raw_offset);
    Main_Claim_Default();
    if (raw_scope == 0)
      compiled_variable = make_LCL_M(offset);
    else
      { scope  = make_NBR(raw_scope);
        compiled_variable = make_GLB_M(scope,
                                       offset); }
    Main_Release_Default();
    Stack_Push_C(compiled_variable); }

/*----------------------------------------------------------------------------*/
/*----------------------------------- value ----------------------------------*/
/*----------------------------------------------------------------------------*/

static NIL_type compile_value_C(EXP_type Value)
  { Stack_Push_C(Value); }

/*----------------------------------------------------------------------------*/
/*----------------------------------- while ----------------------------------*/
/*----------------------------------------------------------------------------*/

static NIL_type compile_while_C(PAI_type Operands)
  { EXP_type compiled_body,
             compiled_predicate,
             predicate;
    PAI_type body,
             operands;
    TAG_type tag;
    WHI_type compiled_while;
    tag = Grammar_Tag(Operands);
    switch (tag)
      { case NUL_tag:
          Main_Error_Text(TFX_error,
                          Main_While_String);
        case PAI_tag:
          body = Operands->cdr;
          tag = Grammar_Tag(body);
          switch (tag)
            { case NUL_tag:
                Main_Error_Text(TFX_error,
                                Main_While_String);
              case PAI_tag:
                operands = Stack_Push_C(Operands);
                predicate = operands->car;
                compile_inline_C(predicate);
                operands = Stack_Reduce();
                body = operands->cdr;
                compile_inline_sequence_C(body,
                                          Main_While_String);
                Main_Claim_Default();
                compiled_body = Stack_Pop();
                compiled_predicate = Stack_Peek();
                compiled_while = make_WHI_M(compiled_predicate,
                                            compiled_body);
                Main_Release_Default();
                Stack_Poke(compiled_while);
                return;
              default:
                Main_Error_Text(ITF_error,
                                Main_While_String); }
        default:
          Main_Error_Text(ITF_error,
                          Main_While_String); }}

/*----------------------------------------------------------------------------*/
/*------------------------------------ form ----------------------------------*/
/*----------------------------------------------------------------------------*/

static NIL_type compile_form_C(PAI_type Form)
  { EXP_type operands,
             operator;
    operator = Form->car;
    operands = Form->cdr;
    if (operator == Main_Begin)
      compile_begin_C(operands);
    else if (operator == Main_Define)
      compile_define_C(operands);
    else if (operator == Main_If)
      compile_if_C(operands);
    else if (operator == Main_Lambda)
      compile_lambda_C(operands);
    else if (operator == Main_Protect)
      compile_protect_C(operands);
    else if (operator == Main_Quote)
      compile_quote_C(operands);
    else if (operator == Main_Set)
      compile_set_C(operands);
    else if (operator == Main_Spawn)
      compile_spawn_C(operands);
    else if (operator == Main_While)
      compile_while_C(operands);
    else
      compile_application_C(Form); }

/*----------------------------------------------------------------------------*/
/*--------------------------------- expression -------------------------------*/
/*----------------------------------------------------------------------------*/

static NIL_type compile_expression_C(EXP_type Expression)
  { TAG_type tag;
    tag = Grammar_Tag(Expression);
    switch (tag)
      { case PAI_tag:
          compile_form_C(Expression);
          return;
        case SYM_tag:
          compile_symbol_C(Expression);
          return;
        case CHA_tag:
        case FLS_tag:
        case NBR_tag:
        case NUL_tag:
        case REA_tag:
        case STR_tag:
        case TRU_tag:
        case VEC_tag:
          compile_value_C(Expression);
          return;
        default:
          Main_Error_Tag(IXT_error,
                         tag); }}

/*------------------------------ public functions ----------------------------*/

EXP_type Compile_Compile_C(EXP_type Expression)
  { EXP_type compiled_expression;
    Stack_Reset();
    compile_expression_C(Expression);
    compiled_expression = Stack_Pop();
    return compiled_expression; }

/*----------------------------------------------------------------------------*/
                               
NIL_type Compile_Initialize(NIL_type)
  { Dictionary_Initialize(); }
