                     /*-----------------------------------*/
                     /*             >>>Slip<<<            */
                     /*            Theo D'Hondt           */
                     /*     VUB Software Languages Lab    */
                     /*             (c) 2010              */
                     /*-----------------------------------*/
                     /*      version 14: multicores       */
                     /*-----------------------------------*/
                     /*            concurrency            */
                     /*-----------------------------------*/

#include <errno.h>

#include "SlipConcurrency.h"

/*-------------------------------- private types -----------------------------*/

typedef  int32_t   Sgn_type;
typedef  int32_t * Srf_type;

/*------------------------------ private functions ---------------------------*/

static NIL_type process_return_value(UNS_type Return_value)
  { BYT_type error;
    if (Return_value == 0)
      return;
    switch (Return_value)
      { case EAGAIN:
          error = PT1_error;
          break;
        case EBUSY:
          error = PT2_error;
          break;
        case EINVAL:
          error = PT3_error;
          break;
        case EDEADLK:
          error = PT4_error;
          break;
        case ENOMEM:
          error = PT5_error;
          break;
        case EPERM:
          error = PT6_error;
          break;
        case ESRCH:
          error = PT7_error;
          break;
        case ETIMEDOUT:
          error = PT8_error; 
          break;
        default:
          error = PT9_error; }
    Main_Fail(error); }

/*------------------------------ public functions ----------------------------*/

NIL_type Concurrency_Initialize_Conditional(CRF_type Conditional_reference) 
  { UNS_type return_value;
    return_value = pthread_cond_init(Conditional_reference,
                                     NULL);
    process_return_value(return_value); }

NIL_type Concurrency_Wait_Conditional(CRF_type Conditional_reference,
                                      MRF_type Mutex_reference      )
  { UNS_type return_value;
    return_value = pthread_cond_wait(Conditional_reference,
                                     Mutex_reference);
    process_return_value(return_value); }

NIL_type Concurrency_Broadcast_Conditional(CRF_type Conditional_reference)
  { UNS_type return_value;
    return_value = pthread_cond_broadcast(Conditional_reference);
    process_return_value(return_value); }

/*----------------------------------------------------------------------------*/

NIL_type Concurrency_Cancel(PTH_type Pthread)
  { UNS_type return_value;
    return_value = pthread_cancel(Pthread);
    process_return_value(return_value);
    return_value = pthread_join(Pthread,
                                NULL);
    process_return_value(return_value); }
                 
NIL_type Concurrency_Pthread(PRF_type Pthread_reference,
                             WRK_type Worker_function,
                             ADR_type Argument)
  { UNS_type return_value;
    return_value = pthread_create(Pthread_reference,
                                  NULL,
                                  Worker_function,  
                                  Argument);
    process_return_value(return_value); }
                   
NIL_type Concurrency_Join(PTH_type Pthread)
  { UNS_type return_value;
    return_value = pthread_join(Pthread,
                                NULL);
    process_return_value(return_value); }
                 
NIL_type Concurrency_Kill(NIL_type)
  { pthread_exit(NULL); } 
  
PTH_type Concurrency_Self(NIL_type)
  { return pthread_self(); }

/*----------------------------------------------------------------------------*/

NIL_type Concurrency_Initialize_Mutex(MRF_type Mutex_reference) 
  { UNS_type return_value;
    return_value = pthread_mutex_init(Mutex_reference,
                                      NULL);
    process_return_value(return_value); }

NIL_type Concurrency_Mutex_Lock(MRF_type Mutex_reference)
  { UNS_type return_value;
    return_value = pthread_mutex_lock(Mutex_reference);
    process_return_value(return_value); }
  
NIL_type Concurrency_Mutex_Unlock(MRF_type Mutex_reference)
  { UNS_type return_value;
    return_value = pthread_mutex_unlock(Mutex_reference);
    process_return_value(return_value); }

/*----------------------------------------------------------------------------*/

NIL_type Concurrency_Initialize_Spin(SRF_type Spin_reference) 
  { *Spin_reference = 0; }

NIL_type Concurrency_Spin_Lock(SRF_type Spin_reference)
  { OSSpinLockLock(Spin_reference); }
  
NIL_type Concurrency_Spin_Unlock(SRF_type Spin_reference)
  { OSSpinLockUnlock(Spin_reference); }

NIL_type Concurrency_Increment(REF_type Reference)
  { OSAtomicAdd32((Sgn_type)1, 
                  (Srf_type)Reference); }

NIL_type Concurrency_Decrement(REF_type Reference)
  { OSAtomicAdd32((Sgn_type)-1, 
                  (Srf_type)Reference); }
  
UNS_type Concurrency_Add(REF_type Reference,
                         UNS_type Increment)
  { return OSAtomicAdd32((Sgn_type)Increment, 
                         (Srf_type)Reference); }

UNS_type Concurrency_Subtract(REF_type Reference,
                              UNS_type Decrement)
  { return OSAtomicAdd32((Sgn_type)-Decrement, 
                         (Srf_type)Reference); }

/*----------------------------------------------------------------------------*/
                           
NIL_type Concurrency_Initialize(NIL_type)
  { return; }