                     /*-----------------------------------*/
                     /*             >>>Slip<<<            */
                     /*            Theo D'Hondt           */
                     /*     VUB Software Languages Lab    */
                     /*             (c) 2010              */
                     /*-----------------------------------*/
                     /*      version 14: multicores       */
                     /*-----------------------------------*/
                     /*            dictionary             */
                     /*-----------------------------------*/

#include <stdio.h>

#include "SlipMain.h"

#include "SlipContext.h"
#include "SlipDictionary.h"
        
/*------------------------------- Frm delarations ----------------------------*/

Define_Grammar_Frame(Frm);

Begin_Grammar_Frame(Frm);
  Define_Grammar_Slot(SYM, var);
  Define_Grammar_Slot(Frm, frm);
End_Grammar_Frame(Frm);

static Frm_type Empty_frame;

static Frm_type make_Frame_M(SYM_type Variable,
                             Frm_type Frame)
  { Frm_type frame;
    VEC_type vector;
    vector = make_VEC_M(Frm_size);
    frame = (Frm_type)vector;
    frame->var = Variable;
    frame->frm = Frame;
    return frame; }

static NIL_type initialize_Frm(NIL_type)
  { Empty_frame = (Frm_type)Grammar_Null; }

/*------------------------------- Env delarations ----------------------------*/

Define_Grammar_Frame(Env);

Begin_Grammar_Frame(Env);
  Define_Grammar_Slot(Frm, frm);
  Define_Grammar_Slot(NBR, siz);
  Define_Grammar_Slot(Env, env); 
End_Grammar_Frame(Env);

static Env_type Empty_environment;

static Env_type make_Environment_M(Frm_type Frame,
                                   NBR_type Size,
                                   Env_type Environment)
  { Env_type environment;
    VEC_type vector;
    vector = make_VEC_M(Env_size);
    environment = (Env_type)vector;
    environment->frm = Frame;
    environment->siz = Size;
    environment->env = Environment;
    return environment; }
                         
static NIL_type initialize_Env(NIL_type)
  { Empty_environment = (Env_type)Grammar_Null; }

/*------------------------------ private variables ---------------------------*/

static Env_type Current_environment;
static Frm_type Current_frame;
static Frm_type Global_frame;

static UNS_type Current_frame_size;
static UNS_type Current_scope_level;
static UNS_type Global_frame_size;

/*------------------------------ private functions ---------------------------*/

static UNS_type get_raw_offset(SYM_type Variable,
                               UNS_type Size,
                               Frm_type Frame)
  { Frm_type frame;
    UNS_type raw_offset;
    raw_offset = Size;
    for (frame = Frame;
         !is_NUL(frame);
         frame = frame->frm)
      { if (frame->var == Variable)
          return raw_offset;
        raw_offset -= 1; }
    return 0; }

/*------------------------------ public functions ----------------------------*/

NIL_type Dictionary_Adjust_Global_Frame_C()
  { if (Current_scope_level == 0)
      Context_Adjust_Global_Frame_C(Current_frame_size); }
      
NIL_type Dictionary_Checkpoint(NIL_type)
  { Global_frame = Current_frame;
    Global_frame_size = Current_frame_size; }

UNS_type Dictionary_Define_M(SYM_type Variable)
  { Current_frame = make_Frame_M(Variable,
                                 Current_frame);
    return ++Current_frame_size; }

NIL_type Dictionary_Enter_Nested_Scope_C(NIL_type)
  { NBR_type frame_size;
    frame_size = make_NBR(Current_frame_size);
    Main_Claim_Default();
    Current_environment = make_Environment_M(Current_frame,
                                             frame_size,
                                             Current_environment);
    Main_Release_Default();
    Current_frame = Empty_frame;
    Current_frame_size = 0;
    Current_scope_level += 1; }

NIL_type Dictionary_Enter_Nested_Scope_M(NIL_type)
  { NBR_type frame_size;
    frame_size = make_NBR(Current_frame_size);
    Current_environment = make_Environment_M(Current_frame,
                                             frame_size,
                                             Current_environment);
    Current_frame = Empty_frame;
    Current_frame_size = 0;
    Current_scope_level += 1; }

UNS_type Dictionary_Exit_Nested_Scope(NIL_type)
  { NBR_type frame_size;
    UNS_type previous_frame_size;
    previous_frame_size = Current_frame_size;
    frame_size = Current_environment->siz;
    Current_frame       = Current_environment->frm;
    Current_environment = Current_environment->env;
    Current_scope_level -= 1;
    Current_frame_size = get_NBR(frame_size);
    return previous_frame_size; }

UNS_type Dictionary_Get_Frame_Size(NIL_type)
  { return Current_frame_size; }

UNS_type Dictionary_Get_Global_Frame_Size(NIL_type)
  { return Global_frame_size; }

UNS_type Dictionary_Initially_Define_M(SYM_type Variable)
  { Current_frame = make_Frame_M(Variable,
                                 Current_frame);
    return ++Current_frame_size; }

UNS_type Dictionary_Lexical_Address(SYM_type Variable,
                                    REF_type Scope   )
  { Env_type environment;
    Frm_type frame;
    NBR_type frame_size;
    UNS_type raw_frame_size,
             raw_offset,
             raw_scope;
    raw_offset= get_raw_offset(Variable,
                               Current_frame_size,
                               Current_frame);
    if (raw_offset > 0)
      { *Scope = 0;
        return raw_offset; }
    if (Current_scope_level > 0)
      { raw_scope = Current_scope_level;
        for (environment = Current_environment;
             !is_NUL(environment);
             environment = environment->env)
          { frame_size = environment->siz;
            frame = environment->frm;
            raw_frame_size = get_NBR(frame_size);
            raw_offset = get_raw_offset(Variable,
                                raw_frame_size,
                                frame);
            if (raw_offset > 0)
              { *Scope = raw_scope;
                return raw_offset; }
            raw_scope--; }}
    return 0; }

NIL_type Dictionary_Rollback_M(NIL_type)
  { NBR_type frame_size;
    Current_frame = Global_frame;
    Current_frame_size = Global_frame_size;
    frame_size = make_NBR(Current_frame_size);
    Current_environment = make_Environment_M(Current_frame,
                                             frame_size,
                                             Empty_environment);
    Current_scope_level = 0; }

/*----------------------------------------------------------------------------*/
                               
NIL_type Dictionary_Initialize(NIL_type)
  { initialize_Frm();
    initialize_Env();
    Current_environment = Empty_environment;
    Current_scope_level = 0;
    Global_frame = Current_frame = Empty_frame;
    Global_frame_size = Current_frame_size = 0;
    Main_Register(Current_environment);
    Main_Register(Current_frame);
    Main_Register(Global_frame); }