                                                                                                                                                       /*-----------------------------------*/
                     /*             >>>Slip<<<            */
                     /*            Theo D'Hondt           */
                     /*     VUB Software Languages Lab    */
                     /*             (c) 2010              */
                     /*-----------------------------------*/
                     /*      version 14: multicores       */
                     /*-----------------------------------*/
                     /*            environment            */
                     /*-----------------------------------*/                                                                                                                                                                   /*-----------------------------------*/

#include <stdio.h>

#include "SlipMain.h"

#include "SlipEnvironment.h"

/*------------------------------ EPL declarations ----------------------------*/

typedef
  struct EPL { VEC_type vec[1]; } EPL;

static EPL_type make_EPL(UNS_type Size)
  { EPL_type environment_pool;
    VEC_type vector;
    vector = make_VEC_M(Size);
    environment_pool = (EPL_type)vector;
    return environment_pool; }
                         
/*------------------------------ ENV declarations ----------------------------*/

typedef
  struct ENV { FRM_type frm[1]; } ENV;

static UNS_type size_ENV(ENV_type Environment)
  { UNS_type size;
    VEC_type vector;
    vector = (VEC_type)Environment;
    size = size_VEC(vector);
    return size; }

/*------------------------------ FRM declarations ----------------------------*/

static UNS_type size_FRM(FRM_type Frame)
  { UNS_type size;
    VEC_type vector;
    vector = (VEC_type)Frame;
    size = size_VEC(vector);
    return size; }

/*------------------------------ private constants ---------------------------*/

enum { Global_frame_size_increment =  32,
         Initial_global_frame_size = 128,
                        Pool_size =   16 };

/*------------------------------ private variables ---------------------------*/

static VEC_type Null_vector;
static FRM_type New_global_frame;
       
/*------------------------------ public variables ----------------------------*/

ENV_type Environment_Global_Environment;
FRM_type Environment_Global_Frame;

/*------------------------------ private functions ---------------------------*/

static VEC_type acquire_vector_M(EPL_type Environment_pool,
                                 UNS_type Size)
  { VEC_type next_vector,
             vector;
    if (Size == 0)
      return Grammar_Empty_Vector;
    if (Size > Pool_size)
      vector = Null_vector;
    else
      vector = Environment_pool->vec[Size];
    if (is_NUL(vector))
      vector = make_VEC_M(Size);
    else
      { next_vector = vector[1];
        Environment_pool->vec[Size] = next_vector; }
    return vector; }

static NIL_type release_vector(EPL_type Environment_pool,
                               VEC_type Vector)
  { VEC_type next_vector;
    UNS_type size;
    size = size_VEC(Vector);
    if (size == 0)
      return;
    if (size > Pool_size)
      return;
    if (!is_MRK(Vector))
      { next_vector = Environment_pool->vec[size];
        Vector[1] = next_vector;
        Environment_pool->vec[size] = Vector; }}

static NIL_type flush_pool(EPL_type Environment_pool)
  { UNS_type index;
    for (index = 1;
         index <= Pool_size;
         index++)
       Environment_pool->vec[index] = Null_vector; }

/*------------------------------ public functions ----------------------------*/

BYT_type Environment_Adjust_Global_Frame_C(UNS_type Size)
  { BYT_type overflow;
    UNS_type new_size,
             old_size;
    old_size = size_FRM(Environment_Global_Frame);
    overflow = (old_size < Size);
    if (overflow)
      { new_size = old_size + Global_frame_size_increment;
        Main_Claim_Dynamic(new_size);
        New_global_frame = Main_Initialize_Vector_M(new_size);
        Main_Release_Dynamic(new_size); }
    return overflow; }

NIL_type Environment_Confirm_Global_Frame(NIL_type)
  { Main_Replace(Environment_Global_Frame,
                 New_global_frame);
    Environment_Global_Frame = New_global_frame;
    New_global_frame = Grammar_Null; }

NIL_type Environment_Flush(EPL_type Environment_pool)
  { flush_pool(Environment_pool); }

UNS_type Environment_Get_Environment_size(ENV_type Environment)
  { UNS_type size;
    size = size_ENV(Environment);
    return size; }

EXP_type Environment_Global_Get(ENV_type Environment,
                                UNS_type Scope,
                                UNS_type Offset)                      
  { EXP_type value;
    FRM_type frame;
    frame = Environment->frm[Scope];
    value = frame[Offset];
    return value; }

NIL_type Environment_Global_Set(ENV_type Environment,
                                EXP_type Value,
                                UNS_type Scope,
                                UNS_type Offset)
  { FRM_type frame;
    frame = Environment->frm[Scope];
    frame[Offset] = Value; }

NIL_type Environment_Grow_Environment(ENV_type New_environment,
                                      ENV_type Environment,
                                      FRM_type Frame)
  { FRM_type frame;
    UNS_type index,
             size;
    size = size_ENV(Environment);
    for (index = 1;
         index <= size;
         index++)
      { frame = Environment->frm[index];
        New_environment->frm[index] = frame; }
    apply_MRK(Frame);
    New_environment->frm[size + 1] = Frame; }
    
EXP_type Environment_Local_Get(FRM_type Frame,
                               UNS_type Offset)
  { EXP_type value;
    value = Frame[Offset];
    return value; }

NIL_type Environment_Local_Set(EXP_type Value,
                               FRM_type Frame,
                               UNS_type Offset)
  { Frame[Offset] = Value; }

ENV_type Environment_Make_M(UNS_type Size,
                            EPL_type Environment_pool)
  { ENV_type environment;
    VEC_type vector;
    vector = acquire_vector_M(Environment_pool,
                              Size);
    environment = (ENV_type)vector;
    return environment; }                            

FRM_type Environment_Make_Frame_M(UNS_type Size,
                                  EPL_type Environment_pool)
  { FRM_type frame;
    UNS_type index;
    VEC_type vector;
    vector = acquire_vector_M(Environment_pool,
                              Size);
    for (index = 1;
         index <= Size;
         index++)
      vector[index] = Null_vector;
    frame = (FRM_type)vector;
    return frame; }

UNS_type Environment_Pool_Size(NIL_type)
  { return Pool_size + 1; }

NIL_type Environment_Release_Frame(FRM_type Frame,
                                   EPL_type Environment_pool)
  { VEC_type vector;
    vector = (VEC_type)Frame;
    release_vector(Environment_pool,
                   vector); }

EPL_type Environment_Setup_M(NIL_type)
  { EPL_type environment_pool;
    environment_pool = make_EPL(Pool_size);
    flush_pool(environment_pool);
    return environment_pool; }

/*----------------------------------------------------------------------------*/

NIL_type Environment_Initialize_M(NIL_type)
  { Null_vector = (VEC_type)Grammar_Null;
    Environment_Global_Environment = (ENV_type)Grammar_Empty_Vector;
    Environment_Global_Frame = Main_Initialize_Vector_M(Initial_global_frame_size);
    New_global_frame = Null_vector;
    apply_MRK(Environment_Global_Frame);
    Main_Register(Environment_Global_Environment);
    Main_Register(Environment_Global_Frame);
    Main_Register(New_global_frame); }