                     /*-----------------------------------*/
                     /*             >>>Slip<<<            */
                     /*            Theo D'Hondt           */
                     /*     VUB Software Languages Lab    */
                     /*             (c) 2010              */
                     /*-----------------------------------*/
                     /*      version 14: multicores       */
                     /*-----------------------------------*/
                     /*              evaluate             */
                     /*-----------------------------------*/

#include "SlipMain.h"

#include "SlipContext.h"
#include "SlipEvaluate.h"
#include "SlipPrint.h"
#include "SlipThread.h"

/*------------------------------ private prototypes --------------------------*/

static NIL_type               evaluate_context_C(CID_type,
                                                 EXP_type);
static NIL_type            evaluate_expression_C(EXP_type,
                                                 CID_type,
                                                 EXP_type);
static NIL_type evaluate_direct_procedure_body_C(CID_type,
                                                 FRM_type,
                                                 EXP_type);
static NIL_type        evaluate_procedure_body_C(CID_type,
                                                 FRM_type,
                                                 EXP_type);

/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/
/*--------------------------------- A P P L Y --------------------------------*/
/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
/*------------------------------- continuation -------------------------------*/
/*----------------------------------------------------------------------------*/

static NBR_type Continue_continuation;

Thread_Begin_Frame(cNT);
  Thread_Frame_Slot(CNT, cnt);
Thread_End_Frame(cNT);

static NIL_type continue_continuation(CID_type Context_id)
  { cNT_type continuation_thread; 
    CNT_type continuation;
    ENV_type environment;
    FRM_type frame;
    THR_type thread;
    thread = Context_Thread_Peek(Context_id);
    continuation_thread = (cNT_type)thread;
    continuation = continuation_thread->cnt;
    Context_Thread_Zap(Context_id);
    environment = continuation->env;
    frame       = continuation->frm;
    thread      = continuation->thr;
    Context_Release_Current_Frame(Context_id);
    Context_Environment_Replace(Context_id,
                                environment,
                                frame);
    Context_Thread_Replace(Context_id,
                           thread); }

/*----------------------------------------------------------------------------*/

static NIL_type apply_continuation_C(CID_type Context_id,
                                     VEC_type Operands,
                                     EXP_type Tail_position_status)
  { cNT_type continuation_thread;
    CNT_type continuation;
    EXP_type expression;
    UNS_type operand_size;
    THR_type thread;
    VEC_type operands;
    operand_size = size_VEC(Operands);
    if (operand_size != 1)
      Main_Error(CRQ_error);
    Context_Save(Context_id,
                 Operands);    
    thread = Context_Thread_Poke_C(Context_id,
                                   Continue_continuation,
                                   Tail_position_status,
                                   cNT_size);
    operands = Context_Restore(Context_id);    
    continuation = Context_Get_Expression(Context_id);
    continuation_thread = (cNT_type)thread;
    continuation_thread->cnt = continuation;
    expression = operands[1];
    evaluate_expression_C(expression,
                          Context_id,
                          Grammar_True); }

/*----------------------------------------------------------------------------*/

static NIL_type apply_direct_continuation(CID_type Context_id,
                                          EXP_type Tail_position_status)
  { CNT_type continuation;
    ENV_type environment;
    EXP_type value;
    FRM_type frame;
    PAI_type arguments;
    THR_type thread;
    VEC_type vector;
    vector = Context_Get_Expression(Context_id);
    continuation = vector[1];
    arguments = vector[2];
    if (!is_PAI(arguments))
      Main_Error(CRQ_error);
    value = arguments->car;
    arguments  = arguments->cdr;
    if (!is_NUL(arguments))
      Main_Error(ITA_error);
    Context_Set_Expression(Context_id,
                           value);
    environment = continuation->env;
    frame       = continuation->frm;
    thread      = continuation->thr;      
    Context_Release_Current_Frame(Context_id);
    Context_Environment_Replace(Context_id,
                                environment,
                                frame);
    Context_Thread_Replace(Context_id,
                           thread); }
                        
/*----------------------------------------------------------------------------*/

static NIL_type initialize_continuation(NIL_type)
  { Continue_continuation = Context_Register(continue_continuation,
                                             cNT_size); }

/*----------------------------------------------------------------------------*/
/*----------------------------------- native ---------------------------------*/
/*----------------------------------------------------------------------------*/

static NBR_type Continue_native;

Thread_Begin_Frame(nAT);
  Thread_Frame_Slot(NAT, nat);
  Thread_Frame_Slot(VEC, opd);
  Thread_Frame_Slot(FRM, frm);
  Thread_Frame_Slot(NBR, pos);
Thread_End_Frame(nAT);

static NIL_type continue_native_C(CID_type Context_id)
  { nAT_type native_thread;
    EXP_type expression,
             tail_position_status,
             value;
    FRM_type frame;
    NAT_type native;
    NPR_type native_procedure;
    NBR_type position;
    UNS_type raw_position,
             size;
    THR_type thread;
    VEC_type operands;
    value = Context_Get_Expression(Context_id);
    thread = Context_Thread_Peek(Context_id);
    tail_position_status = Thread_Get_Tail(thread);
    native_thread = (nAT_type)thread;
    frame    = native_thread->frm;
    position = native_thread->pos;
    raw_position = get_NBR(position);
    frame[raw_position] = value;
    size = size_VEC(frame);
    if (raw_position == size)
      { native = native_thread->nat;
        native_procedure = native->npr;
        Context_Thread_Zap(Context_id);
        Context_Set_Expression(Context_id,
                               frame);
        native_procedure(Context_id,
                         tail_position_status);
        return; }
    thread = Context_Thread_Keep_C(Context_id);
    native_thread = (nAT_type)thread;
    operands = native_thread->opd;
    raw_position += 1;
    expression = operands[raw_position];
    position = make_NBR(raw_position);
    native_thread->pos = position;
    if (raw_position < size)
      tail_position_status = Grammar_False;
    evaluate_expression_C(expression,
                          Context_id,
                          tail_position_status); }

/*----------------------------------------------------------------------------*/

static NIL_type apply_native_C(CID_type Context_id,
                               VEC_type Operands,
                               EXP_type Tail_position_status)
  { nAT_type native_thread;
    EXP_type expression,
             tail_position_status;
    FRM_type frame;
    NAT_type native;
    NPR_type native_procedure;
    THR_type thread;
    UNS_type operand_size;
    VEC_type operands;
    operand_size = size_VEC(Operands);
    if (operand_size == 0)
      { native = Context_Get_Expression(Context_id);
        native_procedure = native->npr;
        Context_Thread_Zap(Context_id);
        Context_Set_Expression(Context_id,
                               Grammar_Empty_Vector);
        native_procedure(Context_id,
                         Tail_position_status);
        return; }
    Context_Save(Context_id,
                 Operands);    
    Context_Thread_Poke_C(Context_id,
                          Continue_native,
                          Tail_position_status,
                          nAT_size);
    frame = Context_Make_Frame_C(Context_id,
                                 operand_size);
    operands = Context_Restore(Context_id);    
    native = Context_Get_Expression(Context_id);
    thread = Context_Thread_Peek(Context_id);
    native_thread = (nAT_type)thread;
    native_thread->nat = native;
    native_thread->opd = operands;
    native_thread->frm = frame;
    native_thread->pos = Grammar_One;
    expression = operands[1];
    tail_position_status = (operand_size > 1) 
                             ? Grammar_False 
                             : Tail_position_status;
    evaluate_expression_C(expression,
                          Context_id,
                          tail_position_status); }   

/*----------------------------------------------------------------------------*/

static NIL_type apply_direct_native_C(CID_type Context_id,
                                      EXP_type Tail_position_status)
  { EXP_type value;
    FRM_type frame;
    NAT_type native;
    NPR_type native_procedure;
    PAI_type arguments;
    TXT_type name;
    UNS_type position,
             size;
    VEC_type vector;
    vector = Context_Get_Expression(Context_id);
    size = 0;
    for (arguments = vector[2];
         is_PAI(arguments);
         arguments = arguments->cdr)
      size += 1;
    if (!is_NUL(arguments))
      { native = vector[1];
        name = native->nam;
        Main_Error_Text(ITA_error,
                        name); }
    frame = Context_Make_Frame_C(Context_id,
                                 size);
    vector = Context_Get_Expression(Context_id);
    native    = vector[1];
    arguments = vector[2];
    for (position = 1;
         position <= size;
         position++)
      { value     = arguments->car;
        arguments = arguments->cdr;
        frame[position] = value; }
    Context_Set_Expression(Context_id,
                           frame);
    native_procedure = native->npr;
    native_procedure(Context_id,
                     Tail_position_status); }

/*----------------------------------------------------------------------------*/

static NIL_type initialize_native(NIL_type)
  { Continue_native = Context_Register(continue_native_C,
                                       nAT_size); }

/*----------------------------------------------------------------------------*/
/*---------------------------------- procedure -------------------------------*/
/*----------------------------------------------------------------------------*/

static NBR_type Continue_procedure;

Thread_Begin_Frame(pRC);
  Thread_Frame_Slot(PRC, prc);
  Thread_Frame_Slot(VEC, opd);
  Thread_Frame_Slot(FRM, frm);
  Thread_Frame_Slot(NBR, pos);
Thread_End_Frame(pRC);

/*----------------------------------------------------------------------------*/

static NIL_type continue_procedure_C(CID_type Context_id)
  { pRC_type procedure_thread;
    EXP_type expression,
             tail_position_status,
             value;
    FRM_type frame;
    NBR_type param_count,
             position;
    PRC_type procedure;
    THR_type thread;
    UNS_type raw_param_count,
             raw_position;
    VEC_type operands;
    value = Context_Get_Expression(Context_id);
    thread = Context_Thread_Peek(Context_id);
    tail_position_status = Thread_Get_Tail(thread);
    procedure_thread = (pRC_type)thread;
    procedure = procedure_thread->prc;
    frame     = procedure_thread->frm;
    position  = procedure_thread->pos;
    raw_position = get_NBR(position);
    frame[raw_position] = value;
    param_count = procedure->par;
    raw_param_count = get_NBR(param_count);
    if (raw_position == raw_param_count)
      { Context_Set_Expression(Context_id,
                               procedure);
        evaluate_procedure_body_C(Context_id,
                                  frame,
                                  tail_position_status); 
        return; }
    thread = Context_Thread_Keep_C(Context_id);
    procedure_thread = (pRC_type)thread;
    operands = procedure_thread->opd;
    raw_position += 1;
    expression = operands[raw_position];
    position = make_NBR(raw_position);
    procedure_thread->pos = position;
    if (raw_position < raw_param_count)
      tail_position_status = Grammar_False;
    evaluate_expression_C(expression,
                          Context_id,
                          tail_position_status); }

/*----------------------------------------------------------------------------*/

static NIL_type apply_procedure_C(CID_type Context_id,
                                  VEC_type Operands,
                                  EXP_type Tail_position_status)
  { pRC_type procedure_thread;
    EXP_type expression,
             tail_position_status;
    FRM_type frame;
    NBR_type frame_size,
             param_count;
    PRC_type procedure;
    THR_type thread;
    UNS_type operand_size,
             raw_frame_size,
             raw_param_count;
    VEC_type operands;
    procedure = Context_Get_Expression(Context_id);
    param_count = procedure->par;
    frame_size  = procedure->siz;
    raw_param_count = get_NBR(param_count);
    raw_frame_size = get_NBR(frame_size);
    operand_size = size_VEC(Operands);
    if (raw_param_count > operand_size)
      Main_Error_Procedure(TFO_error,
                           procedure);
    if (raw_param_count < operand_size)
      Main_Error_Procedure(TMO_error,
                           procedure);
    if (operand_size == 0)
      { frame = Context_Make_Frame_C(Context_id,
                                     raw_frame_size);
        evaluate_procedure_body_C(Context_id,
                                  frame,
                                  Tail_position_status);
        return; }
    Context_Save(Context_id,
                 Operands);    
    Context_Thread_Poke_C(Context_id,
                          Continue_procedure,
                          Tail_position_status,
                          pRC_size);
    frame = Context_Make_Frame_C(Context_id,
                                 raw_frame_size);
    operands = Context_Restore(Context_id);    
    procedure = Context_Get_Expression(Context_id);
    thread = Context_Thread_Peek(Context_id);
    procedure_thread = (pRC_type)thread;
    procedure_thread->prc = procedure;
    procedure_thread->opd = operands;
    procedure_thread->frm = frame;
    procedure_thread->pos = Grammar_One;
    expression = operands[1];
    tail_position_status = (operand_size > 1) 
                     ? Grammar_False 
                     : Tail_position_status;
    evaluate_expression_C(expression,
                          Context_id,
                          tail_position_status); }
                        
/*----------------------------------------------------------------------------*/

static NIL_type apply_direct_procedure_C(CID_type Context_id,
                                         EXP_type Tail_position_status)
  { EXP_type value;
    FRM_type frame;
    NBR_type frame_size,
             param_count;
    PAI_type arguments;
    PRC_type procedure;
    UNS_type raw_frame_size,
             raw_param_count,
             position;
    VEC_type vector;
    vector = Context_Get_Expression(Context_id);
    procedure = vector[1];
    frame_size = procedure->siz;
    raw_frame_size = get_NBR(frame_size);
    frame = Context_Make_Frame_C(Context_id,
                                 raw_frame_size);
    vector = Context_Get_Expression(Context_id);
    procedure = vector[1];
    arguments = vector[2];
    param_count = procedure->par;
    raw_param_count = get_NBR(param_count);
    for (position = 1;
         position <= raw_param_count;
         position++)
      { if (is_NUL(arguments))
          Main_Error_Procedure(TFO_error,
                               procedure);
        if (!is_PAI(arguments))
          Main_Error_Procedure(ITA_error,
                               procedure);
        value     = arguments->car;
        arguments = arguments->cdr;
        frame[position] = value; }
    if (!is_NUL(arguments))
      Main_Error_Procedure(TMO_error,
                           procedure);
    Context_Set_Expression(Context_id,
                           procedure);
    evaluate_direct_procedure_body_C(Context_id,
                                     frame,
                                     Tail_position_status); }
                                                                          
/*----------------------------------------------------------------------------*/

static NIL_type initialize_procedure(NIL_type)
  { Continue_procedure = Context_Register(continue_procedure_C,
                                          pRC_size); }

/*----------------------------------------------------------------------------*/
/*------------------------------- vararg procedure ---------------------------*/
/*----------------------------------------------------------------------------*/

static NBR_type Continue_vararg,
                Continue_vararg_procedure;

Thread_Begin_Frame(vRG);
  Thread_Frame_Slot(PRC, prc);
  Thread_Frame_Slot(VEC, opd);
  Thread_Frame_Slot(FRM, frm);
  Thread_Frame_Slot(NBR, pos);
  Thread_Frame_Slot(PAI, lst);
Thread_End_Frame(vRG);

/*----------------------------------------------------------------------------*/

static NIL_type continue_vararg_C(CID_type Context_id)
  { vRG_type procedure_thread;
    EXP_type expression,
             tail_position_status,
             value;
    FRM_type frame;
    NBR_type param_count,
             position;
    PAI_type list,
             pair;
    PRC_type procedure;
    THR_type thread;
    UNS_type operand_size,
             raw_param_count,
             raw_position;
    VEC_type operands;
    Main_Claim_Default();
    pair = make_PAI_M(Grammar_Null,
                      Grammar_Null);
    Main_Release_Default();
    value = Context_Get_Expression(Context_id);
    thread = Context_Thread_Peek(Context_id);
    tail_position_status = Thread_Get_Tail(thread);
    procedure_thread = (vRG_type)thread;
    procedure = procedure_thread->prc;
    operands  = procedure_thread->opd;
    list      = procedure_thread->lst;
    position  = procedure_thread->pos;
    pair->car = value;
    pair->cdr = list;
    raw_position = get_NBR(position);
    operand_size = size_VEC(operands);
    if (raw_position == operand_size)
      { Context_Set_Expression(Context_id,
                               procedure);
        param_count = procedure->par;
        raw_param_count = get_NBR(param_count);
        frame = procedure_thread->frm;
        list = Main_Reverse(pair);
        frame[raw_param_count + 1] = list;
        evaluate_procedure_body_C(Context_id,
                                  frame,
                                  tail_position_status); 
        return; }
    raw_position += 1;
    Context_Save(Context_id,
                 pair);    
    thread = Context_Thread_Keep_C(Context_id);
    list = Context_Restore(Context_id);    
    procedure_thread = (vRG_type)thread;
    operands = procedure_thread->opd;
    position = make_NBR(raw_position);
    procedure_thread->lst = list;
    procedure_thread->pos = position;
    expression = operands[raw_position];
    if (raw_position < operand_size)
      tail_position_status = Grammar_False;
    evaluate_expression_C(expression,
                          Context_id,
                          tail_position_status); }
                        
static NIL_type continue_vararg_procedure_C(CID_type Context_id)
  { vRG_type procedure_thread;
    EXP_type expression,
             tail_position_status,
             value;
    FRM_type frame;
    NBR_type param_count,
             position;
    PRC_type procedure;
    THR_type thread;
    UNS_type operand_size,
             raw_param_count,
             raw_position;             
    VEC_type operands;
    value = Context_Get_Expression(Context_id);
    thread = Context_Thread_Peek(Context_id);
    tail_position_status = Thread_Get_Tail(thread);
    procedure_thread = (vRG_type)thread;
    procedure = procedure_thread->prc;
    operands  = procedure_thread->opd;
    frame     = procedure_thread->frm;
    position  = procedure_thread->pos;
    raw_position = get_NBR(position);
    frame[raw_position] = value;
    operand_size = size_VEC(operands);
    if (raw_position == operand_size)
      { Context_Set_Expression(Context_id,
                               procedure);
        frame[raw_position] = Grammar_Null;
        evaluate_procedure_body_C(Context_id,
                                  frame,
                                  tail_position_status); 
        return; }
    param_count = procedure->par;
    raw_param_count = get_NBR(param_count);
    if (raw_position < raw_param_count)
      thread = Context_Thread_Keep_C(Context_id);
    else
      thread = Context_Thread_Patch_C(Context_id,
                                      Continue_vararg);
    procedure_thread = (vRG_type)thread;
    operands = procedure_thread->opd;
    raw_position += 1;
    expression = operands[raw_position];
    position = make_NBR(raw_position);
    procedure_thread->pos = position;
    if (raw_position < raw_param_count)
      tail_position_status = Grammar_False;
    evaluate_expression_C(expression,
                          Context_id,
                          tail_position_status); }

/*----------------------------------------------------------------------------*/

static NIL_type apply_vararg_procedure_C(CID_type Context_id,
                                         VEC_type Operands,
                                         EXP_type Tail_position_status)
  { vRG_type procedure_thread;
    EXP_type expression,
             tail_position_status;
    FRM_type frame;
    NBR_type c_function,
             frame_size,
             param_count;
    PRC_type procedure;
    THR_type thread;
    UNS_type operand_size,
             raw_frame_size,
             raw_param_count;
    VEC_type operands;
    procedure = Context_Get_Expression(Context_id);
    param_count = procedure->par;
    frame_size  = procedure->siz;
    raw_param_count = get_NBR(param_count);
    raw_frame_size = get_NBR(frame_size);
    operand_size = size_VEC(Operands);
    if (raw_param_count > operand_size)
      Main_Error_Procedure_Vararg(TFO_error,
                                  procedure);
    if (operand_size == 0)
      { frame = Context_Make_Frame_C(Context_id,
                                     raw_frame_size);
        frame[1] = Grammar_Null;
        evaluate_procedure_body_C(Context_id,
                                  frame,
                                  Tail_position_status);
        return; }
    Context_Save(Context_id,
                 Operands);    
    c_function = (raw_param_count == 0) 
                   ? Continue_vararg 
                   : Continue_vararg_procedure;    
    Context_Thread_Poke_C(Context_id,
                          c_function,
                          Tail_position_status,
                          vRG_size);
    frame = Context_Make_Frame_C(Context_id,
                                 raw_frame_size);
    operands = Context_Restore(Context_id);    
    thread = Context_Thread_Peek(Context_id);
    procedure_thread = (vRG_type)thread;
    procedure_thread->prc = procedure;
    procedure_thread->opd = operands;
    procedure_thread->frm = frame;
    procedure_thread->lst = Grammar_Null;
    procedure_thread->pos = Grammar_One;
    expression = operands[1];
    tail_position_status = (operand_size > 1) 
                     ? Grammar_False 
                     : Tail_position_status;
    evaluate_expression_C(expression,
                          Context_id,
                          tail_position_status); }

/*----------------------------------------------------------------------------*/

static NIL_type apply_direct_vararg_procedure_C(CID_type Context_id,
                                                EXP_type Tail_position_status)
  { EXP_type value;
    FRM_type frame;
    NBR_type frame_size,
             param_count;
    PAI_type arguments;
    PRC_type procedure;
    UNS_type position,
             raw_frame_size,
             raw_param_count;
    VEC_type vector;
    vector = Context_Get_Expression(Context_id);
    procedure = vector[1];
    frame_size = procedure->siz;
    raw_frame_size = get_NBR(frame_size);
    frame = Context_Make_Frame_C(Context_id,
                                 raw_frame_size);
    vector = Context_Get_Expression(Context_id);
    procedure = vector[1];
    arguments = vector[2];
    param_count = procedure->par;
    raw_param_count = get_NBR(param_count);
    for (position = 1;
         position <= raw_param_count;
         position++)
      { if (!is_PAI(arguments))
          Main_Error_Procedure_Vararg(TFO_error,
                                      procedure);
        value     = arguments->car;
        arguments = arguments->cdr;
        frame[position] = value; }
    frame[position] = arguments;
    while (is_PAI(arguments))
      arguments = arguments->cdr;
    if (!is_NUL(arguments))
      Main_Error(ITA_error);      
    Context_Set_Expression(Context_id,
                           procedure);
    evaluate_direct_procedure_body_C(Context_id,
                                     frame,
                                     Tail_position_status); }

/*----------------------------------------------------------------------------*/

static NIL_type initialize_vararg_procedure(NIL_type)
  { Continue_vararg = Context_Register(continue_vararg_C,
                                       vRG_size);
    Continue_vararg_procedure = Context_Register(continue_vararg_procedure_C,
                                                 vRG_size); }

/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/
/*------------------------------ E V A L U A T E -----------------------------*/
/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
/*-------------------------------- application -------------------------------*/
/*----------------------------------------------------------------------------*/

static const TXT_type Application_String = "application";

static NBR_type Continue_application;

Thread_Begin_Frame(aPL);
  Thread_Frame_Slot(VEC, opd);
Thread_End_Frame(aPL);

/*----------------------------------------------------------------------------*/

static NIL_type continue_application_C(CID_type Context_id)
  { aPL_type application_thread;
    EXP_type operator,
             tail_position_status;
    TAG_type tag;
    THR_type thread;
    VEC_type operands;
    thread = Context_Thread_Peek(Context_id);
    tail_position_status = Thread_Get_Tail(thread);
    application_thread = (aPL_type)thread;
    operands = application_thread->opd;
    operator = Context_Get_Expression(Context_id);
    tag = Grammar_Tag(operator);
    switch (tag)
      { case CNT_tag:
          apply_continuation_C(Context_id,
                               operands,
                               tail_position_status);
          return;
        case NAT_tag:
          apply_native_C(Context_id,
                         operands,
                         tail_position_status);
          return;
        case PRC_tag:
          apply_procedure_C(Context_id,
                            operands,
                            tail_position_status);
          return;
        case PRZ_tag:
          apply_vararg_procedure_C(Context_id,
                                   operands,
                                   tail_position_status);
          return;
        default:
          Main_Error_Text(OPR_error,
                          Application_String); }}

/*----------------------------------------------------------------------------*/

static NIL_type evaluate_application_C(CID_type Context_id,
                                       EXP_type Tail_position_status)
  { aPL_type application_thread;
    APL_type application_expression;
    EXP_type expression;
    THR_type thread;
    VEC_type operands;
    thread = Context_Thread_Push_C(Context_id,
                                   Continue_application,
                                   Tail_position_status,
                                   aPL_size);
    application_expression = Context_Get_Expression(Context_id);
    expression = application_expression->exp;
    operands   = application_expression->opd;
    application_thread = (aPL_type)thread;
    application_thread->opd = operands;
    evaluate_expression_C(expression,
                          Context_id,
                          Grammar_False); }

/*----------------------------------------------------------------------------*/

static NIL_type initialize_application(NIL_type)
  { Continue_application = Context_Register(continue_application_C,
                                            aPL_size); }

/*----------------------------------------------------------------------------*/
/*----------------------------------- begin ----------------------------------*/
/*----------------------------------------------------------------------------*/

static NIL_type evaluate_begin_C(CID_type Context_id,
                                 EXP_type Tail_position_status)
  { BEG_type begin_expression;
    EXP_type body;
    begin_expression = Context_Get_Expression(Context_id);
    body = begin_expression->bod;
    evaluate_expression_C(body,
                          Context_id,
                          Tail_position_status); }

/*----------------------------------------------------------------------------*/
/*------------------------------------ body ----------------------------------*/
/*----------------------------------------------------------------------------*/

static NBR_type Continue_body;

Thread_Begin_Frame(bOD);
  Thread_Frame_Slot(ENV, env);
  Thread_Frame_Slot(FRM, frm);
Thread_End_Frame(bOD);

static NIL_type continue_body(CID_type Context_id)
  { bOD_type body_thread;
    ENV_type environment;
    FRM_type frame;
    THR_type thread;
    thread = Context_Thread_Peek(Context_id);
    body_thread = (bOD_type)thread;
    environment = body_thread->env;
    frame       = body_thread->frm;
    Context_Thread_Zap(Context_id);
    Context_Release_Current_Frame(Context_id);
    Context_Environment_Replace(Context_id,
                                environment,
                                frame); }

/*----------------------------------------------------------------------------*/

static NIL_type evaluate_direct_procedure_body_C(CID_type Context_id,
                                                 FRM_type Frame,
                                                 EXP_type Tail_position_status)
  { bOD_type body_thread;
    ENV_type environment;
    EXP_type body;
    FRM_type frame;
    PRC_type procedure;
    THR_type thread;
    if (is_FLS(Tail_position_status))
      { Context_Save(Context_id,
                     Frame);    
        thread = Context_Thread_Push_C(Context_id,
                                       Continue_body,
                                       Grammar_False,
                                       bOD_size);
        Frame = Context_Restore(Context_id);    
        body_thread = (bOD_type)thread;
        environment = Context_Get_Environment(Context_id);
        frame = Context_Get_Frame(Context_id);
        body_thread->env = environment;
        body_thread->frm = frame; }
    else
      Context_Release_Current_Frame(Context_id);
    procedure = Context_Get_Expression(Context_id);
    environment = procedure->env;
    body        = procedure->bod;
    Context_Environment_Replace(Context_id,
                                environment,
                                Frame);
    evaluate_expression_C(body,
                          Context_id,
                          Grammar_True); }      

/*----------------------------------------------------------------------------*/

static NIL_type evaluate_procedure_body_C(CID_type Context_id,
                                          FRM_type Frame,
                                          EXP_type Tail_position_status)
  { bOD_type body_thread;
    ENV_type environment;
    EXP_type body;
    FRM_type frame;
    PRC_type procedure;
    THR_type thread;
    if (is_FLS(Tail_position_status))
      { Context_Save(Context_id,
                     Frame);    
        thread = Context_Thread_Poke_C(Context_id,
                                       Continue_body,
                                       Grammar_False,
                                       bOD_size);
        Frame = Context_Restore(Context_id);    
        body_thread = (bOD_type)thread;
        environment = Context_Get_Environment(Context_id);
        frame = Context_Get_Frame(Context_id);
        body_thread->env = environment;
        body_thread->frm = frame; }
    else
      { Context_Release_Current_Frame(Context_id);
        Context_Thread_Zap(Context_id); }
    procedure = Context_Get_Expression(Context_id);
    environment = procedure->env;
    body        = procedure->bod;
    Context_Environment_Replace(Context_id,
                                environment,
                                Frame);
    evaluate_expression_C(body,
                          Context_id,
                          Grammar_True); }      

/*----------------------------------------------------------------------------*/

static NIL_type initialize_body(NIL_type)
  { Continue_body = Context_Register(continue_body,
                                     bOD_size); }

/*----------------------------------------------------------------------------*/
/*----------------------------------- define ---------------------------------*/
/*----------------------------------------------------------------------------*/

static NBR_type Continue_define;

Thread_Begin_Frame(dEF);
  Thread_Frame_Slot(NBR, ofs);
Thread_End_Frame(dEF);

static NIL_type continue_define(CID_type Context_id)
  { dEF_type define_thread;
    NBR_type offset;
    THR_type thread;
    thread = Context_Thread_Peek(Context_id);
    define_thread = (dEF_type)thread;
    offset = define_thread->ofs;
    Context_Thread_Zap(Context_id);
    Context_Local_Set(Context_id,
                      offset); }

/*----------------------------------------------------------------------------*/

static NIL_type evaluate_define_function_C(CID_type Context_id,
                                           EXP_type Tail_position_status)
  { DFF_type define_expression;
    ENV_type environment;
    EXP_type body;
    NBR_type frame_size,
             offset,
             param_count;
    PRC_type procedure;
    SYM_type name;
    Main_Claim_Default();
    define_expression = Context_Get_Expression(Context_id);
    name        = define_expression->nam;
    offset      = define_expression->ofs;
    param_count = define_expression->par;
    frame_size  = define_expression->siz;
    body        = define_expression->bod;
    procedure = make_PRC_M(name,
                           param_count,
                           frame_size,
                           body,
                           Grammar_Null);
    Main_Release_Default();
    Context_Set_Expression(Context_id,
                           procedure);
    environment = Context_Environment_Grow_C(Context_id);
    procedure = Context_Get_Expression(Context_id);
    procedure->env = environment;
    Context_Local_Set(Context_id,
                      offset); }

/*----------------------------------------------------------------------------*/

static NIL_type evaluate_define_function_vararg_C(CID_type Context_id,
                                                  EXP_type Tail_position_status)
  { DFF_type define_expression;
    ENV_type environment;
    EXP_type body;
    NBR_type frame_size,
             offset,
             param_count;
    PRC_type procedure;
    SYM_type name;
    Main_Claim_Default();
    define_expression = Context_Get_Expression(Context_id);
    name        = define_expression->nam;
    offset      = define_expression->ofs;
    param_count = define_expression->par;
    frame_size  = define_expression->siz;
    body        = define_expression->bod;
    procedure = make_PRZ_M(name,
                           param_count,
                           frame_size,
                           body,
                           Grammar_Null);
    Main_Release_Default();
    Context_Set_Expression(Context_id,
                           procedure);
    environment = Context_Environment_Grow_C(Context_id);
    procedure = Context_Get_Expression(Context_id);
    procedure->env = environment;
    Context_Local_Set(Context_id,
                      offset); }

/*----------------------------------------------------------------------------*/

static NIL_type evaluate_define_variable_C(CID_type Context_id,
                                           EXP_type Tail_position_status)
  { dEF_type define_thread;
    DFV_type define_expression;
    EXP_type expression;
    NBR_type offset;
    THR_type thread;
    thread = Context_Thread_Push_C(Context_id,
                                   Continue_define,
                                   Grammar_False,
                                   dEF_size);
    define_expression = Context_Get_Expression(Context_id);
    offset     = define_expression->ofs;
    expression = define_expression->exp;
    define_thread = (dEF_type)thread;
    define_thread->ofs = offset;
    evaluate_expression_C(expression,
                          Context_id,
                          Grammar_False); }

/*----------------------------------------------------------------------------*/

static NIL_type initialize_define(NIL_type)
  { Continue_define = Context_Register(continue_define,
                                       dEF_size); }

/*----------------------------------------------------------------------------*/
/*----------------------------- direct application ---------------------------*/
/*----------------------------------------------------------------------------*/

static const TXT_type Direct_Application_String = "direct application";

/*----------------------------------------------------------------------------*/

static NIL_type evaluate_direct_application_C(CID_type Context_id,
                                              EXP_type Tail_position_status)
  { EXP_type operator;
    TAG_type tag;
    VEC_type vector;
    vector = Context_Get_Expression(Context_id);
    operator = vector[1];
    tag = Grammar_Tag(operator);
    switch (tag)
      { case CNT_tag:
          apply_direct_continuation(Context_id,
                                    Tail_position_status);
          return;
        case NAT_tag:
          apply_direct_native_C(Context_id,
                                Tail_position_status);
          return;
        case PRC_tag:
          apply_direct_procedure_C(Context_id,
                                   Tail_position_status);
          return;
        case PRZ_tag:
          apply_direct_vararg_procedure_C(Context_id,
                                          Tail_position_status);
          return;
        default:
          Main_Error_Text(OPR_error,
                          Direct_Application_String); }}

/*----------------------------------------------------------------------------*/
/*------------------------------ global variable -----------------------------*/
/*----------------------------------------------------------------------------*/

static NIL_type evaluate_global_variable(CID_type Context_id,
                                         EXP_type Tail_position_status)
  { GLB_type global_expression;
    NBR_type offset,
             scope;
    global_expression = Context_Get_Expression(Context_id);
    scope  = global_expression->scp;
    offset = global_expression->ofs;
    Context_Global_Get(Context_id,
                       scope,
                       offset); }

/*----------------------------------------------------------------------------*/
/*------------------------------------ if ------------------------------------*/
/*----------------------------------------------------------------------------*/

static NBR_type Continue_if_double;
static NBR_type Continue_if_single;

Thread_Begin_Frame(iFD);
  Thread_Frame_Slot(EXP, cns);
  Thread_Frame_Slot(EXP, alt);
Thread_End_Frame(iFD);

Thread_Begin_Frame(iFS);
  Thread_Frame_Slot(EXP, cns);
Thread_End_Frame(iFS);

/*----------------------------------------------------------------------------*/

static NIL_type continue_if_double_C(CID_type Context_id)
  { iFD_type if_thread;
    EXP_type boolean,
             expression,
             tail_position_status;
    TAG_type tag;
    THR_type thread;
    boolean = Context_Get_Expression(Context_id);
    thread = Context_Thread_Peek(Context_id);
    if_thread = (iFD_type)thread;
    tail_position_status = Thread_Get_Tail(thread);
    tag = Grammar_Tag(boolean);
    switch (tag)
      { case FLS_tag:
          expression = if_thread->alt;
          Context_Thread_Zap(Context_id);
          evaluate_expression_C(expression,
                                Context_id,
                                tail_position_status);
          return;
        case TRU_tag:
          expression = if_thread->cns;
          Context_Thread_Zap(Context_id);
          evaluate_expression_C(expression,
                                Context_id,
                                tail_position_status);
          return;
        default:
          Main_Error_Text(BRQ_error,
                          Main_If_String);
          return; }}

static NIL_type continue_if_single_C(CID_type Context_id)
  { iFS_type if_thread;
    EXP_type boolean,
             expression,
             tail_position_status;
    TAG_type tag;
    THR_type thread;
    boolean = Context_Get_Expression(Context_id);
    thread = Context_Thread_Peek(Context_id);
    if_thread = (iFS_type)thread;
    tail_position_status = Thread_Get_Tail(thread);
    tag = Grammar_Tag(boolean);
    switch (tag)
      { case FLS_tag:
          Context_Thread_Zap(Context_id);
          Context_Set_Expression(Context_id,
                                 Grammar_Unspecified);
          return;
        case TRU_tag:
          expression = if_thread->cns;
          Context_Thread_Zap(Context_id);
          evaluate_expression_C(expression,
                                Context_id,
                                tail_position_status);
          return;
        default:
          Main_Error_Text(BRQ_error,
                          Main_If_String); }}
                          
/*----------------------------------------------------------------------------*/

static NIL_type evaluate_if_double_C(CID_type Context_id,
                                     EXP_type Tail_position_status)
  { iFD_type if_thread;
    IFF_type if_expression;
    EXP_type alternative,
             consequent,
             predicate;
    THR_type thread;
    thread = Context_Thread_Push_C(Context_id,
                                   Continue_if_double,
                                   Tail_position_status,
                                   iFD_size);
    if_expression = Context_Get_Expression(Context_id);
    predicate   = if_expression->prd;
    consequent  = if_expression->cns;
    alternative = if_expression->alt;
    if_thread = (iFD_type)thread;
    if_thread->cns = consequent;
    if_thread->alt = alternative;
    evaluate_expression_C(predicate,
                          Context_id,
                          Grammar_False); }

/*----------------------------------------------------------------------------*/

static NIL_type evaluate_if_single_C(CID_type Context_id,
                                     EXP_type Tail_position_status)
  { iFS_type if_thread;
    IFZ_type if_expression;
    EXP_type consequent,
             predicate;
    THR_type thread;
    thread = Context_Thread_Push_C(Context_id,
                                   Continue_if_single,
                                   Tail_position_status,
                                   iFS_size);
    if_expression = Context_Get_Expression(Context_id);
    predicate  = if_expression->prd;
    consequent = if_expression->cns;
    if_thread = (iFS_type)thread;
    if_thread->cns = consequent;
    evaluate_expression_C(predicate,
                          Context_id,
                          Grammar_False); }

static NIL_type initialize_if(NIL_type)
  { Continue_if_double = Context_Register(continue_if_double_C,
                                          iFS_size);
    Continue_if_single = Context_Register(continue_if_single_C,
                                          iFD_size); }

/*----------------------------------------------------------------------------*/
/*----------------------------------- lambda ---------------------------------*/
/*----------------------------------------------------------------------------*/

static NIL_type evaluate_lambda_C(CID_type Context_id,
                                  EXP_type Tail_position_status)
  { ENV_type environment;
    EXP_type body;
    LMB_type lambda_expression;
    NBR_type frame_size,
             param_count;
    PRC_type procedure;
    Main_Claim_Default();
    lambda_expression = Context_Get_Expression(Context_id);
    param_count = lambda_expression->par;
    frame_size  = lambda_expression->siz;
    body        = lambda_expression->bod;
    procedure = make_PRC_M(Main_Anonymous,
                           param_count,
                           frame_size,
                           body,
                           Grammar_Null);
    Main_Release_Default();
    Context_Set_Expression(Context_id,
                           procedure);
    environment = Context_Environment_Grow_C(Context_id);
    procedure = Context_Get_Expression(Context_id);
    procedure->env = environment; }

static NIL_type evaluate_lambda_vararg_C(CID_type Context_id,
                                         EXP_type Tail_position_status)
  { ENV_type environment;
    EXP_type body;
    LMB_type lambda_expression;
    NBR_type frame_size,
             param_count;
    PRC_type procedure;
    Main_Claim_Default();
    lambda_expression = Context_Get_Expression(Context_id);
    param_count = lambda_expression->par;
    frame_size  = lambda_expression->siz;
    body        = lambda_expression->bod;
    procedure = make_PRZ_M(Main_Anonymous,
                           param_count,
                           frame_size,
                           body,
                           Grammar_Null);
    Main_Release_Default();
    Context_Set_Expression(Context_id,
                           procedure);
    environment = Context_Environment_Grow_C(Context_id);
    procedure = Context_Get_Expression(Context_id);
    procedure->env = environment; }

/*----------------------------------------------------------------------------*/
/*---------------------------------- literal ---------------------------------*/
/*----------------------------------------------------------------------------*/

static NIL_type evaluate_literal(CID_type Context_id,
                                 EXP_type Tail_position_status)
  { return; }

/*----------------------------------------------------------------------------*/
/*------------------------------- local variable -----------------------------*/
/*----------------------------------------------------------------------------*/

static NIL_type evaluate_local_variable(CID_type Context_id,
                                        EXP_type Tail_position_status)
  { LCL_type local;
    NBR_type offset;
    local = Context_Get_Expression(Context_id);
    offset = local->ofs;
    Context_Local_Get(Context_id,
                      offset); }

/*----------------------------------------------------------------------------*/
/*----------------------------------- print ----------------------------------*/
/*----------------------------------------------------------------------------*/

static NBR_type Continue_print;

Thread_Begin_Frame(pRI);
Thread_End_Frame(pRI);

/*----------------------------------------------------------------------------*/

static NIL_type continue_print(CID_type Context_id)
  { EXP_type value;
    value = Context_Get_Expression(Context_id);
    Print_Print(value);
    Main_Exit(); }

/*----------------------------------------------------------------------------*/

static NIL_type initialize_print(NIL_type)
  { Continue_print = Context_Register(continue_print,
                                      pRI_size); }

/*----------------------------------------------------------------------------*/
/*---------------------------------- protect ---------------------------------*/
/*----------------------------------------------------------------------------*/

static NBR_type Continue_protect;

Thread_Begin_Frame(pRT);
  Thread_Frame_Slot(LCK, lck);
Thread_End_Frame(pRT);

/*----------------------------------------------------------------------------*/

static NIL_type continue_protect(CID_type Context_id)
  { pRT_type protect_thread;
    LCK_type lock;
    THR_type thread;
    thread = Context_Thread_Peek(Context_id);
    protect_thread = (pRT_type)thread;
    lock = protect_thread->lck;
    Context_Thread_Zap(Context_id);
    Main_Unlock_Thread(lock); }             /* should be protected against GC */

/*----------------------------------------------------------------------------*/

static NIL_type evaluate_protect_C(CID_type Context_id,
                                   EXP_type Tail_position_status)
  { pRT_type protect_thread;
    EXP_type body;
    LCK_type lock;
    PRT_type protect_expression;
    THR_type thread;
    thread = Context_Thread_Push_C(Context_id,
                                   Continue_protect,
                                   Tail_position_status,
                                   pRT_size);
    protect_expression = Context_Get_Expression(Context_id);
    body = protect_expression->bod;
    lock = protect_expression->lck;
    protect_thread = (pRT_type)thread;
    protect_thread->lck = lock;
    Main_Lock_Thread(lock);                 /* should be protected against GC */
    evaluate_expression_C(body,
                          Context_id,
                          Tail_position_status); }

/*----------------------------------------------------------------------------*/

static NIL_type initialize_protect(NIL_type)
  { Continue_protect = Context_Register(continue_protect,
                                        pRT_size); }

/*----------------------------------------------------------------------------*/
/*----------------------------------- quote ----------------------------------*/
/*----------------------------------------------------------------------------*/

static NIL_type evaluate_quote(CID_type Context_id,
                               EXP_type Tail_position_status)
  { EXP_type expression;
    QUO_type quote_expression;
    quote_expression = Context_Get_Expression(Context_id);
    expression = quote_expression->exp;
    Context_Set_Expression(Context_id,
                           expression); }        /* clone!!! */

/*----------------------------------------------------------------------------*/
/*---------------------------------- sequence --------------------------------*/
/*----------------------------------------------------------------------------*/

static NBR_type Continue_sequence;

Thread_Begin_Frame(sEQ);
  Thread_Frame_Slot(SEQ, seq);
  Thread_Frame_Slot(NBR, pos);
Thread_End_Frame(sEQ);

/*----------------------------------------------------------------------------*/

static NIL_type continue_sequence_C(CID_type Context_id)
  { sEQ_type sequence_thread;
    EXP_type expression,
             tail_position_status;
    NBR_type position;
    SEQ_type sequence;
    THR_type thread;
    UNS_type raw_position,
             size;
    thread = Context_Thread_Peek(Context_id);
    sequence_thread = (sEQ_type)thread;
    sequence = sequence_thread->seq;
    position = sequence_thread->pos;
    raw_position = get_NBR(position);
    size = size_SEQ(sequence);
    raw_position += 1;
    if (raw_position < size)
      { thread = Context_Thread_Keep_C(Context_id);
        sequence_thread = (sEQ_type)thread;
        sequence = sequence_thread->seq;
        position = make_NBR(raw_position);
        sequence_thread->pos = position;
        tail_position_status = Grammar_False; }
    else
      { Context_Thread_Zap(Context_id);
        tail_position_status = Thread_Get_Tail(thread); }
    expression = sequence[raw_position];
    evaluate_expression_C(expression,
                          Context_id,
                          tail_position_status); }

/*----------------------------------------------------------------------------*/

static NIL_type evaluate_sequence_C(CID_type Context_id,
                                    EXP_type Tail_position_status)
  { sEQ_type sequence_thread;
    EXP_type expression,
             tail_position_status;
    SEQ_type sequence_expression;
    THR_type thread;
    UNS_type size;
    sequence_expression = Context_Get_Expression(Context_id);
    size = size_SEQ(sequence_expression);
    if (size > 1)
      { thread = Context_Thread_Push_C(Context_id,
                                       Continue_sequence,
                                       Tail_position_status,
                                       sEQ_size);
        sequence_expression = Context_Get_Expression(Context_id);
        sequence_thread = (sEQ_type)thread;
        sequence_thread->seq = sequence_expression;
        sequence_thread->pos = Grammar_One;
        tail_position_status = Grammar_False; }
    else
      tail_position_status = Tail_position_status;
    expression = sequence_expression[1];
    evaluate_expression_C(expression,
                          Context_id,
                          tail_position_status); }

/*----------------------------------------------------------------------------*/

static NIL_type initialize_sequence(NIL_type)
  { Continue_sequence = Context_Register(continue_sequence_C,
                                         sEQ_size); }

/*----------------------------------------------------------------------------*/
/*--------------------------------- set global -------------------------------*/
/*----------------------------------------------------------------------------*/

static NBR_type Continue_set_global;

Thread_Begin_Frame(sTG);
  Thread_Frame_Slot(NBR, scp);
  Thread_Frame_Slot(NBR, ofs);
Thread_End_Frame(sTG);

/*----------------------------------------------------------------------------*/

static NIL_type continue_set_global(CID_type Context_id)
  { sTG_type set_thread;
    NBR_type offset,
             scope;
    THR_type thread;
    thread = Context_Thread_Peek(Context_id);
    set_thread = (sTG_type)thread;
    scope  = set_thread->scp;
    offset = set_thread->ofs;
    Context_Thread_Zap(Context_id);
    Context_Global_Set(Context_id,
                       scope,
                       offset); }

/*----------------------------------------------------------------------------*/

static NIL_type evaluate_set_global_C(CID_type Context_id,
                                      EXP_type Tail_position_status)
  { sTG_type set_thread;
    EXP_type expression;
    NBR_type offset,
             scope;
    STG_type set_expression;
    THR_type thread;
    thread = Context_Thread_Push_C(Context_id,
                                   Continue_set_global,
                                   Grammar_False,
                                   sTG_size);
    set_expression = Context_Get_Expression(Context_id);
    scope      = set_expression->scp;
    offset     = set_expression->ofs;
    expression = set_expression->exp;
    set_thread = (sTG_type)thread;
    set_thread->scp = scope;
    set_thread->ofs = offset;
    evaluate_expression_C(expression,
                          Context_id,
                          Grammar_False); }

/*----------------------------------------------------------------------------*/

static NIL_type initialize_set_global(NIL_type)
  { Continue_set_global = Context_Register(continue_set_global,
                                           sTG_size); }

/*----------------------------------------------------------------------------*/
/*--------------------------------- set local --------------------------------*/
/*----------------------------------------------------------------------------*/

static NBR_type Continue_set_local;

Thread_Begin_Frame(sTL);
  Thread_Frame_Slot(NBR, ofs);
Thread_End_Frame(sTL);

/*----------------------------------------------------------------------------*/

static NIL_type continue_set_local(CID_type Context_id)
  { sTL_type set_thread;
    NBR_type offset;
    THR_type thread;
    thread = Context_Thread_Peek(Context_id);
    set_thread = (sTL_type)thread;
    offset = set_thread->ofs;
    Context_Thread_Zap(Context_id);
    Context_Local_Set(Context_id,
                      offset); }

/*----------------------------------------------------------------------------*/

static NIL_type evaluate_set_local_C(CID_type Context_id,
                                     EXP_type Tail_position_status)
  { sTL_type set_thread;
    EXP_type expression;
    NBR_type offset;
    STL_type set_expression;
    THR_type thread;
    thread = Context_Thread_Push_C(Context_id,
                                   Continue_set_local,
                                   Grammar_False,
                                   sTL_size);
    set_expression = Context_Get_Expression(Context_id);
    offset     = set_expression->ofs;
    expression = set_expression->exp;
    set_thread = (sTL_type)thread;
    set_thread->ofs = offset;
    evaluate_expression_C(expression,
                          Context_id,
                          Grammar_False); }

/*----------------------------------------------------------------------------*/

static NIL_type initialize_set_local(NIL_type)
  { Continue_set_local = Context_Register(continue_set_local,
                                          sTL_size); }

/*----------------------------------------------------------------------------*/
/*----------------------------------- spawn ----------------------------------*/
/*----------------------------------------------------------------------------*/

static NBR_type Continue_spawn;

Thread_Begin_Frame(sPN);
Thread_End_Frame(sPN);

/*----------------------------------------------------------------------------*/

static NIL_type continue_spawn(CID_type Context_id)
  { Context_Stop_Task(); }

/*----------------------------------------------------------------------------*/

static EXP_type start_procedure_C(CID_type Context_id)
  { EXP_type expression;
    SPN_type spawn_expression;
    Context_Start_Task();
    Context_Thread_Push_C(Context_id,
                          Continue_spawn,
                          Grammar_False,
                          sPN_size);
    spawn_expression = Context_Get_Expression(Context_id);
    expression = spawn_expression->exp;
    evaluate_expression_C(expression,
                          Context_id,
                          Grammar_False);
    for (;;)
      Context_Proceed(Context_id);
    return Grammar_Unspecified; }

static NIL_type evaluate_spawn_C(CID_type Context_id,
                                 EXP_type Tail_position_status)
  { CID_type new_context_id;
    PRM_type promise;
    new_context_id = Context_Spawn_Task_C(Context_id,
                                          start_procedure_C);
    Main_Claim_Default();
    promise = make_PRM_M(new_context_id);
    Main_Release_Default();
    Context_Set_Expression(Context_id,
                           promise); }

/*----------------------------------------------------------------------------*/

static NIL_type initialize_spawn(NIL_type)
  { Continue_spawn = Context_Register(continue_spawn,
                                      sPN_size); }

/*----------------------------------------------------------------------------*/
/*----------------------------------- thunk ----------------------------------*/
/*----------------------------------------------------------------------------*/

static NIL_type evaluate_thunk_C(CID_type Context_id,
                                 EXP_type Tail_position_status)
  { bOD_type body_thread;
    ENV_type environment;
    EXP_type body;
    FRM_type frame;
    NBR_type frame_size;
    THK_type thunk_expression;
    THR_type thread;
    UNS_type raw_frame_size;
    thunk_expression = Context_Get_Expression(Context_id);
    frame_size = thunk_expression->siz;
    raw_frame_size = get_NBR(frame_size);
    if (is_FLS(Tail_position_status))
      { thread = Context_Thread_Push_C(Context_id,
                                       Continue_body,
                                       Grammar_False,
                                       bOD_size);
        body_thread = (bOD_type)thread;
        environment = Context_Get_Environment(Context_id);
        frame = Context_Get_Frame(Context_id);
        body_thread->env = environment;
        body_thread->frm = frame; }
    environment = Context_Environment_Grow_C(Context_id);
    Context_Save(Context_id,
                 environment);    
    frame = Context_Make_Frame_C(Context_id,
                                 raw_frame_size);
    environment = Context_Restore(Context_id);    
    Context_Environment_Replace(Context_id,
                                environment,
                                frame);
    thunk_expression = Context_Get_Expression(Context_id);
    body = thunk_expression->bod;
    evaluate_expression_C(body,
                          Context_id,
                          Grammar_True); }
                        
/*----------------------------------------------------------------------------*/
/*----------------------------------- value ----------------------------------*/
/*----------------------------------------------------------------------------*/

static NIL_type evaluate_value(CID_type Context_id,
                              EXP_type Tail_position_status)
  { EXP_type expression;
    VAL_type value_expression;
    value_expression = Context_Get_Expression(Context_id);
    expression = value_expression->exp;
    Context_Set_Expression(Context_id,
                           expression); }

/*----------------------------------------------------------------------------*/
/*----------------------------------- while ----------------------------------*/
/*----------------------------------------------------------------------------*/

static NBR_type Continue_while_body,
                Continue_while_predicate;

Thread_Begin_Frame(wHI);
  Thread_Frame_Slot(EXP, prd);
  Thread_Frame_Slot(EXP, bod);
  Thread_Frame_Slot(EXP, res);
Thread_End_Frame(wHI);

/*----------------------------------------------------------------------------*/

static NIL_type continue_while_body_C(CID_type Context_id)
  { wHI_type while_thread;
    EXP_type predicate,
             value;
    THR_type thread;
    thread = Context_Thread_Patch_C(Context_id,
                                    Continue_while_predicate);
    while_thread = (wHI_type)thread;
    value = Context_Get_Expression(Context_id);
    while_thread->res = value;
    predicate = while_thread->prd;
    evaluate_expression_C(predicate,
                          Context_id,
                          Grammar_False); }

static NIL_type continue_while_predicate_C(CID_type Context_id)
  { wHI_type while_thread;
    EXP_type body,
             boolean,
             result;
    THR_type thread;
    boolean = Context_Get_Expression(Context_id);
    if (is_FLS(boolean))
      { thread = Context_Thread_Peek(Context_id);
        while_thread = (wHI_type)thread;
        result = while_thread->res;
        Context_Thread_Zap(Context_id);
        Context_Set_Expression(Context_id,
                               result);
        return; }
    thread = Context_Thread_Patch_C(Context_id,
                                    Continue_while_body);
    while_thread = (wHI_type)thread;
    body = while_thread->bod;
    evaluate_expression_C(body,
                          Context_id,
                          Grammar_False); }

/*----------------------------------------------------------------------------*/

static NIL_type evaluate_while_C(CID_type Context_id,
                                 EXP_type Tail_position_status)
  { wHI_type while_thread;
    EXP_type body,
             predicate;
    THR_type thread;
    WHI_type while_expression;
    thread = Context_Thread_Push_C(Context_id,
                                   Continue_while_predicate,
                                   Grammar_False,
                                   wHI_size);
    while_expression = Context_Get_Expression(Context_id);
    predicate = while_expression->prd;
    body      = while_expression->bod;
    while_thread = (wHI_type)thread;
    while_thread->prd = predicate;
    while_thread->bod = body;
    while_thread->res = Grammar_Unspecified;
    evaluate_expression_C(predicate,
                          Context_id,
                          Grammar_False); }

/*----------------------------------------------------------------------------*/

static NIL_type initialize_while(NIL_type)
  { Continue_while_predicate = Context_Register(continue_while_predicate_C,
                                                wHI_size);
    Continue_while_body = Context_Register(continue_while_body_C,
                                           wHI_size); }

/*----------------------------------------------------------------------------*/
/*--------------------------------- expression -------------------------------*/
/*----------------------------------------------------------------------------*/

static NIL_type evaluate_context_C(CID_type Context_id,
                                 EXP_type Tail_position_status)
  { EXP_type expression;
    TAG_type tag;
    expression = Context_Get_Expression(Context_id);
    tag = Grammar_Tag(expression);
    switch (tag)
      { case APL_tag:
          evaluate_application_C(Context_id,
                                 Tail_position_status);
          return;
        case BEG_tag:
          evaluate_begin_C(Context_id,
                           Tail_position_status);
          return;
        case DFF_tag:
          evaluate_define_function_C(Context_id,
                                     Tail_position_status);
          return;
        case DFV_tag:
          evaluate_define_variable_C(Context_id,
                                     Tail_position_status);
          return;
        case DFZ_tag:
          evaluate_define_function_vararg_C(Context_id,
                                            Tail_position_status);
          return;
        case GLB_tag:
          evaluate_global_variable(Context_id,
                                   Tail_position_status);
          return;
        case IFF_tag:
          evaluate_if_double_C(Context_id,
                               Tail_position_status);
          return;
        case IFZ_tag:
          evaluate_if_single_C(Context_id,
                               Tail_position_status);
          return;
        case LCL_tag:
          evaluate_local_variable(Context_id,
                                  Tail_position_status);
          return;
        case LMB_tag:
          evaluate_lambda_C(Context_id,
                            Tail_position_status);
          return;
        case LMZ_tag:
          evaluate_lambda_vararg_C(Context_id,
                                   Tail_position_status);
          return;
        case PRT_tag:
          evaluate_protect_C(Context_id,
                             Tail_position_status);
          return;
        case QUO_tag:
          evaluate_quote(Context_id,
                         Tail_position_status);
          return;
        case SEQ_tag:
          evaluate_sequence_C(Context_id,
                              Tail_position_status);
          return;
        case SPN_tag:
          evaluate_spawn_C(Context_id,
                           Tail_position_status);
          return;
        case STG_tag:
          evaluate_set_global_C(Context_id,
                                Tail_position_status);
          return;
        case STL_tag:
          evaluate_set_local_C(Context_id,
                               Tail_position_status);
          return;
        case THK_tag:
          evaluate_thunk_C(Context_id,
                           Tail_position_status);
          return;
        case VAL_tag:
          evaluate_value(Context_id,
                         Tail_position_status);
          return;
        case WHI_tag:
          evaluate_while_C(Context_id,
                           Tail_position_status);
          return;
        case CHA_tag:
        case CNT_tag:
        case FLS_tag:
        case NAT_tag:
        case NBR_tag:
        case NUL_tag:
        case PAI_tag:
        case PRC_tag:
        case PRM_tag:
        case PRZ_tag:
        case REA_tag:
        case STR_tag:
        case TRU_tag:
        case USP_tag:
        case VEC_tag:
          evaluate_literal(Context_id,
                           Tail_position_status);
          return;
        default:
          Main_Error_Tag(IXT_error,
                         tag); }}

static NIL_type evaluate_expression_C(EXP_type Expression,
                                    CID_type Context_id,
                                    EXP_type Tail_position_status)
  { Context_Set_Expression(Context_id,
                           Expression);
    evaluate_context_C(Context_id,
                       Tail_position_status); }

/*------------------------------ public functions ----------------------------*/

NIL_type Evaluate_Apply_C(CID_type Context_id,
                          EXP_type Tail_position_status)
  { evaluate_direct_application_C(Context_id,
                                  Tail_position_status); }

NIL_type Evaluate_Eval_C(CID_type Context_id,
                         EXP_type Tail_position_status)
  { Main_Adjust_Global_Frame();
    evaluate_context_C(Context_id,
                       Tail_position_status); }

NIL_type Evaluate_Evaluate(EXP_type Expression)
  { Context_Set_Expression(Context_Main_Id,
                           Expression);
    Context_Thread_Push_C(Context_Main_Id,
                          Continue_print,
                          Grammar_False,
                          0);
    Main_Adjust_Global_Frame();
    evaluate_context_C(Context_Main_Id,
                     Grammar_False);
    for (;;)
      Context_Proceed(Context_Main_Id); }

/*----------------------------------------------------------------------------*/
                               
NIL_type Evaluate_Initialize(NIL_type)
  { initialize_application();
    initialize_body();
    initialize_continuation();
    initialize_define();
    initialize_if();
    initialize_native();
    initialize_print();
    initialize_procedure();
    initialize_protect();
    initialize_sequence();
    initialize_set_global();
    initialize_set_local();
    initialize_spawn();
    initialize_vararg_procedure();
    initialize_while(); }
