                     /*-----------------------------------*/
                     /*             >>>Slip<<<            */
                     /*            Theo D'Hondt           */
                     /*     VUB Software Languages Lab    */
                     /*             (c) 2010              */
                     /*-----------------------------------*/
                     /*      version 14: multicores       */
                     /*-----------------------------------*/
                     /*               grammar             */
                     /*-----------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "SlipGrammar.h"

/*------------------------------- private macros -----------------------------*/

#define make_chunk_with_offset(SIGNATURE, SIZE)                                \
  (SIGNATURE##_type)Context_Make_Chunk(SIGNATURE##_tag, SIGNATURE##_size + SIZE)

#define make_chunk(SIGNATURE)                                                  \
  make_chunk_with_offset(SIGNATURE, 0)

/*------------------------------ private constants ---------------------------*/

enum { Immediate_null_value        = Memory_Void_Value + 0*Memory_Cell_Bias,
       Immediate_true_value        = Memory_Void_Value + 1*Memory_Cell_Bias,
       Immediate_false_value       = Memory_Void_Value + 2*Memory_Cell_Bias,
       Immediate_unspecified_value = Memory_Void_Value + 3*Memory_Cell_Bias,
       Immediate_boundary          = Memory_Void_Value + 3*Memory_Cell_Bias };

/*------------------------------ public variables ----------------------------*/

VEC_type Grammar_Empty_Vector;
FLS_type Grammar_False;
NUL_type Grammar_Null;
NBR_type Grammar_One;
TRU_type Grammar_True;
USP_type Grammar_Unspecified;
NBR_type Grammar_Zero;

/*------------------------------ private functions ---------------------------*/

static BYT_type member_of(EXP_type Expression,
                          TAG_type Tag)
  { BYT_type tag;
    tag = Grammar_Tag(Expression);
    return (tag == Tag); }

static UNS_type string_to_expression_size(TXT_type Raw_string)
  { return strlen(Raw_string) / sizeof(CEL_type) + 1; }

/*------------------------------ public functions ----------------------------*/

TAG_type Grammar_Tag(EXP_type Expression)
  { UNS_type immediate;
    if (Memory_Is_Immediate(Expression))
      return NBR_tag;
    immediate = (UNS_type)Expression;
    if (immediate <= Immediate_boundary)
      switch (immediate)
        { case Immediate_null_value:
            return NUL_tag;
          case Immediate_true_value:
            return TRU_tag;
          case Immediate_false_value:
            return FLS_tag;
          case Immediate_unspecified_value:
            return USP_tag; }
    return (TAG_type)Memory_Get_Tag((PTR_type)Expression); }

/*------------------------------ APL declarations ----------------------------*/

BYT_type is_APL(EXP_type Expression)
  { BYT_type tag = Grammar_Tag(Expression);
    return (tag == APL_tag); }

APL_type make_APL_M(EXP_type Expression,
                    VEC_type Operands)
  { APL_type application = make_chunk(APL);
    application->exp = Expression;
    application->opd = Operands;
    return application; }

/*------------------------------ BEG declarations ----------------------------*/

BYT_type is_BEG(EXP_type Expression)
  { BYT_type tag = Grammar_Tag(Expression);
    return (tag == BEG_tag); }

BEG_type make_BEG_M(EXP_type Body)
  { BEG_type begin = make_chunk(BEG);
    begin->bod = Body;
    return begin; }

/*------------------------------ CHA declarations ----------------------------*/

BYT_type is_CHA(EXP_type Expression)
  { return member_of(Expression,
                     CHA_tag); }

CHA_type make_CHA_M(CHR_type Raw_character)
  { CHA_type character = make_chunk(CHA);
    character->chr = Raw_character;
    return character; }

/*------------------------------ CNT declarations ----------------------------*/

BYT_type is_CNT(EXP_type Expression)
  { BYT_type tag = Grammar_Tag(Expression);
    return (tag == CNT_tag); }

CNT_type make_CNT_M(ENV_type Environment,
                    FRM_type Frame,
                    THR_type Thread)
  { CNT_type continuation = make_chunk(CNT);
    continuation->env = Environment;
    continuation->frm = Frame;
    continuation->thr = Thread;
    return continuation; }

/*------------------------------ DFF declarations ----------------------------*/

BYT_type is_DFF(EXP_type Expression)
  { BYT_type tag = Grammar_Tag(Expression);
    return (tag == DFF_tag); }

BYT_type is_DFZ(EXP_type Expression)
  { BYT_type tag = Grammar_Tag(Expression);
    return (tag == DFZ_tag); }

DFF_type make_DFF_M(SYM_type Name,
                    NBR_type Offset,
                    NBR_type Param_count,
                    NBR_type Frame_size,
                    EXP_type Body)
  { DFF_type definition = make_chunk(DFF);
    definition->nam = Name;
    definition->ofs = Offset;
    definition->par = Param_count;
    definition->siz = Frame_size;
    definition->bod = Body;
    return definition; }
    
DFF_type make_DFZ_M(SYM_type Name,
                    NBR_type Offset,
                    NBR_type Param_count,
                    NBR_type Frame_size,
                    EXP_type Body)
  { DFF_type definition = make_DFF_M(Name,
                                     Offset,
                                     Param_count,
                                     Frame_size,
                                     Body);
    Memory_Set_Tag((PTR_type)definition,
                   DFZ_tag);  
    return definition; }

/*------------------------------ DFV declarations ----------------------------*/

BYT_type is_DFV(EXP_type Expression)
  { BYT_type tag = Grammar_Tag(Expression);
    return (tag == DFV_tag); }

DFV_type make_DFV_M(NBR_type Offset,
                    EXP_type Expression)
  { DFV_type definition = make_chunk(DFV);
    definition->ofs = Offset;
    definition->exp = Expression;
    return definition; }

/*------------------------------ FLS declarations ----------------------------*/

BYT_type is_FLS(EXP_type Expression)
  { return (Expression == (EXP_type)Immediate_false_value); }

FLS_type make_FLS(NIL_type)
  { return (FLS_type)Immediate_false_value; }

/*------------------------------ GLB declarations ----------------------------*/

BYT_type is_GLB(EXP_type Expression)
  { return member_of(Expression,
                     GLB_tag); }

GLB_type make_GLB_M(NBR_type Scope,
                    NBR_type Offset)
  { GLB_type global = make_chunk(GLB);
    global->scp = Scope;
    global->ofs = Offset;
    return global; }

/*------------------------------ IFF declarations ----------------------------*/

BYT_type is_IFF(EXP_type Expression)
  { BYT_type tag = Grammar_Tag(Expression);
    return (tag == IFF_tag); }

IFF_type make_IFF_M(EXP_type Predicate,
                    EXP_type Consequent,
                    EXP_type Alternative)
  { IFF_type if_ = make_chunk(IFF);
    if_->prd = Predicate;
    if_->cns = Consequent;
    if_->alt = Alternative;
    return if_; }

/*------------------------------ IFZ declarations ----------------------------*/

BYT_type is_IFZ(EXP_type Expression)
  { BYT_type tag = Grammar_Tag(Expression);
    return (tag == IFZ_tag); }

IFZ_type make_IFZ_M(EXP_type Predicate,
                    EXP_type Consequent)
  { IFZ_type if_ = make_chunk(IFZ);
    if_->prd = Predicate;
    if_->cns = Consequent;
    return if_; }

/*------------------------------ LCK declarations ----------------------------*/

BYT_type is_LCK(EXP_type Expression)
  { return member_of(Expression,
                     LCK_tag); }

LCK_type make_LCK_M(MTX_type Mutex)
  { LCK_type lock = make_chunk(LCK);
    lock->mtx = Mutex;
    return lock; }

/*------------------------------ LCL declarations ----------------------------*/

BYT_type is_LCL(EXP_type Expression)
  { return member_of(Expression,
                     LCL_tag); }

LCL_type make_LCL_M(NBR_type Offset)
  { LCL_type local = make_chunk(LCL);
    local->ofs = Offset;
    return local; }

/*------------------------------ LMB declarations ----------------------------*/

BYT_type is_LMB(EXP_type Expression)
  { BYT_type tag = Grammar_Tag(Expression);
    return (tag == LMB_tag); }

BYT_type is_LMZ(EXP_type Expression)
  { BYT_type tag = Grammar_Tag(Expression);
    return (tag == LMZ_tag); }

LMB_type make_LMB_M(NBR_type Param_count,
                    NBR_type Frame_size,
                    EXP_type Body)
  { LMB_type lambda = make_chunk(LMB);
    lambda->par = Param_count;
    lambda->siz = Frame_size;
    lambda->bod = Body;
    return lambda; }

LMB_type make_LMZ_M(NBR_type Param_count,
                    NBR_type Frame_size,
                    EXP_type Body)
  { LMB_type lambda = make_LMB_M(Param_count,
                                 Frame_size,
                                 Body);
    Memory_Set_Tag((PTR_type)lambda,
                   LMZ_tag);  
    return lambda; }

/*------------------------------ MRK declarations ----------------------------*/

BYT_type is_MRK(EXP_type Expression)
  { BYT_type tag = Grammar_Tag(Expression);
    return (tag == MRK_tag); }

NIL_type apply_MRK(EXP_type Expression)
  { Memory_Set_Tag((PTR_type)Expression,
                   MRK_tag); }
                                      
/*------------------------------ NAT declarations ----------------------------*/

BYT_type is_NAT(EXP_type Expression)
  { return member_of(Expression,
                     NAT_tag); }

NAT_type make_NAT_M(NPR_type Native_procedure,
                    TXT_type Name)
  { NAT_type native;
    UNS_type size;
    size = string_to_expression_size(Name);
    native = make_chunk_with_offset(NAT,
                                    size);
    native->npr = Native_procedure;
    strcpy(native->nam,
           Name);
    return native; }

/*------------------------------ NBR declarations ----------------------------*/

BYT_type is_NBR(EXP_type Expression)
  { return Memory_Is_Immediate(Expression); }

NBR_type make_NBR(SGN_type Signed_number)
  { NBR_type number;
    number = Memory_Make_Immediate(Signed_number);
    return number; }

SGN_type get_NBR(NBR_type Number)
  { SGN_type Signed_number;
    Signed_number = Memory_Get_Immediate(Number);
    return Signed_number; }

/*------------------------------ NUL declarations ----------------------------*/

BYT_type is_NUL(EXP_type Expression)
  { return (Expression == Immediate_null_value); }

NUL_type make_NUL(NIL_type)
  { return Immediate_null_value; }

/*------------------------------ PAI declarations ----------------------------*/

BYT_type is_PAI(EXP_type Expression)
  { return member_of(Expression,
                     PAI_tag); }

PAI_type make_PAI_M(EXP_type Car,
                    EXP_type Cdr)
  { PAI_type pair = make_chunk(PAI);
    pair->car = Car;
    pair->cdr = Cdr;
    return pair; }

/*------------------------------ PRC declarations ----------------------------*/

BYT_type is_PRC(EXP_type Expression)
  { return member_of(Expression,
                     PRC_tag); }

BYT_type is_PRZ(EXP_type Expression)
  { return member_of(Expression,
                     PRZ_tag); }

PRC_type make_PRC_M(SYM_type Name,
                    NBR_type Param_count,
                    NBR_type Frame_size,
                    EXP_type Body,
                    ENV_type Environment)
  { PRC_type procedure = make_chunk(PRC);
    procedure->nam = Name;
    procedure->par = Param_count;
    procedure->siz = Frame_size;
    procedure->bod = Body;
    procedure->env = Environment;
    return procedure; }

PRC_type make_PRZ_M(SYM_type Name,
                    NBR_type Param_count,
                    NBR_type Frame_size,
                    EXP_type Body,
                    ENV_type Environment)
  { PRC_type procedure = make_PRC_M(Name,
                                    Param_count,
                                    Frame_size,
                                    Body,
                                    Environment);
    Memory_Set_Tag((PTR_type)procedure,
                   PRZ_tag);  
    return procedure; }

/*------------------------------ PRM declarations ----------------------------*/

BYT_type is_PRM(EXP_type Expression)
  { return member_of(Expression,
                     PRM_tag); }

PRM_type make_PRM_M(CID_type Contex_Id)
  { PRM_type promise = make_chunk(PRM);
    promise->cid = Contex_Id;
    return promise; }
                                      
/*------------------------------ PRT declarations ----------------------------*/

BYT_type is_PRT(EXP_type Expression)
  { BYT_type tag = Grammar_Tag(Expression);
    return (tag == PRT_tag); }

PRT_type make_PRT_M(EXP_type Body,
                    LCK_type Lock)
  { PRT_type protect = make_chunk(PRT);
    protect->bod = Body;
    protect->lck = Lock;
    return protect; }

/*------------------------------ QUO declarations ----------------------------*/

BYT_type is_QUO(EXP_type Expression)
  { BYT_type tag = Grammar_Tag(Expression);
    return (tag == QUO_tag); }

QUO_type make_QUO_M(EXP_type Expression)
  { QUO_type quote = make_chunk(QUO);
    quote->exp = Expression;
    return quote; }

/*------------------------------ REA declarations ----------------------------*/

BYT_type is_REA(EXP_type Expression)
  { return member_of(Expression,
                     REA_tag); }

REA_type make_REA_M(FLO_type Raw_real)
  { REA_type real = make_chunk(REA);
    real->flo = Raw_real;
    return real; }

/*------------------------------ SEQ declarations ----------------------------*/

static const UNS_type SEQ_size = 0;

BYT_type is_SEQ(EXP_type Expression)
  { return member_of(Expression,
                     SEQ_tag); }

SEQ_type make_SEQ_M(UNS_type Size)
  { SEQ_type sequence = (SEQ_type)make_chunk_with_offset(SEQ,
                                                         Size);
    return sequence; }

UNS_type size_SEQ(SEQ_type Sequence)
  { return (UNS_type) Memory_Get_Size((PTR_type)Sequence); }

/*------------------------------ SPN declarations ----------------------------*/

BYT_type is_SPN(EXP_type Expression)
  { BYT_type tag = Grammar_Tag(Expression);
    return (tag == SPN_tag); }

SPN_type make_SPN_M(EXP_type Expression)
  { SPN_type detachment = make_chunk(SPN);
    detachment->exp = Expression;
    return detachment; }

/*------------------------------ STG declarations ----------------------------*/

BYT_type is_STG(EXP_type Expression)
  { return member_of(Expression,
                     STG_tag); }

STG_type make_STG_M(NBR_type Scope,
                    NBR_type Offset,
                    EXP_type Expression)
  { STG_type set = make_chunk(STG);
    set->scp = Scope;
    set->ofs = Offset;
    set->exp = Expression;
    return set; }

/*------------------------------ STL declarations ----------------------------*/

BYT_type is_STL(EXP_type Expression)
  { return member_of(Expression,
                     STL_tag); }

STL_type make_STL_M(NBR_type Offset,
                    EXP_type Expression)
  { STL_type set = make_chunk(STL);
    set->ofs = Offset;
    set->exp = Expression;
    return set; }

/*------------------------------ STR declarations ----------------------------*/

BYT_type is_STR(EXP_type Expression)
  { return member_of(Expression,
                     STR_tag); }

STR_type make_STR_M(TXT_type Raw_string)
  { STR_type string;
    UNS_type size;
    size = string_to_expression_size(Raw_string);
    string = make_chunk_with_offset(STR,
                                    size);
    strcpy(string->txt,
           Raw_string);
    return string; }

UNS_type size_STR(TXT_type Raw_string)
  { return (UNS_type)string_to_expression_size(Raw_string); }

/*------------------------------ SYM declarations ----------------------------*/

BYT_type is_SYM(EXP_type Expression)
  { return member_of(Expression,
                     SYM_tag); }

SYM_type make_SYM_M(TXT_type Raw_string)
  { SYM_type symbol;
    UNS_type size;
    size = string_to_expression_size(Raw_string);
    symbol = make_chunk_with_offset(SYM,
                                    size);
    strcpy(symbol->txt,
           Raw_string);
    return symbol; }

UNS_type size_SYM(TXT_type Raw_string)
  { return (UNS_type)string_to_expression_size(Raw_string); }

/*------------------------------ THK declarations ----------------------------*/

BYT_type is_THK(EXP_type Expression)
  { return member_of(Expression,
                     THK_tag); }

THK_type make_THK_M(EXP_type Body,
                    NBR_type Frame_size)
  { THK_type thunk = make_chunk(THK);
    thunk->bod = Body;
    thunk->siz = Frame_size;
    return thunk; }

/*------------------------------ TRU declarations ----------------------------*/

BYT_type is_TRU(EXP_type Expression)
  { return (Expression == (EXP_type)Immediate_true_value); }

TRU_type make_TRU(NIL_type)
  { return (TRU_type)Immediate_true_value; }

/*------------------------------ USP declarations ----------------------------*/

BYT_type is_USP(EXP_type Expression)
  { return (Expression == (EXP_type)Immediate_unspecified_value); }

USP_type make_USP(NIL_type)
  { return (USP_type)Immediate_unspecified_value; }

/*------------------------------ TSK declarations ----------------------------*/

BYT_type is_TSK(EXP_type Expression)
  { BYT_type tag = Grammar_Tag(Expression);
    return (tag == TSK_tag); }

TSK_type make_TSK_M(PTH_type Pthread)
  { TSK_type task = make_chunk(TSK);
    task->pth = Pthread;
    return task; }

UNS_type size_TSK(NIL_type)
{ return TSK_size; }

/*------------------------------ VAL declarations ----------------------------*/

BYT_type is_VAL(EXP_type Expression)
  { BYT_type tag = Grammar_Tag(Expression);
    return (tag == VAL_tag); }

VAL_type convert_VAL(PRM_type Promise,
                     EXP_type Expression)
  { VAL_type value = (VAL_type)Promise;
    Memory_Set_Tag((PTR_type)value,
                   VAL_tag);
    value->exp = Expression;
    return value; }

/*------------------------------ VEC declarations ----------------------------*/

static const UNS_type VEC_size = 0;

BYT_type is_VEC(EXP_type Expression)
  { return member_of(Expression,
                     VEC_tag); }

VEC_type make_VEC_M(UNS_type Size)
  { VEC_type vector = (VEC_type)make_chunk_with_offset(VEC,
                                                       Size);
    return vector; }

UNS_type size_VEC(VEC_type Vector)
  { return (UNS_type) Memory_Get_Size((PTR_type)Vector); }

/*------------------------------ WHI declarations ----------------------------*/

BYT_type is_WHI(EXP_type Expression)
  { BYT_type tag = Grammar_Tag(Expression);
    return (tag == WHI_tag); }

WHI_type make_WHI_M(EXP_type Predicate,
                    EXP_type Body)
  { WHI_type while_ = make_chunk(WHI);
    while_->prd = Predicate;
    while_->bod = Body;
    return while_; }

/*----------------------------------------------------------------------------*/
                               
NIL_type Grammar_Initialize(NIL_type)
  { Grammar_Empty_Vector = make_VEC_M(0);
    Grammar_False        = make_FLS();
    Grammar_Null         = make_NUL();
    Grammar_One          = make_NBR(1);
    Grammar_True         = make_TRU();
    Grammar_Unspecified  = make_USP();
    Grammar_Zero         = make_NBR(0); }
