                     /*-----------------------------------*/
                     /*             >>>Slip<<<            */
                     /*            Theo D'Hondt           */
                     /*     VUB Software Languages Lab    */
                     /*             (c) 2010              */
                     /*-----------------------------------*/
                     /*      version 14: multicores       */
                     /*-----------------------------------*/
                     /*               main                */
                     /*-----------------------------------*/

#include <setjmp.h>
#include <stdio.h>
#include <string.h>

#include "SlipMain.h"

#include "SlipCompile.h"
#include "SlipContext.h"
#include "SlipEvaluate.h"
#include "SlipNative.h"
#include "SlipPool.h"
#include "SlipPrint.h"
#include "SlipRead.h"
#include "SlipStack.h"

/*------------------------------ private constants ---------------------------*/

enum { Root_size = 20 };

static const UNS_type Minimum_number_of_cells = 2048;

/*-------------------------------- private types -----------------------------*/

typedef enum { Initiate_REP = 0,
               Abort_REP    = 1,
               Proceed_REP  = 2} Tra_type;

typedef jmp_buf Exi_type;

/*------------------------------ private variables ---------------------------*/

static VEC_type Root;

static UNS_type Root_counter;
static UNS_type Total_number_of_cells;

static Exi_type Exit;

static ERF_type Root_references[Root_size];

/*------------------------------ public variables ----------------------------*/

CHR_type Main_Buffer[Main_Buffer_Size];

SYM_type Main_Anonymous;
SYM_type Main_Begin;
SYM_type Main_Define;
SYM_type Main_If;
SYM_type Main_Lambda;
SYM_type Main_Protect;
SYM_type Main_Quote;
SYM_type Main_Set;
SYM_type Main_Spawn;
SYM_type Main_While;
        
/*------------------------------ private functions ---------------------------*/

static NIL_type read_eval_print(NIL_type)
  { EXP_type compiled_expression,
             expression;
    Tra_type status;
    TXT_type input;
    Stack_Empty();
    Main_Print("\n>>>");
    input = Slip_Read();
    Dictionary_Checkpoint();
    if ((status = setjmp(Exit)) == Initiate_REP)
      { expression = Read_Parse_C(input);
        compiled_expression = Compile_Compile_C(expression);
        if ((status = setjmp(Exit)) == Initiate_REP)
          Evaluate_Evaluate(compiled_expression);
        if (status == Abort_REP)
          Context_Flush(); }
    if (status == Abort_REP)
      { Context_Memory_Initialize();
        Main_Claim_Default();
        Dictionary_Rollback_M();
        Main_Release_Default(); }}

static NIL_type vector_copy(VEC_type Old_vector, 
                     VEC_type New_vector)
  { UNS_type index,
             new_size,
             old_size;
    old_size = size_VEC(Old_vector);   
    new_size = size_VEC(New_vector);
    for (index = 1; 
         (index <= old_size)&&(index <= new_size);
         index++)
      New_vector[index] = Old_vector[index]; 
    for (; 
         index <= new_size;
         index++)
      New_vector[index] = Grammar_Null; }

/*------------------------------ hidden functions ----------------------------*/

NIL_type _Main_Register(ERF_type Reference)
  { if (Root_counter == Root_size)
      Slip_Fail(TMR_error);
    Root_references[Root_counter++] = Reference; }

/*------------------------------ public functions ----------------------------*/

NIL_type Main_Adjust_Global_Frame(NIL_type)
  { UNS_type global_frame_size;
    global_frame_size = Dictionary_Get_Global_Frame_Size();
    Context_Adjust_Global_Frame_C(global_frame_size); }

EXP_type Main_Error(BYT_type Error)
  { return Main_Error_Text(Error,
                           0); }

EXP_type Main_Error_Procedure(BYT_type Error    ,
                              PRC_type Procedure)
  { SYM_type symbol;
    symbol = Procedure->nam;
    return Main_Error_Symbol(Error,
                             symbol); }

EXP_type Main_Error_Procedure_Vararg(BYT_type Error    ,
                                     PRC_type Procedure)
  { SYM_type symbol;
    symbol = Procedure->nam;
    return Main_Error_Symbol(Error,
                             symbol); }

EXP_type Main_Error_Symbol(BYT_type Error ,
                           SYM_type Symbol)
  { TXT_type text;
    text = Symbol->txt;
    return Main_Error_Text(Error,
                           text); }

EXP_type Main_Error_Tag(BYT_type Error,
                        TAG_type Tag  )
  { sprintf(Main_Buffer,
            "%d",
            Tag);
    return Main_Error_Text(Error,
                           Main_Buffer); }

EXP_type Main_Error_Text(BYT_type Error,
                         TXT_type Text )
  { Slip_Error(Error,
               Text);
    longjmp(Exit,
            Abort_REP);
    return Grammar_Unspecified; }

NIL_type Main_Exit(NIL_type)
  { longjmp(Exit,
            Proceed_REP); }

NIL_type Main_Fail(BYT_type Error)
  { Slip_Fail(Error); }

VEC_type Main_Initialize_Vector_M(UNS_type Size)
  { UNS_type index;
    VEC_type vector = make_VEC_M(Size);
    for (index = 1;
         index <= Size;
         index++)
      vector[index] = Grammar_Null;
    return vector; }

NIL_type Main_Lock_Thread(LCK_type Lock)
  { Concurrency_Mutex_Lock(&Lock->mtx); }

LCK_type Main_Make_Lock_M(NIL_type)
  { LCK_type lock;
    MTX_type mutex;
    Concurrency_Initialize_Mutex(&mutex);
    lock = make_LCK_M(mutex);   /* --- */
    return lock; }

NIL_type Main_Print(TXT_type Text)
  { Slip_Print(Text); }

TXT_type Main_Read(TXT_type Name)
  { TXT_type input;
    if (Name == 0)
      input = Slip_Read();
    else
      input = Slip_Load(Name);
    return input; }

NIL_type Main_Reallocate_Vector_M(VRF_type Vector_reference,
                                  UNS_type Increment)
  { UNS_type new_size,
             old_size;
    VEC_type new_vector,
             old_vector;
    old_vector = *Vector_reference;
    old_size = size_VEC(old_vector);   
    new_size = old_size + Increment;
    new_vector = make_VEC_M(new_size); 
    vector_copy(old_vector,
                new_vector);
    *Vector_reference = new_vector; }

NIL_type Main_Reclaim_C(NIL_type)
  { Context_Collect_C(); }

NIL_type Main_Replace(VEC_type Old_vector, 
                      VEC_type New_vector)
  { Memory_Replace((PTR_type)Old_vector,
                   (PTR_type)New_vector);
    vector_copy(Old_vector,
                New_vector); }

PAI_type Main_Reverse(PAI_type This)
  { PAI_type next,
             pair,
             previous;
    previous = Grammar_Null;
    for (pair = This;
         is_PAI(pair);
         pair = next)
      { next = pair->cdr;
        pair->cdr = previous;
        previous = pair; }
    return previous; }
                                                
NIL_type Main_Terminate(NIL_type)
  { longjmp(Exit,
            Abort_REP); }

NIL_type Main_Unlock_Thread(LCK_type Lock)
  { Concurrency_Mutex_Unlock(&Lock->mtx); }

/*---------------------------- implemented callbacks -------------------------*/

NIL_type Memory_After_Reclaim(PTR_type New_root)
  { EXP_type expression;
    ERF_type reference;
    UNS_type index;
    Root = (EXP_type)New_root;
    for (index = 0;
         index < Root_counter;
         index++)
      { reference = Root_references[index];
        expression = Root[index + 1];
        *reference = expression; } }

PTR_type Memory_Before_Reclaim(NIL_type)
  { ERF_type reference;
    UNS_type index;
    for (index = 0;
         index < Root_counter;
         index++)
      { reference = Root_references[index];
        Root[index + 1] = *reference; }
    return (PTR_type)Root; }

/*------------------------------ exported functions --------------------------*/

void Slip_REP(char * Memory,
              int    Size)
  { Total_number_of_cells = Size / sizeof(CEL_type);
    Memory_Initialize(Memory,
                      Total_number_of_cells);
    Context_Memory_Initialize();
    Main_Claim_Dynamic(Minimum_number_of_cells);
    Grammar_Initialize();
    Root = Main_Initialize_Vector_M(Root_size);
    Root_counter = 0;
    Pool_Initialize_M();
    Main_Anonymous = Pool_Enter_M(Main_Anonymous_String);
    Main_Begin     = Pool_Enter_M(Main_Begin_String);
    Main_Define    = Pool_Enter_M(Main_Define_String);
    Main_If        = Pool_Enter_M(Main_If_String);
    Main_Lambda    = Pool_Enter_M(Main_Lambda_String);
    Main_Protect   = Pool_Enter_M(Main_Protect_String);
    Main_Quote     = Pool_Enter_M(Main_Quote_String);
    Main_Set       = Pool_Enter_M(Main_Set_String);
    Main_Spawn     = Pool_Enter_M(Main_Spawn_String);
    Main_While     = Pool_Enter_M(Main_While_String);
    Stack_Initialize_M();
    Read_Initialize();
    Compile_Initialize();
    Context_Initialize_M();
    Evaluate_Initialize();
    Native_Initialize_M();
    Print_Initialize();
    Main_Release_Dynamic(Minimum_number_of_cells);
    Main_Print("cpSlip/c version 14: multicores");
    Main_Register(Grammar_Empty_Vector);
    Main_Register(Main_Anonymous);
    Main_Register(Main_Begin);
    Main_Register(Main_Define);
    Main_Register(Main_If);
    Main_Register(Main_Lambda);
    Main_Register(Main_Protect);
    Main_Register(Main_Quote);
    Main_Register(Main_Set);
    Main_Register(Main_Spawn);
    Main_Register(Main_While);
    for (;;)
      read_eval_print(); }