                     /*-----------------------------------*/
                     /*             >>>Slip<<<            */
                     /*            Theo D'Hondt           */
                     /*     VUB Software Languages Lab    */
                     /*             (c) 2010              */
                     /*-----------------------------------*/
                     /*      version 14: multicores       */
                     /*-----------------------------------*/
                     /*              natives              */
                     /*-----------------------------------*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "SlipMain.h"

#include "SlipCompile.h"
#include "SlipContext.h"
#include "SlipDictionary.h"
#include "SlipEvaluate.h"
#include "SlipNative.h"
#include "SlipPool.h"
#include "SlipPrint.h"
#include "SlipRead.h"
#include "SlipThread.h"

/*------------------------------ private constants ---------------------------*/

enum { Natives_table_size = 47 };

/*------------------------------ private variables ---------------------------*/

static TXT_type Natives_table [Natives_table_size];

/*------------------------------ private functions ---------------------------*/

static NIL_type enter_native_name(TXT_type String,
                                  UNS_type Offset)
  { if (Offset >= Natives_table_size)
      Slip_Fail(NTO_error);
    Natives_table[Offset - 1] = String; }

static TXT_type get_native_name(UNS_type Offset)
  { if (Offset >= Natives_table_size)
      return 0;
    return Natives_table[Offset - 1]; }
    
static REA_type make_real_number_C(FLO_type Raw_real)
  { REA_type real;
    Main_Claim_Default();
    real = make_REA_M(Raw_real);
    Main_Release_Default();
    return real; }

static NIL_type native_define_M(TXT_type String,
                                NPR_type Native_procedure)
  { EXP_type native;
    SYM_type variable;
    UNS_type raw_offset;
    variable = Pool_Enter_M(String);
    native = make_NAT_M(Native_procedure,
                        String);
    raw_offset = Dictionary_Define_M(variable);
    enter_native_name(String,
                      raw_offset);
    Context_Global_Frame_Set(native,
                             raw_offset); }

/*----------------------------------------------------------------------------*/
/*--------------------------------- variables --------------------------------*/
/*----------------------------------------------------------------------------*/

/*---------------------------- meta-circularity-level ------------------------*/

static const TXT_type mcl_string = "circularity-level";

static NIL_type initialize_circularity_level_M(NIL_type)
  { SYM_type variable;
    UNS_type raw_offset;
    variable = Pool_Enter_M(mcl_string);
    raw_offset = Dictionary_Define_M(variable);
    enter_native_name(mcl_string,
                      raw_offset);
    Context_Global_Frame_Set(Grammar_Zero,
                             raw_offset); }

/*----------------------------------------------------------------------------*/
/*-------------------------------- arithmetic --------------------------------*/
/*----------------------------------------------------------------------------*/

/*------------------------------------ + -------------------------------------*/

static const TXT_type add_string = "+";

static EXP_type add_floats_C(FLO_type Raw_real,
                             FRM_type Arguments,
                             UNS_type Index)
  { EXP_type value;
    FLO_type raw_real,
             raw_real_sum;
    SGN_type signed_number;
    REA_type real;
    TAG_type tag;
    UNS_type index,
             size;
    raw_real_sum = Raw_real;
    size = size_VEC(Arguments);
    for (index = Index;
         index <= size;
         index++)
      { value = Arguments[index];
        tag = Grammar_Tag(value);
        switch (tag)
          { case NBR_tag:
              signed_number = get_NBR(value);
              raw_real_sum += signed_number;
              break;
            case REA_tag:
              real = value;
              raw_real = real->flo;
              raw_real_sum += raw_real;
              break;
            default:
              return Main_Error_Text(ARN_error,
                                     add_string); }}
    return make_real_number_C(raw_real_sum); }

static EXP_type add_integers_C(NBR_type Number,
                               FRM_type Arguments)
  { EXP_type value;
    FLO_type raw_real,
             raw_real_sum;
    SGN_type signed_number,
             raw_number_sum;
    REA_type real;
    TAG_type tag;
    UNS_type index,
             size;
    raw_number_sum = get_NBR(Number);
    size = size_VEC(Arguments);
    for (index = 2;
         index <= size;
         index++)
      { value = Arguments[index];
        tag = Grammar_Tag(value);
        switch (tag)
          { case NBR_tag:
              signed_number = get_NBR(value);
              raw_number_sum += signed_number;
              break;
            case REA_tag:
              real = value;
              raw_real = real->flo;
              raw_real_sum = raw_number_sum + raw_real;
              return add_floats_C(raw_real_sum,
                                  Arguments,
                                  index + 1);
            default:
              return Main_Error_Text(ARN_error,
                                     add_string); }}
    return make_NBR(raw_number_sum); }

static NIL_type add_native_C(CID_type Context_id,
                             EXP_type Tailposition)
  { EXP_type value;
    FLO_type raw_real;
    FRM_type arguments;
    REA_type real;
    TAG_type tag;
    UNS_type size;
    arguments = Context_Get_Expression(Context_id);
    size = size_VEC(arguments);
    if (size == 0)
      value = Grammar_Zero;
    else
      { value = arguments[1];
        if (size > 1)
          { tag = Grammar_Tag(value);
            switch (tag)
              { case NBR_tag:
                  value = add_integers_C(value,
                                         arguments);
                  break;
                case REA_tag:
                  real = value;
                  raw_real = real->flo;
                  value = add_floats_C(raw_real,
                                       arguments,
                                       2);
                  break;
               default:
                 Main_Error_Text(ARN_error,
                                 add_string); }}}
    Context_Release_Frame(Context_id,
                          arguments);
    Context_Set_Expression(Context_id,
                           value); }
                                    
/*----------------------------------------------------------------------------*/

static NIL_type initialize_add_native_M(NIL_type)
  { native_define_M(add_string,
                    add_native_C); }

/*------------------------------------ - -------------------------------------*/

static const TXT_type sub_string = "-";

static EXP_type subtract_floats_C(FLO_type Raw_real,
                                  FRM_type Arguments,
                                  UNS_type Index)
  { EXP_type value;
    FLO_type raw_real,
             raw_real_sum;
    SGN_type signed_number;
    REA_type real;
    TAG_type tag;
    UNS_type index,
             size;
    raw_real_sum = Raw_real;
    size = size_VEC(Arguments);
    for (index = Index;
         index <= size;
         index++)
      { value = Arguments[index];
        tag = Grammar_Tag(value);
        switch (tag)
          { case NBR_tag:
              signed_number = get_NBR(value);
              raw_real_sum -= signed_number;
              break;
            case REA_tag:
              real = value;
              raw_real = real->flo;
              raw_real_sum -= raw_real;
              break;
            default:
              return Main_Error_Text(ARN_error,
                                     sub_string); }}
    return make_real_number_C(raw_real_sum); }

static EXP_type subtract_integers_C(NBR_type Number,
                                    FRM_type Arguments)
  { EXP_type value;
    FLO_type raw_real,
             raw_real_sum;
    SGN_type signed_number,
             raw_number_sum;
    REA_type real;
    TAG_type tag;
    UNS_type index,
             size;
    raw_number_sum = get_NBR(Number);
    size = size_VEC(Arguments);
    for (index = 2;
         index <= size;
         index++)
      { value = Arguments[index];
        tag = Grammar_Tag(value);
        switch (tag)
          { case NBR_tag:
              signed_number = get_NBR(value);
              raw_number_sum -= signed_number;
              break;
            case REA_tag:
              real = value;
              raw_real = real->flo;
              raw_real_sum = raw_number_sum - raw_real;
              return subtract_floats_C(raw_real_sum,
                                       Arguments,
                                       index + 1);
            default:
              return Main_Error_Text(ARN_error,
                                     sub_string); }}
    return make_NBR(raw_number_sum); }

static NIL_type subtract_native_C(CID_type Context_id,
                                EXP_type Tailposition)
  { EXP_type value;
    FLO_type raw_real;
    FRM_type arguments;
    SGN_type signed_number;
    REA_type real;
    TAG_type tag;
    UNS_type size;
    arguments = Context_Get_Expression(Context_id);
    size = size_VEC(arguments);
    if (size == 0)
      Main_Error_Text(ALO_error,
                      sub_string);
    value = arguments[1];
    tag = Grammar_Tag(value);
    if (size == 1)
      switch (tag)
        { case NBR_tag:
            signed_number = get_NBR(value);
            value = make_NBR(- signed_number);
            break;
          case REA_tag:
            real = value;
            raw_real = real->flo;
            value = make_real_number_C(- raw_real);
            break;
          default:
            Main_Error_Text(ARN_error,
                            sub_string); }
    else
      switch (tag)
        { case NBR_tag:
            value = subtract_integers_C(value,
                                        arguments);
            break;
          case REA_tag:
            real = value;
            raw_real = real->flo;
            value = subtract_floats_C(raw_real,
                                      arguments,
                                      2);
            break;
        default:
          Main_Error_Text(ARN_error,
                          sub_string); }                              
    Context_Release_Frame(Context_id,
                          arguments);
    Context_Set_Expression(Context_id,
                           value); }

/*----------------------------------------------------------------------------*/

static NIL_type initialize_subtract_native_M(NIL_type)
  { native_define_M(sub_string,
                    subtract_native_C); }

/*------------------------------------ * -------------------------------------*/

static const TXT_type mul_string = "*";

static EXP_type multiply_floats_C(FLO_type Raw_real,
                                  FRM_type Arguments,
                                  UNS_type Index)
  { EXP_type value;
    FLO_type raw_real,
             raw_real_product;
    SGN_type signed_number;
    REA_type real;
    TAG_type tag;
    UNS_type index,
             size;
    raw_real_product = Raw_real;
    size = size_VEC(Arguments);
    for (index = Index;
         index <= size;
         index++)
      { value = Arguments[index];
        tag   = Grammar_Tag(value);
        switch (tag)
          { case NBR_tag:
              signed_number = get_NBR(value);
              raw_real_product *= signed_number;
              break;
            case REA_tag:
              real = value;
              raw_real = real->flo;
              raw_real_product *= raw_real;
              break;
            default:
              return Main_Error_Text(ARN_error,
                                     mul_string); }}
    return make_real_number_C(raw_real_product); }

static EXP_type multiply_integers_C(NBR_type Number,
                                    FRM_type Arguments)
  { EXP_type value;
    FLO_type raw_real,
             raw_real_product;
    SGN_type signed_number,
             raw_number_product;
    REA_type real;
    TAG_type tag;
    UNS_type index,
             size;
    raw_number_product = get_NBR(Number);
    size = size_VEC(Arguments);
    for (index = 2;
         index <= size;
         index++)
      { value  = Arguments[index];
        tag = Grammar_Tag(value);
        switch (tag)
          { case NBR_tag:
              signed_number = get_NBR(value);
              raw_number_product *= signed_number;
              break;
            case REA_tag:
              real = value;
              raw_real = real->flo;
              raw_real_product = raw_number_product * raw_real;
              return multiply_floats_C(raw_real_product,
                                       Arguments,
                                       index + 1);
            default:
              return Main_Error_Text(ARN_error,
                                     mul_string); }}
    return make_NBR(raw_number_product); }

static NIL_type multiply_native_C(CID_type Context_id,
                                  EXP_type Tailposition)
  { EXP_type value;
    FLO_type raw_real;
    FRM_type arguments;
    REA_type real;
    TAG_type tag;
    UNS_type size;
    arguments = Context_Get_Expression(Context_id);
    size = size_VEC(arguments);
    if (size == 0)
      value = Grammar_One;
    else
      { value = arguments[1];
        if (size > 1)
          { tag = Grammar_Tag(value);
            switch (tag)
              { case NBR_tag:
                  value = multiply_integers_C(value,
                                              arguments);
                  break;
                case REA_tag:
                  real = value;
                  raw_real = real->flo;
                  value = multiply_floats_C(raw_real,
                                            arguments,
                                            2);
                default:
                  Main_Error_Text(ARN_error,
                                  mul_string); }}}
    Context_Release_Frame(Context_id,
                          arguments);
    Context_Set_Expression(Context_id,
                           value); }

/*----------------------------------------------------------------------------*/

static NIL_type initialize_multiply_native_M(NIL_type)
  { native_define_M(mul_string,
                    multiply_native_C); }

/*------------------------------------ / -------------------------------------*/

static const TXT_type div_string = "/";

static EXP_type divide_floats_C(FLO_type Raw_real,
                                FRM_type Arguments)
  { EXP_type value;
    FLO_type raw_real,
             raw_real_product;
    SGN_type signed_number;
    REA_type real;
    TAG_type tag;
    UNS_type index,
             size;
    raw_real_product = Raw_real;
    size = size_VEC(Arguments);
    for (index = 2;
         index <= size;
         index++)
      { value = Arguments[index];
        tag = Grammar_Tag(value);
        switch (tag)
          { case NBR_tag:
              signed_number = get_NBR(value);
              raw_real_product /= signed_number;
              break;
            case REA_tag:
              real = value;
              raw_real = real->flo;
              raw_real_product /= raw_real;
              break;
            default:
              return Main_Error_Text(ARN_error,
                                     div_string); }}
    return make_real_number_C(raw_real_product); }

static NIL_type divide_native_C(CID_type Context_id,
                                EXP_type Tailposition)
  { EXP_type value;
    FLO_type raw_real;
    FRM_type arguments;
    SGN_type signed_number;
    REA_type real;
    TAG_type tag;
    UNS_type size;
    arguments = Context_Get_Expression(Context_id);
    size = size_VEC(arguments);
    if (size == 0)
      Main_Error_Text(ALO_error,
                      div_string);
    value = arguments[1];
    tag = Grammar_Tag(value);
    if (size == 1)
      switch (tag)
        { case NBR_tag:
            signed_number = get_NBR(value);
            value = make_real_number_C(1.0 / signed_number);
            break;
          case REA_tag:
            real = value;
            raw_real = real->flo;
            value = make_real_number_C(1.0 / raw_real);
            break;
          default:
            Main_Error_Text(ARN_error,
                            div_string); }
    else
      switch (tag)
        { case NBR_tag:
            value = divide_floats_C(get_NBR(value),
                                    arguments);
            break;
          case REA_tag:
            real = value;
            raw_real = real->flo;
            value = divide_floats_C(raw_real,
                                    arguments);
            break;
          default:
            Main_Error_Text(ARN_error,
                            div_string); }
    Context_Release_Frame(Context_id,
                          arguments);
    Context_Set_Expression(Context_id,
                           value); }

/*----------------------------------------------------------------------------*/

static NIL_type initialize_divide_native_M(NIL_type)
  { native_define_M(div_string,
                    divide_native_C); }

/*--------------------------------- quotient ---------------------------------*/

static const TXT_type quo_string = "quotient";

static NIL_type quotient_native(CID_type Context_id,
                                EXP_type Tailposition)
  { FRM_type arguments;
    SGN_type raw_left,
             signed_number,
             raw_right;
    NBR_type left,
             number,
             right;
    UNS_type size;
    arguments = Context_Get_Expression(Context_id);
    size = size_VEC(arguments);
    if (size != 2)
      Main_Error_Text(EX2_error,
                          quo_string);
    left = arguments[1];
    right = arguments[2];
    Context_Release_Frame(Context_id,
                          arguments);
    if (!is_NBR(left) || !is_NBR(right))
      Main_Error_Text(ARN_error,
                          quo_string);
    raw_left = get_NBR(left);
    raw_right = get_NBR(right);
    signed_number = raw_left / raw_right;
    number = make_NBR(signed_number);
    Context_Set_Expression(Context_id,
                           number); }

/*----------------------------------------------------------------------------*/

static NIL_type initialize_quotient_native_M(NIL_type)
  { native_define_M(quo_string,
                    quotient_native); }

/*-------------------------------- remainder ---------------------------------*/

static const TXT_type rem_string = "remainder";

static NIL_type remainder_native(CID_type Context_id,
                                 EXP_type Tailposition)
  { FRM_type arguments;
    SGN_type raw_left,
             signed_number,
             raw_right;
    NBR_type left,
             number,
             right;
    UNS_type size;
    arguments = Context_Get_Expression(Context_id);
    size = size_VEC(arguments);
    if (size != 2)
      Main_Error_Text(EX2_error,
                          rem_string);
    left = arguments[1];
    right = arguments[2];
    Context_Release_Frame(Context_id,
                          arguments);
    if (!is_NBR(left) || !is_NBR(right))
      Main_Error_Text(ARN_error,
                      rem_string);
    raw_left = get_NBR(left);
    raw_right = get_NBR(right);
    signed_number = raw_left % raw_right;
    number = make_NBR(signed_number);
    Context_Set_Expression(Context_id,
                           number); }
                                    
/*----------------------------------------------------------------------------*/

static NIL_type initialize_remainder_native_M(NIL_type)
  { native_define_M(rem_string,
                    remainder_native); }
                  
/*----------------------------------------------------------------------------*/
/*---------------------------- utility functions -----------------------------*/
/*----------------------------------------------------------------------------*/

/*------------------------------------ abs -----------------------------------*/

static const TXT_type abs_string = "abs";

static NIL_type abs_native_C(CID_type Context_id,
                             EXP_type Tailposition)
  { EXP_type value;
    FRM_type arguments;
    FLO_type raw_float;
    SGN_type signed_number;
    REA_type real;
    TAG_type tag;
    UNS_type size;
    arguments = Context_Get_Expression(Context_id);
    size = size_VEC(arguments);
    if (size != 1)
      Main_Error_Text(EX1_error,
                          abs_string);
    value = arguments[1];
    Context_Release_Frame(Context_id,
                          arguments);
    tag = Grammar_Tag(value);
    switch (tag)
      { case NBR_tag:
          signed_number = get_NBR(value);
          signed_number = abs(signed_number);
          value = make_NBR(signed_number);
          break;
        case REA_tag:
          real = value;
          raw_float = real->flo;
          raw_float = abs(raw_float);
          value = make_real_number_C(raw_float);
          break;
        default:
          Main_Error_Text(ARN_error,
                          abs_string); }
    Context_Set_Expression(Context_id,
                           value); }
        
/*----------------------------------------------------------------------------*/

static NIL_type initialize_abs_native_M(NIL_type)
  { native_define_M(abs_string,
                    abs_native_C); }

/*----------------------------------------------------------------------------*/
/*------------------------- transcendent functions ---------------------------*/
/*----------------------------------------------------------------------------*/

/*----------------------------------- sqrt -----------------------------------*/

static const TXT_type sqt_string = "sqrt";

static NIL_type sqrt_native_C(CID_type Context_id,
                              EXP_type Tailposition)
  { EXP_type value;
    FRM_type arguments;
    FLO_type raw_float;
    SGN_type signed_number;
    REA_type real;
    TAG_type tag;
    UNS_type size;
    arguments = Context_Get_Expression(Context_id);
    size = size_VEC(arguments);
    if (size != 1)
      Main_Error_Text(EX1_error,
                      sqt_string);
    value = arguments[1];
    Context_Release_Frame(Context_id,
                          arguments);
    tag = Grammar_Tag(value);
    switch (tag)
      { case NBR_tag:
          signed_number = get_NBR(value);
          raw_float = sqrt(signed_number);
          break;
        case REA_tag:
          real = value;
          raw_float = real->flo;
          raw_float = sqrt(raw_float);
          break;
        default:
          raw_float = 0;
          Main_Error_Text(ARN_error,
                          sqt_string); }
    real = make_real_number_C(raw_float);
    Context_Set_Expression(Context_id,
                           real); }
        
/*----------------------------------------------------------------------------*/

static NIL_type initialize_sqrt_native_M(NIL_type)
  { native_define_M(sqt_string,
                    sqrt_native_C); }

/*----------------------------------------------------------------------------*/
/*-------------------------------- comparison --------------------------------*/
/*----------------------------------------------------------------------------*/

/*------------------------------------ = -------------------------------------*/

static const TXT_type eql_string = "=";

static EXP_type equal_floats(FLO_type Raw_real,
                             FRM_type Arguments,
                             UNS_type Index)
  { EXP_type value;
    FLO_type raw_previous_real,
             raw_real;
    REA_type real;
    TAG_type tag;
    UNS_type index,
             size;
    raw_previous_real = Raw_real;
    size = size_VEC(Arguments);
    for (index = Index;
         index <= size;
         index++)
      { value = Arguments[index];
        tag = Grammar_Tag(value);
        switch (tag)
          { case NBR_tag:
              raw_real = get_NBR(value);
              break;
            case REA_tag:
              real = value;
              raw_real = real->flo;
              break;
            default:
              return Main_Error_Text(ARN_error,
                                     eql_string); }
        if (raw_previous_real != raw_real)
          return Grammar_False;
        raw_previous_real = raw_real; }
    return Grammar_True; }
   
static EXP_type equal_integers(NBR_type Number,
                               FRM_type Arguments)
  { EXP_type value;
    FLO_type raw_real;
    SGN_type signed_number,
             raw_previous_number;
    REA_type real;
    TAG_type tag;
    UNS_type index,
             size;
    raw_previous_number = get_NBR(Number);
    size = size_VEC(Arguments);
    for (index = 2;
         index <= size;
         index++)
      { value = Arguments[index];
        tag = Grammar_Tag(value);
        switch (tag)
          { case NBR_tag:
              signed_number = get_NBR(value);
              if (raw_previous_number != signed_number)
                return Grammar_False;
              break;
            case REA_tag:
              real = value;
              raw_real = real->flo;
              if (raw_previous_number != raw_real)
                return Grammar_False;
              return equal_floats(raw_real,
                                  Arguments,
                                  index + 1);
            default:
              return Main_Error_Text(ARN_error,
                                     eql_string); }
        raw_previous_number = signed_number; }
    return Grammar_True; }
    
static NIL_type equal_native(CID_type Context_id,
                             EXP_type Tailposition)
  { EXP_type boolean,
             value;
    FLO_type raw_real;
    FRM_type arguments;
    REA_type real;
    TAG_type tag;
    UNS_type size;
    arguments = Context_Get_Expression(Context_id);
    size = size_VEC(arguments);
    if (size < 2)
      Main_Error_Text(AL2_error,
                      eql_string);
    value = arguments[1];
    tag = Grammar_Tag(value);
    switch (tag)
      { case NBR_tag:
          boolean = equal_integers(value,
                                   arguments);
          break;
        case REA_tag:
          real = value;
          raw_real = real->flo;
          boolean = equal_floats(raw_real,
                                 arguments,
                                 2);
          break;
        default:
          boolean = Grammar_False;
          Main_Error_Text(ARN_error,
                          eql_string); }
    Context_Release_Frame(Context_id,
                          arguments);
    Context_Set_Expression(Context_id,
                           boolean); }
    
/*----------------------------------------------------------------------------*/

static NIL_type initialize_equal_native_M(NIL_type)
  { native_define_M(eql_string,
                    equal_native); }

/*------------------------------------ > -------------------------------------*/

static const TXT_type grt_string = ">";

static EXP_type greater_floats(FLO_type Raw_real,
                               FRM_type Arguments,
                               UNS_type Index)
  { EXP_type value;
    FLO_type raw_previous_real,
             raw_real;
    REA_type real;
    TAG_type tag;
    UNS_type index,
             size;
    raw_previous_real = Raw_real;
    size = size_VEC(Arguments);
    for (index = Index;
         index <= size;
         index++)
      { value = Arguments[index];
        tag = Grammar_Tag(value);
        switch (tag)
          { case NBR_tag:
              raw_real = get_NBR(value);
              break;
            case REA_tag:
              real = value;
              raw_real = real->flo;
              break;
            default:
              return Main_Error_Text(ARN_error,
                                     grt_string); }
        if (raw_previous_real <= raw_real)
          return Grammar_False;
        raw_previous_real = raw_real; }
   return Grammar_True; }

static EXP_type greater_integers(NBR_type Number,
                                 FRM_type Arguments)
  { EXP_type value;
    FLO_type raw_real;
    SGN_type signed_number,
             raw_previous_number;
    REA_type real;
    TAG_type tag;
    UNS_type index,
             size;
    raw_previous_number = get_NBR(Number);
    size = size_VEC(Arguments);
    for (index = 2;
         index <= size;
         index++)
      { value = Arguments[index];
        tag = Grammar_Tag(value);
        switch (tag)
          { case NBR_tag:
              signed_number = get_NBR(value);
              if (raw_previous_number <= signed_number)
                return Grammar_False;
              break;
            case REA_tag:
              real = value;
              raw_real = real->flo;
              if (raw_previous_number <= raw_real)
                return Grammar_False;
              return greater_floats(raw_real,
                                    Arguments,
                                    index + 1);
            default:
              return Main_Error_Text(ARN_error,
                                     grt_string); }
        raw_previous_number = signed_number; }
    return Grammar_True; }

static NIL_type greater_native(CID_type Context_id,
                               EXP_type Tailposition)
  { EXP_type boolean,
             value;
    FLO_type raw_real;
    FRM_type arguments;
    REA_type real;
    TAG_type tag;
    UNS_type size;
    arguments = Context_Get_Expression(Context_id);
    size = size_VEC(arguments);
    if (size < 2)
      Main_Error_Text(AL2_error,
                      grt_string);
    value = arguments[1];
    tag = Grammar_Tag(value);
    switch (tag)
      { case NBR_tag:
          boolean = greater_integers(value,
                                     arguments);
          break;
        case REA_tag:
          real = value;
          raw_real = real->flo;
          boolean = greater_floats(raw_real,
                                   arguments,
                                   2);
          break;
        default:
          boolean = Grammar_False;
          Main_Error_Text(ARN_error,
                          grt_string); }
    Context_Release_Frame(Context_id,
                          arguments);
    Context_Set_Expression(Context_id,
                           boolean); }

/*----------------------------------------------------------------------------*/

static NIL_type initialize_greater_native_M(NIL_type)
  { native_define_M(grt_string,
                    greater_native); }

/*------------------------------------ < -------------------------------------*/

static const TXT_type lss_string = "<";

static EXP_type less_floats(FLO_type Raw_real,
                            FRM_type Arguments,
                            UNS_type Index)
  { EXP_type value;
    FLO_type raw_previous_real,
             raw_real;
    REA_type real;
    TAG_type tag;
    UNS_type index,
             size;
    raw_previous_real = Raw_real;
    size = size_VEC(Arguments);
    for (index = Index;
         index <= size;
         index++)
      { value = Arguments[index];
        tag = Grammar_Tag(value);
        switch (tag)
          { case NBR_tag:
              raw_real = get_NBR(value);
              break;
            case REA_tag:
              real = value;
              raw_real = real->flo;
              break;
            default:
              return Main_Error_Text(ARN_error,
                                     lss_string); }
        if (raw_previous_real >= raw_real)
          return Grammar_False;
        raw_previous_real = raw_real; }
   return Grammar_True; }

static EXP_type less_integers(NBR_type Number,
                              FRM_type Arguments)
  { EXP_type value;
    FLO_type raw_real;
    SGN_type signed_number,
             raw_previous_number;
    REA_type real;
    TAG_type tag;
    UNS_type index,
             size;
    raw_previous_number = get_NBR(Number);
    size = size_VEC(Arguments);
    for (index = 2;
         index <= size;
         index++)
      { value = Arguments[index];
        tag = Grammar_Tag(value);
        switch (tag)
          { case NBR_tag:
              signed_number = get_NBR(value);
              if (raw_previous_number >= signed_number)
                return Grammar_False;
              break;
            case REA_tag:
              real = value;
              raw_real = real->flo;
              if (raw_previous_number >= raw_real)
                return Grammar_False;
              return less_floats(raw_real,
                                 Arguments,
                                 index + 1);
            default:
              return Main_Error_Text(ARN_error,
                                     lss_string); }
        raw_previous_number = signed_number; }
    return Grammar_True; }
    
static NIL_type less_native(CID_type Context_id,
                            EXP_type Tailposition)
  { EXP_type boolean,
             value;
    FLO_type raw_real;
    FRM_type arguments;
    REA_type real;
    TAG_type tag;
    UNS_type size;
    arguments = Context_Get_Expression(Context_id);
    size = size_VEC(arguments);
    if (size < 2)
      Main_Error_Text(AL2_error,
                      lss_string);
    value = arguments[1];
    tag = Grammar_Tag(value);
    switch (tag)
      { case NBR_tag:
          boolean = less_integers(value,
                                  arguments);
          break;
        case REA_tag:
          real = value;
          raw_real = real->flo;
          boolean = less_floats(raw_real,
                                arguments,
                                2);
          break;
        default:
          boolean = Grammar_False;
          Main_Error_Text(ARN_error,
                          lss_string); }
    Context_Release_Frame(Context_id,
                          arguments);
    Context_Set_Expression(Context_id,
                           boolean); }
    
/*----------------------------------------------------------------------------*/

static NIL_type initialize_less_native_M(NIL_type)
  { native_define_M(lss_string,
                    less_native); }

/*------------------------------------ >= ------------------------------------*/

static const TXT_type geq_string = ">=";

static EXP_type greater_or_equal_floats(FLO_type Raw_real,
                                        FRM_type Arguments,
                                        UNS_type Index)
  { EXP_type value;
    FLO_type raw_previous_real,
             raw_real;
    REA_type real;
    TAG_type tag;
    UNS_type index,
             size;
    raw_previous_real = Raw_real;
    size = size_VEC(Arguments);
    for (index = Index;
         index <= size;
         index++)
      { value = Arguments[index];
        tag = Grammar_Tag(value);
        switch (tag)
          { case NBR_tag:
              raw_real = get_NBR(value);
              break;
            case REA_tag:
              real = value;
              raw_real = real->flo;
              break;
            default:
              return Main_Error_Text(ARN_error,
                                     geq_string); }
        if (raw_previous_real < raw_real)
          return Grammar_False;
        raw_previous_real = raw_real; }
   return Grammar_True; }

static EXP_type greater_or_equal_integers(NBR_type Number,
                                          FRM_type Arguments)
  { EXP_type value;
    FLO_type raw_real;
    SGN_type signed_number,
             raw_previous_number;
    REA_type real;
    TAG_type tag;
    UNS_type index,
             size;
    raw_previous_number = get_NBR(Number);
    size = size_VEC(Arguments);
    for (index = 2;
         index <= size;
         index++)
      { value = Arguments[index];
        tag = Grammar_Tag(value);
        switch (tag)
          { case NBR_tag:
              signed_number = get_NBR(value);
              if (raw_previous_number < signed_number)
                return Grammar_False;
              break;
            case REA_tag:
              real = value;
              raw_real = real->flo;
              if (raw_previous_number < raw_real)
                return Grammar_False;
              return greater_or_equal_floats(raw_real,
                                             Arguments,
                                             index + 1);
            default:
              return Main_Error_Text(ARN_error,
                                     geq_string); }
        raw_previous_number = signed_number; }
    return Grammar_True; }

static NIL_type greater_or_equal_native(CID_type Context_id,
                                        EXP_type Tailposition)
  { EXP_type boolean,
             value;
    FLO_type raw_real;
    FRM_type arguments;
    REA_type real;
    TAG_type tag;
    UNS_type size;
    arguments = Context_Get_Expression(Context_id);
    size = size_VEC(arguments);
    if (size < 2)
      Main_Error_Text(AL2_error,
                      geq_string);
    value = arguments[1];
    tag = Grammar_Tag(value);
    switch (tag)
      { case NBR_tag:
          boolean = greater_or_equal_integers(value,
                                              arguments);
          break;
        case REA_tag:
          real = value;
          raw_real = real->flo;
          boolean = greater_or_equal_floats(raw_real,
                                            arguments,
                                            2);
          break;
        default:
          boolean = Grammar_False;
          Main_Error_Text(ARN_error,
                          geq_string); }
    Context_Release_Frame(Context_id,
                          arguments);
    Context_Set_Expression(Context_id,
                           boolean); }
    
/*----------------------------------------------------------------------------*/

static NIL_type initialize_greater_or_equal_native_M(NIL_type)
  { native_define_M(geq_string,
                    greater_or_equal_native); }

/*------------------------------------ <= ------------------------------------*/

static const TXT_type leq_string = "<=";

static EXP_type less_or_equal_floats(FLO_type Raw_real,
                                     FRM_type Arguments,
                                     UNS_type Index)
  { EXP_type value;
    FLO_type raw_previous_real,
             raw_real;
    REA_type real;
    TAG_type tag;
    UNS_type index,
             size;
    raw_previous_real = Raw_real;
    size = size_VEC(Arguments);
    for (index = Index;
         index <= size;
         index++)
      { value = Arguments[index];
        tag = Grammar_Tag(value);
        switch (tag)
          { case NBR_tag:
              raw_real = get_NBR(value);
              break;
            case REA_tag:
              real = value;
              raw_real = real->flo;
              break;
            default:
              return Main_Error_Text(ARN_error,
                                     leq_string); }
        if (raw_previous_real > raw_real)
          return Grammar_False;
        raw_previous_real = raw_real; }
   return Grammar_True; }

static EXP_type less_or_equal_integers(NBR_type Number,
                                       FRM_type Arguments)
  { EXP_type value;
    FLO_type raw_real;
    SGN_type signed_number,
             raw_previous_number;
    REA_type real;
    TAG_type tag;
    UNS_type index,
             size;
    raw_previous_number = get_NBR(Number);
    size = size_VEC(Arguments);
    for (index = 2;
         index <= size;
         index++)
      { value = Arguments[index];
        tag = Grammar_Tag(value);
        switch (tag)
          { case NBR_tag:
              signed_number = get_NBR(value);
              if (raw_previous_number > signed_number)
                return Grammar_False;
              break;
            case REA_tag:
              real = value;
              raw_real = real->flo;
              if (raw_previous_number > raw_real)
                return Grammar_False;
              return less_or_equal_floats(raw_real,
                                          Arguments,
                                          index + 1);
            default:
              return Main_Error_Text(ARN_error,
                                     grt_string); }
        raw_previous_number = signed_number; }
    return Grammar_True; }
    
static NIL_type less_or_equal_native(CID_type Context_id,
                                     EXP_type Tailposition)
  { EXP_type boolean,
             value;
    FLO_type raw_real;
    FRM_type arguments;
    REA_type real;
    TAG_type tag;
    UNS_type size;
    arguments = Context_Get_Expression(Context_id);
    size = size_VEC(arguments);
    if (size < 2)
      Main_Error_Text(AL2_error,
                      leq_string);
    value = arguments[1];
    tag = Grammar_Tag(value);
    switch (tag)
      { case NBR_tag:
          boolean = less_or_equal_integers(value,
                                           arguments);
          break;
        case REA_tag:
          real = value;
          raw_real = real->flo;
          boolean = less_or_equal_floats(raw_real,
                                         arguments,
                                         2);
          break;
        default:
          boolean = Grammar_False;
          Main_Error_Text(ARN_error,
                          leq_string); }
    Context_Release_Frame(Context_id,
                          arguments);
    Context_Set_Expression(Context_id,
                           boolean); }
    
/*----------------------------------------------------------------------------*/

static NIL_type initialize_less_or_equal_native_M(NIL_type)
  { native_define_M(leq_string,
                    less_or_equal_native); }

/*------------------------------------ eq? -----------------------------------*/

static const TXT_type eqq_string = "eq?";

static NIL_type equivalent_native(CID_type Context_id,
                                  EXP_type Tailposition)
  { ADR_type left,
             right;
    EXP_type boolean;
    FRM_type arguments;
    UNS_type size;
    arguments = Context_Get_Expression(Context_id);
    size = size_VEC(arguments);
    if (size != 2)
      Main_Error_Text(EX2_error,
                      eqq_string);
    left  = arguments[1],
    right = arguments[2];
    Context_Release_Frame(Context_id,
                          arguments);
    if (left == right)
      boolean = Grammar_True;
    else
     boolean = Grammar_False;
    Context_Set_Expression(Context_id,
                           boolean); }

/*----------------------------------------------------------------------------*/

static NIL_type initialize_equivalent_native_M(NIL_type)
  { native_define_M(eqq_string,
                    equivalent_native); }

/*----------------------------------------------------------------------------*/
/*--------------------------------- booleans ---------------------------------*/
/*----------------------------------------------------------------------------*/

/*------------------------------------ not -----------------------------------*/

static const TXT_type not_string = "not";

static NIL_type not_native(CID_type Context_id,
                           EXP_type Tailposition)
  { EXP_type boolean,
             value;
    FRM_type arguments;
    UNS_type size;
    arguments = Context_Get_Expression(Context_id);
    size = size_VEC(arguments);
    if (size != 1)
      Main_Error_Text(EX1_error,
                      not_string);
    value = arguments[1];
    Context_Release_Frame(Context_id,
                          arguments);
    if (is_FLS(value))
      boolean = Grammar_True;
    else if (is_TRU(value))
      boolean = Grammar_False;
    else
      { boolean = Grammar_False;
        Main_Error_Text(NAB_error,
                        not_string); }
    Context_Set_Expression(Context_id,
                           boolean); }

/*----------------------------------------------------------------------------*/

static NIL_type initialize_not_native_M(NIL_type)
  { native_define_M(not_string,
                    not_native); }

/*----------------------------------------------------------------------------*/
/*-------------------------------- predicates --------------------------------*/
/*----------------------------------------------------------------------------*/

/*----------------------------------- null? ----------------------------------*/

static const TXT_type inl_string = "null?";

static NIL_type is_null_native(CID_type Context_id,
                               EXP_type Tailposition)
  { EXP_type boolean,
             value;
    FRM_type arguments;
    UNS_type size;
    arguments = Context_Get_Expression(Context_id);
    size = size_VEC(arguments);
    if (size != 1)
      Main_Error_Text(EX1_error,
                      inl_string);
    value = arguments[1];
    Context_Release_Frame(Context_id,
                          arguments);
    if (is_NUL(value))
      boolean = Grammar_True;
    else
      boolean = Grammar_False;
    Context_Set_Expression(Context_id,
                           boolean); }

/*----------------------------------------------------------------------------*/

static NIL_type initialize_is_null_native_M(NIL_type)
  { native_define_M(inl_string,
                    is_null_native); }

/*----------------------------------- pair? ----------------------------------*/

static const TXT_type ipr_string = "pair?";

static NIL_type is_pair_native(CID_type Context_id,
                               EXP_type Tailposition)
  { EXP_type boolean,
             value;
    FRM_type arguments;
    UNS_type size;
    arguments = Context_Get_Expression(Context_id);
    size = size_VEC(arguments);
    if (size != 1)
      Main_Error_Text(EX1_error,
                      ipr_string);
    value = arguments[1];
    Context_Release_Frame(Context_id,
                          arguments);
    if (is_PAI(value))
      boolean = Grammar_True;
    else
      boolean = Grammar_False;
    Context_Set_Expression(Context_id,
                           boolean); }

/*----------------------------------------------------------------------------*/

static NIL_type initialize_is_pair_native_M(NIL_type)
  { native_define_M(ipr_string,
                    is_pair_native); }

/*---------------------------------- symbol? ---------------------------------*/

static const TXT_type ism_string = "symbol?";

static NIL_type is_symbol_native(CID_type Context_id,
                                 EXP_type Tailposition)
  { EXP_type boolean,
             value;
    FRM_type arguments;
    UNS_type size;
    arguments = Context_Get_Expression(Context_id);
    size = size_VEC(arguments);
    if (size != 1)
      Main_Error_Text(EX1_error,
                      ism_string);
    value = arguments[1];
    Context_Release_Frame(Context_id,
                          arguments);
    if (is_SYM(value))
      boolean = Grammar_True;
    else
      boolean = Grammar_False;
    Context_Set_Expression(Context_id,
                           boolean); }

/*----------------------------------------------------------------------------*/

static NIL_type initialize_is_symbol_native_M(NIL_type)
  { native_define_M(ism_string,
                    is_symbol_native); }

/*---------------------------------- vector? ---------------------------------*/

static const TXT_type ivc_string = "vector?";

static NIL_type is_vector_native(CID_type Context_id,
                                 EXP_type Tailposition)
  { EXP_type boolean,
             value;
    FRM_type arguments;
    UNS_type size;
    arguments = Context_Get_Expression(Context_id);
    size = size_VEC(arguments);
    if (size != 1)
      Main_Error_Text(EX1_error,
                      ivc_string);
    value = arguments[1];
    Context_Release_Frame(Context_id,
                          arguments);
    if (is_VEC(value))
      boolean = Grammar_True;
    else
      boolean = Grammar_False;
    Context_Set_Expression(Context_id,
                           boolean); }

/*----------------------------------------------------------------------------*/

static NIL_type initialize_is_vector_native_M(NIL_type)
  { native_define_M(ivc_string,
                    is_vector_native); }

/*----------------------------------------------------------------------------*/
/*---------------------------------- pairs -----------------------------------*/
/*----------------------------------------------------------------------------*/

/*----------------------------------- cons -----------------------------------*/

static const TXT_type cns_string = "cons";

static NIL_type cons_native_C(CID_type Context_id,
                              EXP_type Tailposition)
  { EXP_type car,
             cdr;
    FRM_type arguments;
    PAI_type pair;
    UNS_type size;
    Main_Claim_Default();
    arguments = Context_Get_Expression(Context_id);
    size = size_VEC(arguments);
    if (size != 2)
      Main_Error_Text(EX2_error,
                      cns_string);
    car = arguments[1];
    cdr = arguments[2];
    Context_Release_Frame(Context_id,
                          arguments);
    pair = make_PAI_M(car,
                      cdr);
    Main_Release_Default();
    Context_Set_Expression(Context_id,
                           pair); }

/*----------------------------------------------------------------------------*/

static NIL_type initialize_cons_native_M(NIL_type)
  { native_define_M(cns_string,
                    cons_native_C); }

/*------------------------------------ car -----------------------------------*/

static const TXT_type car_string = "car";

static NIL_type car_native(CID_type Context_id,
                           EXP_type Tailposition)
  { EXP_type value;
    FRM_type arguments;
    PAI_type pair;
    UNS_type size;
    arguments = Context_Get_Expression(Context_id);
    size = size_VEC(arguments);
    if (size != 1)
      Main_Error_Text(EX1_error,
                      car_string);
    pair = arguments[1];
    Context_Release_Frame(Context_id,
                          arguments);
    if (!is_PAI(pair))
      Main_Error_Text(NAP_error,
                      car_string);
    value = pair->car;
    Context_Set_Expression(Context_id,
                           value); }

/*----------------------------------------------------------------------------*/

static NIL_type initialize_car_native_M(NIL_type)
  { native_define_M(car_string,
                    car_native); }

/*------------------------------------ cdr -----------------------------------*/

static const TXT_type cdr_string = "cdr";

static NIL_type cdr_native(CID_type Context_id,
                           EXP_type Tailposition)
  { EXP_type value;
    FRM_type arguments;
    PAI_type pair;
    UNS_type size;
    arguments = Context_Get_Expression(Context_id);
    size = size_VEC(arguments);
    if (size != 1)
      Main_Error_Text(EX1_error,
                      cdr_string);
    pair = arguments[1];
    Context_Release_Frame(Context_id,
                          arguments);
    if (!is_PAI(pair))
      Main_Error_Text(NAP_error,
                      cdr_string);
    value = pair->cdr;
    Context_Set_Expression(Context_id,
                           value); }

/*----------------------------------------------------------------------------*/

static NIL_type initialize_cdr_native_M(NIL_type)
  { native_define_M(cdr_string,
                    cdr_native); }

/*--------------------------------- set-car! ---------------------------------*/

static const TXT_type sca_string = "set-car!";

static NIL_type set_car_native(CID_type Context_id,
                               EXP_type Tailposition)
  { EXP_type value;
    FRM_type arguments;
    PAI_type pair;
    UNS_type size;
    arguments = Context_Get_Expression(Context_id);
    size = size_VEC(arguments);
    if (size != 2)
      Main_Error_Text(EX2_error,
                      sca_string);
    pair = arguments[1];
    value = arguments[2];
    Context_Release_Frame(Context_id,
                          arguments);
    if (!is_PAI(pair))
      Main_Error_Text(NAP_error,
                      sca_string);
    pair->car = value;
    Context_Set_Expression(Context_id,
                           pair); }

/*----------------------------------------------------------------------------*/

static NIL_type initialize_set_car_native_M(NIL_type)
  { native_define_M(sca_string,
                    set_car_native); }

/*--------------------------------- set-cdr! ---------------------------------*/

static const TXT_type scd_string = "set-cdr!";

static NIL_type set_cdr_native(CID_type Context_id,
                               EXP_type Tailposition)
  { EXP_type value;
    FRM_type arguments;
    PAI_type pair;
    UNS_type size;
    arguments = Context_Get_Expression(Context_id);
    size = size_VEC(arguments);
    if (size != 2)
      Main_Error_Text(EX2_error,
                      scd_string);
    pair = arguments[1];
    value = arguments[2];
    Context_Release_Frame(Context_id,
                          arguments);
    if (!is_PAI(pair))
      Main_Error_Text(NAP_error,
                      scd_string);
    pair->cdr = value;
    Context_Set_Expression(Context_id,
                           pair); }

/*----------------------------------------------------------------------------*/

static NIL_type initialize_set_cdr_native_M(NIL_type)
  { native_define_M(scd_string,
                    set_cdr_native); }

/*----------------------------------- list -----------------------------------*/

static const TXT_type list_string = "list";

static NIL_type list_native_C(CID_type Context_id,
                              EXP_type Tailposition)
  { EXP_type value;
    FRM_type arguments;
    PAI_type list;
    UNS_type index,
             size;
    arguments = Context_Get_Expression(Context_id);
    size = size_VEC(arguments);
    if (size == 0)
      { Context_Release_Frame(Context_id,
                              arguments);
        Context_Set_Expression(Context_id,
                               Grammar_Null); 
        return; }
    list = Grammar_Null;
    for (index = 1;
         index <= size;
         index++)
      { Context_Save(Context_id,
                     list);
        Main_Claim_Default();
        list = Context_Restore(Context_id);
        arguments = Context_Get_Expression(Context_id);
        value = arguments[index];
        list = make_PAI_M(value,
                          list);
        Main_Release_Default(); }
    list = Main_Reverse(list);
    Context_Release_Frame(Context_id,
                          arguments);
    Context_Set_Expression(Context_id,
                           list); }

/*----------------------------------------------------------------------------*/

static NIL_type initialize_list_native_M(NIL_type)
  { native_define_M(list_string,
                    list_native_C); }

/*----------------------------------- assoc ----------------------------------*/

static const TXT_type ass_string = "assoc";

static NIL_type assoc_native(CID_type Context_id,
                             EXP_type Tailposition)
  { EXP_type value;
    FRM_type arguments;
    PAI_type list,
             pair;
    TAG_type tag;
    UNS_type size;
    arguments = Context_Get_Expression(Context_id);
    size = size_VEC(arguments);
    if (size != 2)
      Main_Error_Text(EX2_error,
                      ass_string);
    value = arguments[1];
    list = arguments[2];
    Context_Release_Frame(Context_id,
                          arguments);
    for (;;)
      { tag = Grammar_Tag(list);
        switch (tag)
          { case NUL_tag:
              Context_Set_Expression(Context_id,
                                     Grammar_False);
              return;
            case PAI_tag:
              pair = list->car;
              if (!is_PAI(pair))
                Main_Error_Text(NAP_error,
                                ass_string);
              if (value == pair->car)
                { Context_Set_Expression(Context_id,
                                         pair);
                  return; }
              list = list->cdr;
              break;
            default:
              Main_Error_Text(NAL_error,
                              ass_string); }}}

/*----------------------------------------------------------------------------*/

static NIL_type initialize_assoc_native_M(NIL_type)
  { native_define_M(ass_string,
                    assoc_native); }

/*----------------------------------------------------------------------------*/
/*--------------------------------- vectors ----------------------------------*/
/*----------------------------------------------------------------------------*/

/*-------------------------------- make-vector -------------------------------*/

static const TXT_type mkv_string = "make-vector";

static NIL_type make_vector_native_C(CID_type Context_id,
                                     EXP_type Tailposition)
  { EXP_type value;
    FRM_type arguments;
    SGN_type signed_number;
    NBR_type number;
    UNS_type index,
             size;
    VEC_type vector;
    arguments = Context_Get_Expression(Context_id);
    size = size_VEC(arguments);
    if (size == 0)
      Main_Error_Text(ALO_error,
                      mkv_string);
    if (size > 2)
      Main_Error_Text(AM2_error,
                      mkv_string);
    if (size == 1)
      value = Grammar_Unspecified;
    else
      value = arguments[2];
    number = arguments[1];
    Context_Release_Frame(Context_id,
                          arguments);
    if (!is_NBR(number))
      Main_Error_Text(NAN_error,
                      mkv_string);
    signed_number = get_NBR(number);
    if (signed_number < 0)
      Main_Error_Text(NNN_error,
                      mkv_string);
    if (signed_number == 0)
      { Context_Set_Expression(Context_id,
                               Grammar_Empty_Vector);
        return; }
    Context_Save(Context_id,
                 value);
    Main_Claim_Dynamic(signed_number);
    value = Context_Restore(Context_id);
    vector = make_VEC_M(signed_number);
    Main_Release_Dynamic(signed_number);
    for (index = 1;
         index <= signed_number;
         index++)
      vector[index] = value;
    Context_Set_Expression(Context_id,
                           vector); }

/*----------------------------------------------------------------------------*/

static NIL_type initialize_make_VEC_Mtor_native_M(NIL_type)
  { native_define_M(mkv_string,
                    make_vector_native_C); }

/*-------------------------------- vector-ref --------------------------------*/

static const TXT_type vrf_string = "vector-ref";

static NIL_type vector_ref_native(CID_type Context_id,
                                  EXP_type Tailposition)
  { EXP_type value;
    FRM_type arguments;
    SGN_type signed_number;
    NBR_type number;
    UNS_type size;
    VEC_type vector;
    arguments = Context_Get_Expression(Context_id);
    size = size_VEC(arguments);
    if (size != 2)
      Main_Error_Text(EX2_error,
                      vrf_string);
    vector = arguments[1];
    number = arguments[2];
    Context_Release_Frame(Context_id,
                          arguments);
    if (!is_VEC(vector))
      Main_Error_Text(NAV_error,
                      vrf_string);
    size = size_VEC(vector);
    if (!is_NBR(number))
      Main_Error_Text(NAN_error,
                      vrf_string);
    signed_number = get_NBR(number);
    if (signed_number < 0)
      Main_Error_Text(NNN_error,
                      vrf_string);
    if (signed_number >= size)
      Main_Error_Text(IOR_error,
                      vrf_string);
    value = vector[signed_number + 1];
    Context_Set_Expression(Context_id,
                           value); }

/*----------------------------------------------------------------------------*/

static NIL_type initialize_vector_ref_native_M(NIL_type)
  { native_define_M(vrf_string,
                    vector_ref_native); }

/*-------------------------------- vector-set! -------------------------------*/

static const TXT_type vst_string = "vector-set!";

static NIL_type vector_set_native(CID_type Context_id,
                                  EXP_type Tailposition)
  { EXP_type value;
    FRM_type arguments;
    SGN_type signed_number;
    NBR_type number;
    UNS_type size;
    VEC_type vector;
    arguments = Context_Get_Expression(Context_id);
    size = size_VEC(arguments);
    if (size != 3)
      Main_Error_Text(EX3_error,
                      vst_string);
    vector = arguments[1];
    number = arguments[2];
    value =  arguments[3];
    Context_Release_Frame(Context_id,
                          arguments);
    if (!is_VEC(vector))
      Main_Error_Text(NAV_error,
                      vst_string);
    size = size_VEC(vector);
    if (!is_NBR(number))
      Main_Error_Text(NAN_error,
                      vst_string);
    signed_number = get_NBR(number);
    if (signed_number < 0)
      Main_Error_Text(NNN_error,
                      vst_string);
    if (signed_number >= size)
      Main_Error_Text(IOR_error,
                      vst_string);
    vector[signed_number + 1] = value;
    Context_Set_Expression(Context_id,
                           vector); }

/*----------------------------------------------------------------------------*/

static NIL_type initialize_vector_set_native_M(NIL_type)
  { native_define_M(vst_string,
                    vector_set_native); }

/*---------------------------------- vector ----------------------------------*/

static const TXT_type vec_string = "vector";

static NIL_type vector_native(CID_type Context_id,
                              EXP_type Tailposition)
  { return; }

/*----------------------------------------------------------------------------*/

static NIL_type initialize_vector_native_M(NIL_type)
  { native_define_M(vec_string,
                    vector_native); }

/*------------------------------- vector-length ------------------------------*/

static const TXT_type vcl_string = "vector-length";

static NIL_type vector_length_native(CID_type Context_id,
                                     EXP_type Tailposition)
  { NBR_type number;
    FRM_type arguments;
    UNS_type size;
    VEC_type vector;
    arguments = Context_Get_Expression(Context_id);
    size = size_VEC(arguments);
    if (size != 1)
      Main_Error_Text(EX1_error,
                      vcl_string);
    vector = arguments[1];
    Context_Release_Frame(Context_id,
                          arguments);
    if (!is_VEC(vector))
      Main_Error_Text(NAV_error,
                      vcl_string);
    size = size_VEC(vector);
    number = make_NBR(size);
    Context_Set_Expression(Context_id,
                           number); }

/*----------------------------------------------------------------------------*/

static NIL_type initialize_vector_length_native_M(NIL_type)
  { native_define_M(vcl_string,
                    vector_length_native); }

/*----------------------------------------------------------------------------*/
/*--------------------------------- control ----------------------------------*/
/*----------------------------------------------------------------------------*/

/*---------------------------------- apply -----------------------------------*/

static const TXT_type apl_string = "apply";

static NIL_type apply_native(CID_type Context_id,
                             EXP_type Tailposition)
  { UNS_type size;
    FRM_type arguments;
    arguments = Context_Get_Expression(Context_id);
    size = size_VEC(arguments);
    if (size != 2)
      Main_Error_Text(EX2_error,
                      apl_string);
    Evaluate_Apply_C(Context_id,               /* reuse <arguments> as vector */
                     Tailposition); }

/*----------------------------------------------------------------------------*/

static NIL_type initialize_apply_native_M(NIL_type)
  { native_define_M(apl_string,
                    apply_native); }

/*--------------------------------- for-each ---------------------------------*/

static const TXT_type for_string = "for-each";

static NBR_type Continue_for_each;

Thread_Begin_Frame(fOR);
  Thread_Frame_Slot(VEC, arg);
  Thread_Frame_Slot(PAI, lst);
Thread_End_Frame(fOR);

static NIL_type continue_for_each_C(CID_type Context_id)
  { fOR_type for_each_thread;
    EXP_type value;
    FRM_type arguments;
    PAI_type list,
             pair;
    TAG_type tag;
    THR_type thread;
    thread = Context_Thread_Peek(Context_id);
    for_each_thread = (fOR_type)thread;
    list      = for_each_thread->lst;
    value = Context_Get_Expression(Context_id);
    tag = Grammar_Tag(list);
    switch (tag)
      { case NUL_tag:
          arguments = for_each_thread->arg;
          Context_Thread_Zap(Context_id);
          Context_Release_Frame(Context_id,
                                arguments);
          Context_Set_Expression(Context_id,
                                 value);
          return;
        case PAI_tag:
          thread = Context_Thread_Keep_C(Context_id);
          for_each_thread = (fOR_type)thread;
          arguments = for_each_thread->arg;
          list      = for_each_thread->lst;
          value = list->car;
          list  = list->cdr;
          pair = arguments[2];
          pair->car = value;
          for_each_thread->lst = list;
          Context_Set_Expression(Context_id,
                                 arguments);
          Evaluate_Apply_C(Context_id,
                           Grammar_False);           /* optimize if last call */
          return;
        default:
          Main_Error_Text(NAL_error,
                          for_string); }}

/*----------------------------------------------------------------------------*/

static NIL_type for_each_native_C(CID_type Context_id,
                                  EXP_type Tailposition)
  { fOR_type for_each_thread;
    EXP_type value;
    FRM_type arguments;
    PAI_type list,
             pair;
    TAG_type tag;
    THR_type thread;
    UNS_type size;
    arguments = Context_Get_Expression(Context_id);
    size = size_VEC(arguments);
    if (size != 2)
      Main_Error_Text(EX2_error,
                      for_string);
    list = arguments[2];
    tag = Grammar_Tag(list);
    switch (tag)
      { case NUL_tag:
          Context_Release_Frame(Context_id,
                                arguments);
          Context_Set_Expression(Context_id,
                                 Grammar_Unspecified);
          return;
        case PAI_tag:
          Context_Thread_Push_C(Context_id,
                                Continue_for_each,
                                Tailposition,
                                fOR_size);
          Main_Claim_Default();
          arguments = Context_Get_Expression(Context_id);
          list = arguments[2];
          value = list->car;
          pair = make_PAI_M(value,
                            Grammar_Null);
          Main_Release_Default();
          list = list->cdr;
          thread = Context_Thread_Peek(Context_id);
          for_each_thread = (fOR_type)thread;
          for_each_thread->arg = arguments;
          for_each_thread->lst = list;
          arguments[2] = pair;                 /* reuse <arguments> as vector */
          Context_Set_Expression(Context_id,
                                 arguments); 
          Evaluate_Apply_C(Context_id,      
                           Grammar_False);         /* optimize if only 1 call */
          return;
        default:
          Main_Error_Text(NAL_error,
                          for_string); }}

/*----------------------------------------------------------------------------*/

static NIL_type initialize_for_each_native_M(NIL_type)
  { Continue_for_each = Context_Register(continue_for_each_C,
                                         fOR_size);
    native_define_M(for_string,
                    for_each_native_C); }

/*----------------------------------- map ------------------------------------*/

static const TXT_type map_string = "map";

static NBR_type Continue_map;

Thread_Begin_Frame(mAP);
  Thread_Frame_Slot(VEC, arg);
  Thread_Frame_Slot(PAI, lst);
  Thread_Frame_Slot(PAI, res);
Thread_End_Frame(mAP);

static NIL_type continue_map_C(CID_type Context_id)
  { mAP_type map_thread;
    EXP_type value;
    FRM_type arguments;
    PAI_type list,
             pair,
             result;
    TAG_type tag;
    THR_type thread;
    Main_Claim_Default();
    thread = Context_Thread_Peek(Context_id);
    map_thread = (mAP_type)thread;
    list      = map_thread->lst;
    result    = map_thread->res;
    value = Context_Get_Expression(Context_id);
    pair = make_PAI_M(value,
                      result);
    Main_Release_Default();
    tag = Grammar_Tag(list);
    switch (tag)
      { case NUL_tag:
          arguments = map_thread->arg;
          Context_Thread_Zap(Context_id);
          result = Main_Reverse(pair);
          Context_Release_Frame(Context_id,
                                arguments);
          Context_Set_Expression(Context_id,
                                 result);
          return;
        case PAI_tag:
          map_thread->res = pair;
          thread = Context_Thread_Keep_C(Context_id);
          map_thread = (mAP_type)thread;
          arguments = map_thread->arg;
          list      = map_thread->lst;
          value = list->car;
          list  = list->cdr;
          map_thread->lst = list;
          pair = arguments[2];
          pair->car = value;
          Context_Set_Expression(Context_id,
                                 arguments);
          Evaluate_Apply_C(Context_id,
                           Grammar_False);           /* optimize if last call */
          return;
        default:
          Main_Error_Text(NAL_error,
                          map_string); }}

/*----------------------------------------------------------------------------*/

static NIL_type map_native_C(CID_type Context_id,
                             EXP_type Tailposition)
  { mAP_type map_thread;
    EXP_type value;
    FRM_type arguments;
    PAI_type list,
             pair;
    TAG_type tag;
    THR_type thread;
    UNS_type size;
    arguments = Context_Get_Expression(Context_id);
    size = size_VEC(arguments);
    if (size != 2)
      Main_Error_Text(EX2_error,
                      map_string);
    list = arguments[2];
    tag = Grammar_Tag(list);
    switch (tag)
      { case NUL_tag:
          Context_Release_Frame(Context_id,
                                arguments);
          Context_Set_Expression(Context_id,
                                 Grammar_Null);
          return;
        case PAI_tag:
          Context_Thread_Push_C(Context_id,
                                Continue_map,
                                Tailposition,
                                mAP_size);
          Main_Claim_Default();
          arguments = Context_Get_Expression(Context_id);
          list = arguments[2];
          value = list->car;
          pair = make_PAI_M(value,
                            Grammar_Null);
          Main_Release_Default();
          list  = list->cdr;
          thread = Context_Thread_Peek(Context_id);
          map_thread = (mAP_type)thread;
          map_thread->arg = arguments;
          map_thread->lst = list;
          map_thread->res = Grammar_Null;
          arguments[2] = pair;
          Context_Set_Expression(Context_id,
                                 arguments);   /* reuse <arguments> as vector */
          Evaluate_Apply_C(Context_id,   
                           Grammar_False);         /* optimize if only 1 call */
          return;
        default:
          Main_Error_Text(NAL_error,
                          map_string); }}

/*----------------------------------------------------------------------------*/

static NIL_type initialize_map_native_M(NIL_type)
  { Continue_map = Context_Register(continue_map_C,
                                    mAP_size);
    native_define_M(map_string,
                    map_native_C); }

/*----------------------------------- eval -----------------------------------*/

static const TXT_type evl_string = "eval";

static NBR_type Continue_eval;

Thread_Begin_Frame(eVA);
  Thread_Frame_Slot(ENV, env);
  Thread_Frame_Slot(FRM, frm);
Thread_End_Frame(eVA);

static NIL_type continue_eval(CID_type Context_id)
  { eVA_type eval_thread;
    ENV_type environment;
    FRM_type frame;
    THR_type thread;
    Context_Release_Current_Frame(Context_id);
    thread = Context_Thread_Peek(Context_id);
    eval_thread = (eVA_type)thread;
    environment = eval_thread->env;
    frame       = eval_thread->frm;
    Context_Thread_Zap(Context_id);
    Context_Environment_Replace(Context_id,
                                environment,
                                frame); }

static NIL_type eval_native_C(CID_type Context_id,
                              EXP_type Tailposition)
  { eVA_type eval_thread;
    ENV_type environment;
    EXP_type compiled_expression,
             expression;
    FRM_type arguments,
             frame;
    THR_type thread;
    UNS_type size;
    arguments = Context_Get_Expression(Context_id);
    size = size_VEC(arguments);
    if (size != 1)
      Main_Error_Text(EX1_error,
                      evl_string);
    expression = arguments[1];
    Context_Release_Frame(Context_id,
                          arguments);
    compiled_expression = Compile_Compile_C(expression);
    Context_Set_Expression(Context_id,
                           compiled_expression);
    thread = Context_Thread_Push_C(Context_id,
                                   Continue_eval,
                                   Grammar_False,
                                   eVA_size);
    eval_thread = (eVA_type)thread;
    environment = Context_Get_Environment(Context_id);
    frame = Context_Get_Frame(Context_id);
    eval_thread->env = environment;
    eval_thread->frm = frame;
    Context_Reset_Environment(Context_id);
    Evaluate_Eval_C(Context_id,
                    Tailposition); }

/*----------------------------------------------------------------------------*/

static NIL_type initialize_eval_native_M(NIL_type)
  { Continue_eval = Context_Register(continue_eval,
                                     eVA_size);
    native_define_M(evl_string,
                    eval_native_C); }

/*--------------------------------- call-cc ----------------------------------*/

static const TXT_type ccc_string = "call-cc";

static NIL_type call_cc_native_C(CID_type Context_id,
                                 EXP_type Tailposition)
  { CNT_type continuation;
    ENV_type environment;
    EXP_type procedure;
    FRM_type arguments,
             frame;
    PAI_type pair;
    THR_type thread;
    UNS_type size;
    VEC_type vector;
    Main_Claim_Default();
    arguments = Context_Get_Expression(Context_id);
    size = size_VEC(arguments);
    if (size != 1)
      Main_Error_Text(EX1_error,
                      ccc_string);
    procedure = arguments[1];
    Context_Release_Frame(Context_id,
                          arguments);
    environment = Context_Get_Environment(Context_id);
    frame = Context_Mark_Frame(Context_id);
    thread = Context_Mark_Thread(Context_id);
    continuation = make_CNT_M(environment,
                              frame,
                              thread);
    pair = make_PAI_M(continuation,
                      Grammar_Null);
    vector = make_VEC_M(2);
    Main_Release_Default();
    vector[1] = procedure;
    vector[2] = pair;
    Context_Set_Expression(Context_id,
                           frame);
    Evaluate_Apply_C(Context_id,
                     Tailposition); }

/*----------------------------------------------------------------------------*/

static NIL_type initialize_call_cc_native_M(NIL_type)
  { native_define_M(ccc_string,
                    call_cc_native_C); }

/*----------------------------------- sync -----------------------------------*/

static const TXT_type snc_string = "sync";

static NIL_type sync_native(CID_type Context_id,
                            EXP_type Tailposition)
  { CID_type context_id;
    EXP_type expression;
    FRM_type arguments;
    PRM_type promise;
    UNS_type size;
    VAL_type value;
    arguments = Context_Get_Expression(Context_id);
    size = size_VEC(arguments);
    if (size != 1)
      Main_Error_Text(EX1_error,
                      snc_string);
    expression = arguments[1];
    Context_Release_Frame(Context_id,
                          arguments);
    if (is_PRM(expression))
      { promise = expression;
        context_id = promise->cid;
        expression = Context_Sync_Task(context_id);
        value = convert_VAL(promise,
                            expression);        
        Context_Set_Expression(Context_id,
                               value); }
    else if (is_VAL(expression))
      { value = expression;
        expression = value->exp;
        Context_Set_Expression(Context_id,
                               expression); }
    else
      Context_Set_Expression(Context_id,                 /* promise required? */
                             expression); }
          
/*----------------------------------------------------------------------------*/

static NIL_type initialize_sync_native_M(NIL_type)
  { native_define_M(snc_string,
                    sync_native); }

/*----------------------------------------------------------------------------*/
/*---------------------------------- input -----------------------------------*/
/*----------------------------------------------------------------------------*/

/*----------------------------------- read -----------------------------------*/

static const TXT_type rdd_string = "read";

static NIL_type read_native_C(CID_type Context_id,
                              EXP_type Tailposition)
  { EXP_type expression;
    FRM_type arguments;
    STR_type name;
    TXT_type input,
             text;
    UNS_type size;
    arguments = Context_Get_Expression(Context_id);
    size = size_VEC(arguments);
    if (size > 1)
      Main_Error_Text(AM1_error,
                      rdd_string);
    if (size == 0)
      text = 0;
    else
      { name = arguments[1];
        Context_Release_Frame(Context_id,
                              arguments);
        if (!is_STR(name))
          Main_Error_Text(NAS_error,
                          rdd_string);
        text = name->txt; }
    input = Context_Read(text);    
    expression = Read_Parse_C(input);
    Context_Set_Expression(Context_id,
                           expression); }

/*----------------------------------------------------------------------------*/

static NIL_type initialize_read_native_M(NIL_type)
  { native_define_M(rdd_string,
                    read_native_C); }

/*----------------------------------------------------------------------------*/
/*---------------------------------- output ----------------------------------*/
/*----------------------------------------------------------------------------*/

/*---------------------------------- display ---------------------------------*/

static const TXT_type dis_string = "display";

static NIL_type display_native(CID_type Context_id,
                               EXP_type Tailposition)
  { EXP_type value;
    FRM_type arguments;
    UNS_type index,
             size;
    arguments = Context_Get_Expression(Context_id);
    size = size_VEC(arguments);
    if (size == 0)
      Main_Error_Text(ALO_error,
                      dis_string);
    for (index = 1;
         index <= size;
         index++)
      { value = arguments[index];
        Print_Display(value); }
    Context_Release_Frame(Context_id,
                          arguments);
    Context_Set_Expression(Context_id,
                           Grammar_Unspecified); }

/*----------------------------------------------------------------------------*/

static NIL_type initialize_display_native_M(NIL_type)
  { native_define_M(dis_string,
                    display_native); }

/*---------------------------------- newline ---------------------------------*/

static const TXT_type nwl_string = "newline";

static NIL_type newline_native(CID_type Context_id,
                               EXP_type Tailposition)
  { UNS_type size;
    FRM_type arguments;
    arguments = Context_Get_Expression(Context_id);
    size = size_VEC(arguments);
    if (size != 0)
      Main_Error_Text(EX0_error,
                      nwl_string);
    Print_Newline();
    Context_Set_Expression(Context_id,
                           Grammar_Unspecified); }

/*----------------------------------------------------------------------------*/

static NIL_type initialize_newline_native_M(NIL_type)
  { native_define_M(nwl_string,
                    newline_native); }

/*---------------------------------- pretty ----------------------------------*/

static const TXT_type prt_string = "pretty";

static NIL_type pretty_native_C(CID_type Context_id,
                                EXP_type Tailposition)
  { EXP_type compiled_expression,
             expression;
    FRM_type arguments;
    UNS_type size;
    arguments = Context_Get_Expression(Context_id);
    size = size_VEC(arguments);
    if (size != 1)
      Main_Error_Text(EX1_error,
                      prt_string);
    expression = arguments[1];
    Context_Release_Frame(Context_id,
                          arguments);
    compiled_expression = Compile_Compile_C(expression);
    Print_Newline();
    Print_Print(compiled_expression);
    Print_Newline();
    Context_Set_Expression(Context_id,
                           Grammar_Unspecified); }

/*----------------------------------------------------------------------------*/

static NIL_type initialize_pretty_native_M(NIL_type)
  { native_define_M(prt_string,
                    pretty_native_C); }

/*----------------------------------------------------------------------------*/
/*--------------------------------- service ----------------------------------*/
/*----------------------------------------------------------------------------*/

/*---------------------------------- clock -----------------------------------*/

static const TXT_type clk_string = "clock";

static NIL_type clock_native(CID_type Context_id,
                             EXP_type Tailposition)
  { NBR_type number;
    FRM_type arguments;
    UNS_type size,
             seconds;
    arguments = Context_Get_Expression(Context_id);
    size = size_VEC(arguments);
    if (size != 0)
      Main_Error_Text(EX0_error,
                      clk_string);
    seconds = time(NULL);
    number = make_NBR(seconds);
    Context_Set_Expression(Context_id,
                           number); }

/*----------------------------------------------------------------------------*/

static NIL_type initialize_clock_native_M(NIL_type)
  { native_define_M(clk_string,
                    clock_native); }

/*---------------------------------- random ----------------------------------*/

static const TXT_type rnd_string = "random";

static NIL_type random_native(CID_type Context_id,
                              EXP_type Tailposition)
  { NBR_type number;
    FRM_type arguments;
    UNS_type size,
             random;
    arguments = Context_Get_Expression(Context_id);
    size = size_VEC(arguments);
    if (size != 0)
      Main_Error_Text(EX0_error,
                      rnd_string);
    random = rand();
    random >>= 2;
    number = make_NBR(random);
    Context_Set_Expression(Context_id,
                           number); }

/*----------------------------------------------------------------------------*/

static NIL_type initialize_random_native_M(NIL_type)
  { native_define_M(rnd_string,
                    random_native); }

/*--------------------------------- collect ----------------------------------*/

static const TXT_type col_string = "collect";

static NIL_type collect_native_C(CID_type Context_id,
                                 EXP_type Tailposition)
  { FRM_type arguments;
    UNS_type size;
    arguments = Context_Get_Expression(Context_id);
    size = size_VEC(arguments);
    if (size != 0)
      Main_Error_Text(EX0_error,
                      col_string);
    Main_Reclaim_C();
    Context_Set_Expression(Context_id,
                           Grammar_Unspecified); }

/*----------------------------------------------------------------------------*/

static NIL_type initialize_collect_native_M(NIL_type)
  { native_define_M(col_string,
                    collect_native_C); }

/*----------------------------------- exit -----------------------------------*/

static const TXT_type exi_string = "exit";

static NIL_type exit_native(CID_type Context_id,
                            EXP_type Tailposition)
  { UNS_type size;
    FRM_type arguments;
    arguments = Context_Get_Expression(Context_id);
    size = size_VEC(arguments);
    if (size != 0)
      Main_Error_Text(EX0_error,
                      exi_string);
    Main_Terminate(); }

/*----------------------------------------------------------------------------*/

static NIL_type initialize_exit_native_M(NIL_type)
  { native_define_M(exi_string,
                    exit_native); }

/*------------------------------ public functions ----------------------------*/

TXT_type Native_Get_Name(UNS_type Offset)
  { return get_native_name(Offset); }

/*----------------------------------------------------------------------------*/
                               
NIL_type Native_Initialize_M(NIL_type)
  { initialize_circularity_level_M();
    initialize_abs_native_M();
    initialize_add_native_M();
    initialize_apply_native_M();
    initialize_assoc_native_M();
    initialize_call_cc_native_M();
    initialize_car_native_M();
    initialize_cdr_native_M();
    initialize_clock_native_M();
    initialize_collect_native_M();
    initialize_cons_native_M();
    initialize_display_native_M();
    initialize_divide_native_M();
    initialize_equal_native_M();
    initialize_equivalent_native_M();
    initialize_eval_native_M();
    initialize_exit_native_M();
    initialize_for_each_native_M();
    initialize_greater_native_M();
    initialize_greater_or_equal_native_M();
    initialize_is_null_native_M();
    initialize_is_pair_native_M();
    initialize_is_symbol_native_M();
    initialize_is_vector_native_M();
    initialize_less_native_M();
    initialize_less_or_equal_native_M();
    initialize_list_native_M();
    initialize_make_VEC_Mtor_native_M();
    initialize_map_native_M();
    initialize_multiply_native_M();
    initialize_newline_native_M();
    initialize_not_native_M();
    initialize_pretty_native_M();
    initialize_quotient_native_M();
    initialize_random_native_M();
    initialize_read_native_M();
    initialize_remainder_native_M();
    initialize_set_car_native_M();
    initialize_set_cdr_native_M();
    initialize_sqrt_native_M();
    initialize_subtract_native_M();
    initialize_sync_native_M();
    initialize_vector_length_native_M();
    initialize_vector_native_M();
    initialize_vector_ref_native_M();
    initialize_vector_set_native_M(); }