                     /*-----------------------------------*/
                     /*             >>>Slip<<<            */
                     /*            Theo D'Hondt           */
                     /*     VUB Software Languages Lab    */
                     /*             (c) 2010              */
                     /*-----------------------------------*/
                     /*      version 14: multicores       */
                     /*-----------------------------------*/
                     /*              reader               */
                     /*-----------------------------------*/

#include <stdlib.h>

#include "SlipMain.h"

#include "SlipPool.h"
#include "SlipRead.h"
#include "SlipScan.h"
#include "SlipStack.h"

/*------------------------------ private prototypes --------------------------*/

static NIL_type read_expression_C(NIL_type);

/*------------------------------ private variables ---------------------------*/

static SCA_type Token;

/*------------------------------ private functions ---------------------------*/

static NIL_type next_token(NIL_type)
  { Token = Scan_Next(); }

/*----------------------------------------------------------------------------*/

static NIL_type read_bracket(NIL_type)
  { Main_Error(IRB_error); }

static NIL_type read_character_C(NIL_type)
  { CHA_type character;
    CHR_type raw_character;
    raw_character = Scan_String[0];
    Main_Claim_Default();
    character = make_CHA_M(raw_character);
    Main_Release_Default();
    Stack_Push_C(character);
    next_token(); }

static NIL_type read_end_program(NIL_type)
  { Main_Error(PEP_error); }

static NIL_type read_false_C(NIL_type)
  { Stack_Push_C(Grammar_False);
    next_token(); }

static NIL_type read_list_C(NIL_type)
  { EXP_type expression;
    PAI_type pair;
    UNS_type count;
    next_token();
    for (count = 0;
         ;
         count++)
      if (Token == RPR_token)
        { Stack_Push_C(Grammar_Null);
          break; }
      else
        { read_expression_C();
          if (Token == PER_token)
            { next_token();
              read_expression_C();
              if (Token != RPR_token)
                Main_Error(MRP_error);
              count += 1;
              break; }}
    next_token();
    for (;
         count > 0;
         count--)
      { Main_Claim_Default();
        pair = Stack_Pop();
        expression = Stack_Peek();
        pair = make_PAI_M(expression,
                          pair);
        Main_Release_Default();
        Stack_Poke(pair); }}

static NIL_type read_number_C(NIL_type)
  { SGN_type signed_number;
    NBR_type number;
    signed_number = atol(Scan_String);
    if (abs(signed_number) > Main_Maximum_Number)
      Main_Error(NTL_error);    
    number = make_NBR(signed_number);
    Stack_Push_C(number);
    next_token(); }

static NIL_type read_parenthesis(NIL_type)
  { Main_Error(IRP_error); }

static NIL_type read_real_C(NIL_type)
  { FLO_type raw_real;
    REA_type real;
    raw_real = atof(Scan_String);
    Main_Claim_Default();
    real = make_REA_M(raw_real);
    Main_Release_Default();
    Stack_Push_C(real);
    next_token(); }

static NIL_type read_period(NIL_type)
  { Main_Error(IUP_error); }

static NIL_type read_quote_C(NIL_type)
  { EXP_type expression;
    PAI_type pair;
    next_token();
		read_expression_C();
    Main_Claim_Default();
    expression = Stack_Peek();
    pair = make_PAI_M(expression,
                      Grammar_Null);
    pair = make_PAI_M(Main_Quote,
                      pair);
    Main_Release_Default();
    Stack_Poke(pair); }

static NIL_type read_string_C(NIL_type)
  { STR_type string;
    UNS_type size;
    size = size_STR(Scan_String);
    Main_Claim_Dynamic(size);
    string = make_STR_M(Scan_String);
    Main_Release_Dynamic(size);
    Stack_Push_C(string);
    next_token(); }

static NIL_type read_symbol_C(NIL_type)
  { SYM_type symbol;
    symbol = Pool_Enter_C(Scan_String);
    Stack_Push_C(symbol);
    next_token(); }

static NIL_type read_true_C(NIL_type)
  { Stack_Push_C(Grammar_True);
    next_token(); }

static NIL_type read_vector_C(NIL_type)
  { EXP_type expression;
    UNS_type count;
    VEC_type vector;
    next_token();
    if (Token == RBR_token)
      vector = Grammar_Empty_Vector;
    else
      { for (count = 0;
             Token != RBR_token;
             count++)
          read_expression_C();
        Main_Claim_Dynamic(count);
        vector = make_VEC_M(count);
        Main_Release_Dynamic(count);
        for (;
             count > 0;
             count--)
          { expression = Stack_Pop();
            vector[count] = expression; }}
    Stack_Push_C(vector);
    next_token(); }

static NIL_type read_expression_C(NIL_type)
  { switch (Token)
      { case CHA_token:
          read_character_C();
          return;
        case END_token:
          read_end_program();
          return;
        case FLS_token:
          read_false_C();
          return;
        case IDT_token:
          read_symbol_C();
          return;
        case LBR_token:
          read_vector_C();
          return;
        case LPR_token:
          read_list_C();
          return;
        case NBR_token:
          read_number_C();
          return;
        case PER_token:
          read_period();
          return;
        case QUO_token:
          read_quote_C();
          return;
        case RBR_token:
          read_bracket();
          return;
        case RPR_token:
          read_parenthesis();
          return;
        case REA_token:
          read_real_C();
          return;
        case STR_token:
          read_string_C();
          return;
        case TRU_token:
          read_true_C();
          return;
        default:
          Main_Error(IXT_error); }}

/*------------------------------ public functions ----------------------------*/

EXP_type Read_Parse_C(TXT_type Input) /* should be atomic! */
  { EXP_type expression;
    Stack_Reset();
    Scan_Preset(Input);
    next_token();
    read_expression_C();
    if (Token != END_token)
      Main_Error(XCT_error);
    expression = Stack_Pop();
    return expression; }

/*----------------------------------------------------------------------------*/

NIL_type Read_Initialize(NIL_type)
  { Scan_Initialize(); }                               
