                     /*-----------------------------------*/
                     /*             >>>Slip<<<            */
                     /*            Theo D'Hondt           */
                     /*     VUB Software Languages Lab    */
                     /*             (c) 2010              */
                     /*-----------------------------------*/
                     /*      version 14: multicores       */
                     /*-----------------------------------*/
                     /*              threads              */
                     /*-----------------------------------*/

#include "SlipMain.h"

#include "SlipThread.h"

/*------------------------------ TPL declarations ----------------------------*/

typedef
  struct TPL { THR_type thr[1]; } TPL;

static TPL_type make_TPL(UNS_type Size)
  { TPL_type thread_pool;
    VEC_type vector;
    vector = make_VEC_M(Size);
    thread_pool = (TPL_type)vector;
    return thread_pool; }
                         
/*------------------------------ THR declarations ----------------------------*/

Begin_Grammar_Frame(THR);
  Define_Grammar_Slot(NBR, tid);
  Define_Grammar_Slot(THR, thr);
  Define_Grammar_Slot(EXP, tps);
  Define_Grammar_Slot(EXP, exp[]);
End_Grammar_Frame(THR);
               
static THR_type make_THR(NBR_type Thread_id,
                         THR_type Thread,
                         EXP_type Tail_position_status,
                         UNS_type Size)
  { THR_type thread;
    UNS_type index;
    VEC_type vector;
    vector = make_VEC_M(THR_size + Size);
    thread = (THR_type)vector;
    thread->tid = Thread_id;
    thread->thr = Thread;
    thread->tps = Tail_position_status;
    for (index = 0;    
         index < Size;
         index++)
      thread->exp[index] = Grammar_Null;
    return thread; }
               
static UNS_type size_THR(THR_type Thread)
  { UNS_type size;
    VEC_type vector;
    vector = (VEC_type)Thread;
    size = size_VEC(vector);
    size -= THR_size;
    return size; }

/*------------------------------ private constants ---------------------------*/

enum { Pool_Size = 6 };
       
/*------------------------------ public variables ----------------------------*/

THR_type Thread_Null;

/*------------------------------ private functions ---------------------------*/

static THR_type acquire_thread_M(TPL_type Thread_pool,
                                 UNS_type Size,
                                 NBR_type Thread_id,
                                 THR_type Thread,
                                 EXP_type Tail_position_status)
  { THR_type next_thread,
             thread;
    UNS_type index;
    index = Size + 1;
    thread = Thread_pool->thr[index];
    if (is_NUL(thread))
      thread = make_THR(Thread_id,
                        Thread,
                        Tail_position_status,
                        Size);
    else
      { next_thread = thread->thr;
        Thread_pool->thr[index] = next_thread;
        thread->tid = Thread_id;
        thread->thr = Thread; }
    return thread; }

static THR_type clone_thread_M(TPL_type Thread_pool,
                               THR_type Thread)
  { EXP_type tail_position_status,
             value;
    NBR_type thread_id;
    THR_type thread;
    UNS_type index,
             size;
    if (!is_MRK(Thread))
      return Thread;
    size = size_THR(Thread);
    thread_id            = Thread->tid;
    tail_position_status = Thread->tps;
    thread               = Thread->thr;
    thread = acquire_thread_M(Thread_pool,
                              size,
                              thread_id,
                              thread,
                              tail_position_status);
    for (index = 0;
         index < size;
         index++)
     { value = Thread->exp[index];
       thread->exp[index] = value; }
    return thread; }

static NIL_type release_thread(TPL_type Thread_pool,
                               THR_type Thread)
  { THR_type thread;
    UNS_type index,
             size;
    if (!is_MRK(Thread))
      { size = size_THR(Thread);
        index = size + 1;
        thread = Thread_pool->thr[index];
        Thread->thr = thread;
        Thread_pool->thr[index] = Thread; }}

static NIL_type flush_pool(TPL_type Thread_pool)
  { UNS_type index;
    for (index = 1;
         index <= Pool_Size;
         index++)
       Thread_pool->thr[index] = Thread_Null; }

/*------------------------------ public functions ----------------------------*/

NIL_type Thread_Flush(TPL_type Thread_pool)
  { flush_pool(Thread_pool); }

NBR_type Thread_Get_Id(THR_type Thread)
  { NBR_type thread_id;
    thread_id = Thread->tid;
    return thread_id; }

EXP_type Thread_Get_Tail(THR_type Thread)
  { EXP_type tail_position_status;
    tail_position_status = Thread->tps;
    return tail_position_status; }

THR_type Thread_Keep_M(THR_type Thread,
                       TPL_type Thread_pool)
  { THR_type thread;
    thread = clone_thread_M(Thread_pool,
                            Thread);
    return thread; }
   
THR_type Thread_Make_M(NBR_type Thread_id,
                       EXP_type Tail_position_status,
                       UNS_type Size,
                       TPL_type Thread_pool)
  { THR_type thread;
    thread = acquire_thread_M(Thread_pool,
                              Size,
                              Thread_id,
                              Thread_Null,
                              Tail_position_status);
    return thread; }
    
NIL_type Thread_Mark(THR_type Thread)
  { THR_type thread;
    for (thread = Thread;
         !is_NUL(thread);
         thread = thread->thr)
      apply_MRK(thread); }

THR_type Thread_Patch_M(THR_type Thread,
                        NBR_type Thread_id,
                        TPL_type Thread_pool)
  { THR_type thread;
    thread = clone_thread_M(Thread_pool,
                            Thread);
    thread->tid = Thread_id;
    return thread; }

NIL_type Thread_Poke(THR_type Thread,
                     THR_type Old_thread,
                     TPL_type Thread_pool)
  { THR_type next_thread;
    next_thread = Old_thread->thr;
    Thread->thr = next_thread;
    release_thread(Thread_pool,
                   Old_thread); }
   
UNS_type Thread_Pool_Size(NIL_type)
  { return Pool_Size + 1; }
  
NIL_type Thread_Push(THR_type New_thread,
                     THR_type Thread)
  { New_thread->thr = Thread; }
   
NIL_type Thread_Release(THR_type Thread,
                        TPL_type Thread_pool)
  { THR_type next_thread,
             thread;
    for (thread = Thread;
         !is_NUL(thread);
         thread = next_thread)
      { next_thread = thread->thr;
        release_thread(Thread_pool,
                       thread); }}

TPL_type Thread_Setup_M(NIL_type)
  { TPL_type thread_pool;
    thread_pool = make_TPL(Pool_Size);
    flush_pool(thread_pool);
    return thread_pool; }

BYT_type Thread_Valid_Size(UNS_type Size)
  { return (Size <= Pool_Size); }

NIL_type Thread_Zap(TRF_type Thread_reference,
                    TPL_type Thread_pool)
  { THR_type next_thread,
             thread;
    thread = *Thread_reference;
    next_thread = thread->thr;
    release_thread(Thread_pool,
                   thread);
    *Thread_reference = next_thread; }

/*----------------------------------------------------------------------------*/

NIL_type Thread_Initialize(NIL_type)
  { Thread_Null = (THR_type)Grammar_Null; }